//
// ElevationGrid class:
//
// simple functionality to support generic grids of elevation data
//

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fstream.h>

#include "dem/ddgegrid.h"

#include "dem/spheroids.h"

extern "C" {
#include "dem/utm.h"
//#include "netcdf.h"
}

#ifndef max
#define max(a,b)            (((a) > (b)) ? (a) : (b))
#endif

#ifndef min
#define min(a,b)            (((a) < (b)) ? (a) : (b))
#endif

#define meters_per_latitude	111300.0f

//////////////////////////////////////////////////

ElevationGrid::ElevationGrid()
{
	m_pData = NULL;
	m_pFData = NULL;
	m_iColumns = 0;
	m_iRows = 0;
	m_bFloatMode = false;
}

ElevationGrid::ElevationGrid(DRECT area, int iColumns, int iRows, bool bFloatMode)
{
	m_area = area;								// raw extents
	m_iColumns = iColumns;
	m_iRows = iRows;
	m_bFloatMode = bFloatMode;
	AllocateArray();
//	m_pData = (short *)calloc(m_iColumns, m_iRows*sizeof(short));
//	m_pFData = NULL;
}

ElevationGrid::ElevationGrid(ElevationGrid *pOld, int iNewZone)
{
	int i, j;

	// many fields are simple to set
	m_bUTM = true;
	m_iUTMZone = iNewZone;
	m_fGRes = pOld->m_fGRes;
	m_fVRes = pOld->m_fVRes;
	strcpy(m_szDemName, pOld->GetName());

	// find where the corners are going to be in the new terrain
	// we assume that the old terrain already has corner information in
	// both UTM and latlon.

	for (i = 0; i < 4; i++)
		m_Corners_lat[i] = pOld->m_Corners_lat[i];
	ConvertCornersToUTM(iNewZone);

	// now, how large an array will we need for the new terrain?
	ComputeExtentsFromCorners();
	double flColumns = (m_area.right - m_area.left)/m_fGRes;
	double flRows = (m_area.top - m_area.bottom)/m_fGRes;

	// round up to the nearest integer
	m_iColumns = (int)(flColumns + 0.999);
	m_iRows = (int)(flRows + 0.999);

	m_bFloatMode = false;
	AllocateArray();

	// convert each bit of data from the old array to the new
	int iOldCols, iOldRows, new_i, new_j;
	DPoint2 p, old_step, lat, step;
	float value; 

	pOld->GetDimensions(iOldCols, iOldRows);
	old_step.x = (pOld->m_area.right - pOld->m_area.left) / (iOldCols - 1);
	old_step.y = (pOld->m_area.top - pOld->m_area.bottom) / (iOldRows - 1);
	step.x = (m_area.right - m_area.left) / (m_iColumns - 1);
	step.y = (m_area.top - m_area.bottom) / (m_iRows - 1);

	for (i = 0; i < iOldCols; i++)
	{
		for (j = 0; j < iOldRows; j++)
		{
			value = pOld->GetFValue(i, j);
			if (value == INVALID_ELEVATION) continue;

			// find UTM coordinates of this datapoint
			p.x = pOld->m_area.left + i * old_step.x;
			p.y = pOld->m_area.bottom + j * old_step.y;

			// convert utm -> latlon
			utminvint(major[0], minor[0], UTM_SCALE, pOld->m_iUTMZone);
			utminv(p.x, p.y, &lat.x, &lat.y);

			// convert latlon -> utm
			utmforint(major[0], minor[0], UTM_SCALE, iNewZone);
			utmfor(lat.x, lat.y, &p.x, &p.y);

			// find place in new grid
			new_i = (int)((p.x - m_area.left) / step.x);
			new_j = (int)((p.y - m_area.bottom) / step.y);

			// set value in new grid
			if (new_i > 0 && new_i < m_iColumns &&
				new_j >= 0 && new_j < m_iRows)
				SetFValue(new_i, new_j, value);
		}
	}
}

ElevationGrid::~ElevationGrid()
{
	if (m_pData)
		free(m_pData);
	if (m_pFData)
		free(m_pFData);
}

void ElevationGrid::ConvertGeoToUTM(ElevationGrid *pOld, int iZone,
									void progress_callback(int am))
{
	int i, j;

	// some fields are simple to set
	m_bUTM = true;
	m_iUTMZone = iZone;
	strcpy(m_szDemName, pOld->GetName());

	// Find where the corners are going to be in the new terrain
	// We assume that the old terrain already has corner information
	// in latlon at least
	for (i = 0; i < 4; i++)
		m_Corners_lat[i] = pOld->m_Corners_lat[i];
	ConvertCornersToUTM(iZone);
	ComputeExtentsFromCorners();

	// Get information about the old terrain
	int iOldCols, iOldRows;
	DPoint2 old_size, old_step;
	pOld->GetDimensions(iOldCols, iOldRows);
	old_size.x = (pOld->m_area.right - pOld->m_area.left);
	old_size.y = (pOld->m_area.top - pOld->m_area.bottom);
	old_step.x = old_size.x / (iOldCols - 1);
	old_step.y = old_size.y / (iOldRows - 1);

	// Get information about the new terrain
	DPoint2 size, step, meter_spacing;
	size.x = (m_area.right - m_area.left);
	size.y = (m_area.top - m_area.bottom);

	meter_spacing.x = old_step.x * meters_per_latitude * cos(m_Corners_lat[0].y / 180.0 * PI);	// convert degrees to meters (approximately)
	meter_spacing.y = old_step.y * meters_per_latitude;	// convert degrees to meters (approximately)
	m_fGRes = meter_spacing.x;
	m_fVRes = pOld->m_fVRes;

	// how large an array will we need for the new terrain?
	double flColumns = size.x/meter_spacing.x;
	double flRows = size.y/meter_spacing.y;

	// round up to the nearest integer
	m_iColumns = (int)(flColumns + 0.9999);
	m_iRows = (int)(flRows + 0.9999);

	m_bFloatMode = pOld->m_bFloatMode;
	AllocateArray();

	// convert each bit of data from the old array to the new
	int new_i, new_j;
	DPoint2 p, lat;
	float value; 

	step.x = size.x / (m_iColumns - 1);
	step.y = size.y / (m_iRows - 1);

	// convert latlon -> utm
	utmforint(major[0], minor[0], UTM_SCALE, iZone);

	for (i = 0; i < iOldCols; i++)
	{
		if (progress_callback != NULL) progress_callback(i*100/iOldCols);
		for (j = 0; j < iOldRows; j++)
		{
			value = pOld->GetFValue(i, j);
			if (value == INVALID_ELEVATION) continue;

			// find Geographic coordinates of this datapoint
			lat.x = pOld->m_area.left + i * old_step.x;
			lat.y = pOld->m_area.bottom + j * old_step.y;

			// convert latlon -> utm
			utmfor_degrees(lat.x, lat.y, &p.x, &p.y);

			// find place in new grid
			new_i = (int)((p.x - m_area.left) / step.x);
			new_j = (int)((p.y - m_area.bottom) / step.y);

			// set value in new grid
			if (new_i >= 0 && new_i < m_iColumns &&
				new_j >= 0 && new_j < m_iRows)
				SetFValue(new_i, new_j, value);
		}
	}
}


// ************** DConvert - DEM Helper fn ****************

double DConvert(FILE *File, int nmax)
{
	int  np;
	char cChar;
	char szCharString [64];

	np=0;
	while ((cChar=fgetc(File)) == ' ')  {
		szCharString[np] = cChar;
		np++;
        }

	szCharString[np] = cChar;
	np++;

	while ((cChar=fgetc(File)) != ' ')  {
        	if (cChar=='D') cChar='E';
		szCharString[np] = cChar;
		np++;
		if (np==nmax) break;
	}  	
	szCharString[np] = 0;

	return atof(szCharString);
}


bool ElevationGrid::LoadFromDEM(const char *szFileName,
								void progress_callback(int am))
{
    int		i, j;
	FILE	*InDem;
	long	lygap;
	int		njunk, nCPoints;
	double	djunk, dxStart, dyStart;
	int		nRow, nColumn;
	int		nElev;
	int		nVUnit, nGUnit;
	double	flColumns, flRows;
	double 	dxdelta, dydelta, dzdelta;
	double	dElevMax, dElevMin;
	int		nElevMin, nElevMax;
	bool	bNewFormat;
	int		nCoordSystem;
	int		nProfiles;
	char	szDateBuffer[5];
	DPoint2	corners[4];			// SW, NW, NE, SE
	DPoint2	extent_min, extent_max;
	double	fGMeters, fVMeters;			// ground and vertical units
	int		nDatum;
	int		iDataStartOffset;

	if (progress_callback != NULL) progress_callback(0);

	if (NULL == (InDem = fopen(szFileName,"rb")))
	{
		// Cannot Open File
		return false;
	}

	// check for version of DEM format
	fseek(InDem, 864, 0);		     			// Read DEM into matrix
	fscanf(InDem, "%d", &nRow);
	fscanf(InDem, "%d", &nColumn);
	bNewFormat = ((nRow!=1)||(nColumn!=1));
	if (bNewFormat)
	{
		fseek(InDem, 1024, 0); 	// New Format
		fscanf(InDem, "%d", &i);
		fscanf(InDem, "%d", &j);
		if ((i!=1)||(j!=1))			// File OK?
		{
			fseek(InDem, 893, 0); 	// Undocumented Format
			fscanf(InDem, "%d", &i);
			fscanf(InDem, "%d", &j);
			if ((i!=1)||(j!=1))			// File OK?
			{
				// Not a DEM file
				fclose(InDem);
				return false;
			}
			else
				iDataStartOffset = 893;
		}
		else
			iDataStartOffset = 1024;
	}
	else
		iDataStartOffset = 864;

	for (i=0; i<80; i++)  m_szDemName[i]=0;
	fseek(InDem, 0, 0);                       		//Read DEM Name
	fgets(m_szDemName, 40, InDem);

	fseek(InDem, 156, 0);
	fscanf(InDem, "%d", &nCoordSystem);
	fscanf(InDem, "%d", &m_iUTMZone);

	if (nCoordSystem == 0)	// geographic (lat-lon)
	{
		m_bUTM = false;
		m_iUTMZone = -1;
	}
	if (nCoordSystem == 1)	// utm
		m_bUTM = true;

	fseek(InDem, 528, 0);
	fscanf(InDem, "%d", &nGUnit);
	fscanf(InDem, "%d", &nVUnit);

	// Ground Units in meters
	if (nGUnit==1) {
		fGMeters = 0.3048;		// 1 = feet
	}
	if (nGUnit==2) { 
		fGMeters = 1.0;			// 2 = meters
	}
	if (nGUnit==3) {
		fGMeters = 30.922;		// 3 = arc-seconds
	}

	// Vertical Units in meters
	if (nVUnit==1)
		fVMeters = 0.3048f;	// feet to meter conversion
	else if (nVUnit==2)
		fVMeters = 1.0f;	// meters == meters
	else
		fVMeters = 1.0f;	// anything else, assume meters

	fseek(InDem, 816, 0);
	dxdelta = DConvert(InDem, 12);
	dydelta = DConvert(InDem, 12);
	dzdelta = DConvert(InDem, 12);

	m_fGRes = dydelta;		   			// G and V Resolution
	m_fVRes = dzdelta;

	if (nVUnit==1 || m_fVRes < 1.0f)
		m_bFloatMode = true;	//store elevation values as floats.
	else
		m_bFloatMode = false;

	// Read the coordinates of the 4 corners
	fseek(InDem, 546, 0);
	for (i = 0; i < 4; i++)
	{
		corners[i].x = DConvert(InDem, 48);
		corners[i].y = DConvert(InDem, 48);
	}

	// find absolute extents of raw vales
	extent_min.x = min(corners[0].x, corners[1].x);
	extent_max.x = max(corners[2].x, corners[3].x);
	extent_min.y = min(corners[0].y, corners[3].y);
	extent_max.y = max(corners[1].y, corners[2].y);

	if (m_bUTM)
	{
		for (i = 0; i < 4; i++)
		{
			m_Corners_utm[i].x = (int)corners[i].x;
			m_Corners_utm[i].y = (int)corners[i].y;
		}
		ConvertCornersToLatlon();
	}
	else
	{
		for (i = 0; i < 4; i++)
		{
			// convert arcseconds to degrees
			m_Corners_lat[i].x = corners[i].x / 3600.0;
			m_Corners_lat[i].y = corners[i].y / 3600.0;
		}
	}

	dElevMin = DConvert(InDem, 48);
	dElevMax = DConvert(InDem, 48);
	nElevMin = (int)dElevMin;
	nElevMax = (int)dElevMax;

	fseek(InDem, 858, 0);
	fscanf(InDem, "%d", &nProfiles);

	// OLD format header ends at byte 864

	if (bNewFormat)
	{
		// year of data compilation
		fseek(InDem, 876, 0);
		fread(szDateBuffer, 4, 1, InDem);
		szDateBuffer[4] = 0;

		// Horizontal datum
		// 1=North American Datum 1927 (NAD 27)
		// 2=World Geodetic System 1972 (WGS 72)
		// 3=WGS 84
		// 4=NAD 83
		// 5=Old Hawaii Datum
		// 6=Puerto Rico Datum
		fseek(InDem, 890, 0);
		fscanf(InDem, "%d", &nDatum);
		switch (nDatum) {
		case 1:
		case 2:
		case 3:
			break;
		case 4:
			//adjustment for kealakekua area.  made need other values for other areas.
			//adjust the extents...
			for (i = 0; i < 4; i++) {
				m_Corners_utm[i].x -=300;
				m_Corners_utm[i].y +=225;
			}
			ConvertCornersToLatlon();			
			break;
		case 5:
		case 6:
			break;
		}
	}

	ComputeExtentsFromCorners();

	if (m_bUTM)
	{
		flColumns = (m_area.right - m_area.left)/dxdelta;
		flRows = (m_area.top - m_area.bottom)/dydelta;

		// round up to the nearest integer
		m_iColumns = (int)(flColumns + 0.999) + 1;
		m_iRows = (int)(flRows + 0.999) + 1;
	}
	else
	{
		flColumns = (float)(m_area.right - m_area.left) * 1200.0f;
		flRows = (float)(m_area.top - m_area.bottom) * 1200.0f;

		// 1 more than you might expect
		m_iColumns = (int)flColumns + 1;
		m_iRows = (int)flRows + 1;
	}

#if 0
	if (bNewFormat)
		fseek(InDem, 1024, 0); 	// New Format
	else
		fseek(InDem, 864, 0);	// Old format
#else
	fseek(InDem, iDataStartOffset, 0);
#endif

	AllocateArray();

	for (i=0; i<m_iColumns; i++)
	{
		if (progress_callback != NULL) progress_callback(i*100/m_iColumns);
		fscanf(InDem, "%d", &njunk);
		fscanf(InDem, "%d", &njunk);
		fscanf(InDem, "%d", &nCPoints);
		fscanf(InDem, "%d", &njunk);
		dxStart = DConvert(InDem, 48);
		dyStart = DConvert(InDem, 48);
		djunk = DConvert(InDem, 48);
		djunk = DConvert(InDem, 48);
		djunk = DConvert(InDem, 48);

		lygap = (long)((dyStart-extent_min.y)/dydelta);

		for (j=(int)lygap; j<(nCPoints+(int)lygap); j++) {
				fscanf(InDem, "%d", &nElev);
				SetFValue(i, j, (float)(nElev*fVMeters*m_fVRes));
		}
	}
	fclose(InDem);

	FindHeightExtents();

	return true;
}

/*
bool ElevationGrid::LoadFromCDF(const char *szFileName,
								void progress_callback(int am))
{
	int id;

	// open existing netCDF dataset 
	int status = nc_open(szFileName, NC_NOWRITE, &id);
	if (status != NC_NOERR)
		return false;

	if (progress_callback != NULL) progress_callback(0);

	// get dimension IDs
	int id_side, id_xysize;
	nc_inq_dimid(id, "side", &id_side);
	nc_inq_dimid(id, "xysize", &id_xysize);
	size_t xysize_length;
	nc_inq_dimlen(id, id_xysize, &xysize_length);

	// get variable IDs
	int id_xrange, id_yrange, id_zrange;
	int id_spacing, id_dimension, id_z;
	nc_inq_varid(id, "x_range", &id_xrange);
	nc_inq_varid(id, "y_range", &id_yrange);
	nc_inq_varid(id, "z_range", &id_zrange);
	nc_inq_varid(id, "spacing", &id_spacing);
	nc_inq_varid(id, "dimension", &id_dimension);
	nc_inq_varid(id, "z", &id_z);

	// get values of variables
	double xrange[2], yrange[2], zrange[2], spacing[2];
	int dimension[2];
	nc_get_var_double(id, id_xrange, xrange);
	nc_get_var_double(id, id_yrange, yrange);
	nc_get_var_double(id, id_zrange, zrange);
	nc_get_var_double(id, id_spacing, spacing);
	nc_get_var_int(id, id_dimension, dimension);

	double *z = new double[xysize_length];
	if (progress_callback != NULL) progress_callback(20);

	nc_get_var_double(id, id_z, z);
	if (progress_callback != NULL) progress_callback(60);

    nc_close(id);				// close netCDF dataset

	// Now copy the values into the ElevationGrid object
	m_iColumns = dimension[0];
	m_iRows = dimension[1];

	m_bFloatMode = false;
	AllocateArray();
	if (progress_callback != NULL) progress_callback(80);

	int i, j;
	for (i = 0; i < m_iColumns; i++)
	{
		for (j = 0; j < m_iRows; j++)
		{
			SetValue(i, m_iRows-1-j, (short)z[j*m_iColumns+i]);
		}
	}
	if (progress_callback != NULL) progress_callback(90);

	m_fGRes = spacing[0];		// G and V Resolution
	m_fGRes *= 3600.0f;			// convert from degrees to arc-seconds
	m_fVRes = 1.0f;				// meters
	m_bUTM = false;
	m_area.left = xrange[0];
	m_area.right = xrange[1];
	m_area.top = yrange[1];
	m_area.bottom = yrange[0];

	ComputeCornersFromExtents();

	// delete temporary storage
	delete z;

	return true;
}
*/

//
// Helper for DTED reader: read an 8-byte number in the
// form DDDMMSSH, degrees-minute-seconds-hemisphere
//
float get_dms8(FILE *fp)
{
	float f;

	char buf[8];
	fread(buf, 8, 1, fp);

	buf[3] = 0;
	f = (float)atof(buf);
	if (buf[7] == 'W') f = -f;
	return f;
}

//
// Read a DTED file.
//
// chunks/bytes:
//  UHL 80
//  DSI 648
//  ACC 2700
//  Data (variable)
//
bool ElevationGrid::LoadFromDTED(const char *szFileName,
								 void progress_callback(int am))
{
	FILE *fp;
	char buf[80];

	if (!(fp = fopen(szFileName, "rb")))
	{
		// Cannot Open File
		return false;
	}

	// all DTEDs are geographic and in integer meters
	// datum is always WGS84
	m_bUTM = false;
	m_bFloatMode = false;

	// check for correct format
	fread(buf, 4, 1, fp);
	if (strncmp(buf, "UHL1", 4))
	{
		// Not a DTED file
		fclose(fp);
		return false;
	}
	m_Corners_lat[0].x = get_dms8(fp);
	m_Corners_lat[0].y = get_dms8(fp);

	// imply other corners
	m_Corners_lat[1].x = m_Corners_lat[0].x;
	m_Corners_lat[1].y = m_Corners_lat[0].y + 1.0f;
	m_Corners_lat[2].x = m_Corners_lat[0].x + 1.0f;
	m_Corners_lat[2].y = m_Corners_lat[0].y + 1.0f;
	m_Corners_lat[3].x = m_Corners_lat[0].x + 1.0f;
	m_Corners_lat[3].y = m_Corners_lat[0].y;
	ComputeExtentsFromCorners();

	// could read spacing here, but just assume 30 arc-second for now
	m_fGRes = 30.0f;
	m_fVRes = 1.0f;

	// get dimensions
	fseek(fp, 47, 0);
	fread(buf, 4, 1, fp);
	buf[4] = 0;
	m_iColumns = atoi(buf);
	fread(buf, 4, 1, fp);
	m_iRows = atoi(buf);

	AllocateArray();

	int line_length = 12 + 2 * m_iRows;
	unsigned char *linebuf = new unsigned char[line_length];
	unsigned char swap[2];

	fseek(fp, 80 + 648 + 2700, 0);
	int i, j, offset;
	for (i = 0; i < m_iColumns; i++)
	{
		if (progress_callback != NULL)
			progress_callback(i * 100 / m_iColumns);

		fread(linebuf, line_length, 1, fp);
		if (*linebuf != 0xaa)
		{
			break;
		}
		offset = 8;
		for (j = 0; j < m_iRows; j++)
		{
			swap[1] = *(linebuf + offset++);
			swap[0] = *(linebuf + offset++);
			int z = *((unsigned short *)swap);
			SetValue(i, j, z);
//			offset += 2;
		}
	}
	delete linebuf;
	return true;
}

//
// Read a GTOPO30 file - actually composed of at least 2 files,
//	a header with a .hdr extension and data with a .dem
//
bool ElevationGrid::LoadFromGTOPO30(const char *szFileName,
									void progress_callback(int am))
{
	if (progress_callback != NULL) progress_callback(1);

	// Open the header file
	ifstream hdrFile(szFileName);
	if (!hdrFile.is_open())
	  return false;

	// Parse the file
	char strName[30];
	char strValue[30];

	GTOPOHeader gh;

	// Read the byte order
	hdrFile >> strName >> gh.ByteOrder;

	// Read the layout
	hdrFile >> strName >> gh.Layout;
	// Read the number of rows
	hdrFile >> strName >> strValue;
	gh.NumRows = atol( strValue );

	// Read the number of columns
	hdrFile >> strName >> strValue;
	gh.NumCols = atol( strValue );

	// Read the number of bands
	hdrFile >> strName >> gh.Bands;

	// Read the number of bits per pixel
	hdrFile >> strName >> gh.Bits;

	// Read the band row bytes
	hdrFile >> strName >> gh.BandRowBytes;

	// Read the total row bytes
	hdrFile >> strName >> gh.TotalRowBytes;

	// Read the band gap bytes
	hdrFile >> strName >> gh.BandGapBytes;

	// Read the no data value
	hdrFile >> strName >> strValue;
	gh.NoData = atoi( strValue );

	// Read the upper left x coordinate
	hdrFile >> strName >> strValue;
	gh.ULXMap = atof( strValue );

	// Read the upper left y coordinate
	hdrFile >> strName >> strValue;
	gh.ULYMap = atof( strValue );

	// Read the x pixel spacing
	hdrFile >> strName >> strValue;
	gh.XDim = atof( strValue );

	// Read the y pixel spacing
	hdrFile >> strName >> strValue;
	gh.YDim = atof( strValue );

	// Close the file
	hdrFile.close();

	// make the corresponding filename for the DEM
	char dem_fname[200];
	strcpy(dem_fname, szFileName);
	char *ext = strrchr(dem_fname, '.');
	if (!ext)
		return false;
	strcpy(ext, ".dem");
	FILE *fp = fopen(dem_fname, "rb");
	if (!fp)
		return false;

	if (progress_callback != NULL) progress_callback(5);

	// set up for an array of the indicated size
	strcpy(m_szDemName, szFileName);
	m_bUTM = false;
	m_bFloatMode = false;
	m_area.left = gh.ULXMap;
	m_area.top = gh.ULYMap;
	m_area.right = m_area.left + (gh.XDim * (gh.NumCols-1));
	m_area.bottom = m_area.top - (gh.YDim * (gh.NumRows-1));
	ComputeCornersFromExtents();
	m_fGRes = 30.0f;	// 30 arc seconds
	m_fVRes = 1.0f;		// height is meters

#if 1
	m_iColumns = gh.NumCols;
	m_iRows = gh.NumRows;
#else
	m_iColumns = gh.NumRows;
	m_iRows = gh.NumCols;
#endif
	int lNumberOfCells = gh.NumCols * gh.NumRows;
	AllocateArray();

	// read the file
//	int dwNumRead = fread(m_pData, sizeof(short), lNumberOfCells, fp);
	int i, j;
	short z;
	char *cp = (char *)&z, temp;
	for (j = 0; j < m_iRows; j++)
	{
		if (progress_callback != NULL) progress_callback(10 + j * 90 / m_iRows);
		for (i = 0; i < m_iColumns; i++)
		{
			fread(&z, sizeof(short), 1, fp);
			temp = cp[0];
			cp[0] = cp[1];
			cp[1] = temp;
			SetValue(i, m_iRows-1-j, (z == gh.NoData) ? 0 : z);
		}
	}

	return false;
}

bool ElevationGrid::LoadFromBT(const char *szFileName)
{
	int i, j;
	int datasize;
	int value;
	float fvalue;
	short svalue;
	float ftmp;
	char buf[10];

	m_bFloatMode = false;

	FILE *fp = fopen(szFileName, "rb");
	if (!fp)
		return false;		// Cannot Open File

	fread(buf, 10, 1, fp);
	if (strncmp(buf, "binterr1.0", 10))
	{
		fclose(fp);
		return false;		// Not a current BT file
	}

	fread(&m_iColumns, 4, 1, fp);
	fread(&m_iRows, 4, 1, fp);
	fread(&datasize, 4, 1, fp);

	// UTM flag
	fread(&svalue, 2, 1, fp);
	m_bUTM = (svalue == 1);

	// UTM zone
	fread(&svalue, 2, 1, fp);
	m_iUTMZone = svalue;

	// coordinate extents left-right
	fread(&ftmp, 4, 1, fp);
	m_area.left = ftmp;
	fread(&ftmp, 4, 1, fp);
	m_area.right = ftmp;

	// coordinate extents bottom-top
	fread(&ftmp, 4, 1, fp);
	m_area.bottom = ftmp;
	fread(&ftmp, 4, 1, fp);
	m_area.top = ftmp;

	ComputeCornersFromExtents();

	//is the data floating point or integers?
	fread(&m_bFloatMode, 4, 1, fp);

	if (m_bFloatMode != true) {
		m_bFloatMode = false;
	}

	fseek(fp, 256, SEEK_SET);

	AllocateArray();

	for (i = 0; i < m_iColumns; i++)
		for (j = 0; j < m_iRows; j++)
		{
			if (m_bFloatMode) {
				fread(&fvalue, 4, 1, fp);
				SetFValue(i, j, fvalue);
			} else {
				if (datasize == 4)
				{
					fread(&value, 4, 1, fp);
					SetValue(i, j, value);
				}
				else
				{
					fread(&svalue, 2, 1, fp);
					SetValue(i, j, svalue);
				}
			}
		}

	// be sure to close the file!
	fclose(fp);

	FindHeightExtents();
	return true;
}


void ElevationGrid::SaveToBTFile(FILE *fp)
{
	int i, j;
	int svalue;
	float fvalue;
	
	int datasize = 2;
	if (m_bFloatMode)
		datasize = 4;

	int w = m_iColumns;
	int h = m_iRows;

	// be careful to save doubles as floats
	float ftmp;

	// header is 256 bytes
	fwrite("binterr1.0", 10, 1, fp);
	fwrite(&w, 4, 1, fp);
	fwrite(&h, 4, 1, fp);
	fwrite(&datasize, 4, 1, fp);

	// UTM true/false
	svalue = m_bUTM;
	fwrite(&svalue, 2, 1, fp);

	// UTM zone
	svalue = m_iUTMZone;
	fwrite(&svalue, 2, 1, fp);

	// min and max longitude
	ftmp = (float)m_area.left;
	fwrite(&ftmp, 4, 1, fp);
	ftmp = (float)m_area.right;
	fwrite(&ftmp, 4, 1, fp);

	// min and max latitude
	ftmp = (float)m_area.bottom;
	fwrite(&ftmp, 4, 1, fp);
	ftmp = (float)m_area.top;
	fwrite(&ftmp, 4, 1, fp);

	//float mode
	fwrite(&m_bFloatMode, 4, 1, fp);

	int filler = 256-10-4-4-4-4-4-4-4-4-4;
	char *zeros = new char[filler];
	for (i = 0; i < filler; i++) {
		zeros[i] = 0;
	}
	fwrite(zeros, filler, 1, fp);

	for (i = 0; i < w; i++) {
		for (j = 0; j < h; j++)
		{
			if (m_bFloatMode) {
				fvalue = GetFValue(i, j);
				fwrite(&fvalue, datasize, 1, fp);
			} else {
				svalue = GetValue(i, j);
				fwrite(&svalue, datasize, 1, fp);
			}
		}
	}
}


void ElevationGrid::FindHeightExtents()
{
	int i, j;

	m_fMinHeight = 100000.0f;
	m_fMaxHeight = -100000.0f;

	for (i=0; i<m_iColumns; i++)
	{
		for (j=0; j<m_iRows; j++)
		{
			float value = GetFValue(i, j);
			if (value > m_fMaxHeight) m_fMaxHeight = value;
			if (value < m_fMinHeight) m_fMinHeight = value;
		}
	}
}


void ElevationGrid::GetDimensions(int &nColumns, int &nRows)
{
	nColumns = m_iColumns;
	nRows = m_iRows;
}

void ElevationGrid::SetValue(int i, int j, short value)
{
	if (m_bFloatMode)
		m_pFData[i*m_iRows+j] = (float)value;
	else
		m_pData[i*m_iRows+j] = value;
}

void ElevationGrid::SetFValue(int i, int j, float value)
{
	if (m_bFloatMode)
		m_pFData[i*m_iRows+j] = value;
	else
		m_pData[i*m_iRows+j] = (short)value;
}

int ElevationGrid::GetValue(int i, int j)
{
	if (m_bFloatMode)
		return (int) m_pFData[i*m_iRows+j];
	else
		return m_pData[i*m_iRows+j];
}

float ElevationGrid::GetFValue(int i, int j)
{
	if (m_bFloatMode)
		return m_pFData[i*m_iRows+j];
	else
		return (float) m_pData[i*m_iRows+j];
}

void ElevationGrid::ConvertCornersToUTM(int iToZone)
{
	// setup transformation from lat-lon
	utmforint(major[0], minor[0], UTM_SCALE, iToZone);

	double x, y;
	for (int i = 0; i < 4; i++)
	{
		utmfor_degrees(m_Corners_lat[i].x, m_Corners_lat[i].y,
					  &x, &y);
		m_Corners_utm[i].x = (int)x;
		m_Corners_utm[i].y = (int)y;
	}
}

void ElevationGrid::ConvertCornersToLatlon()
{
	// setup transformation to lat-lon
	utminvint(major[0], minor[0], UTM_SCALE, m_iUTMZone);

	for (int i = 0; i < 4; i++)
		utminv_degrees(m_Corners_utm[i].x, m_Corners_utm[i].y,
					  &m_Corners_lat[i].x, &m_Corners_lat[i].y);
}

void ElevationGrid::ComputeExtentsFromCorners()
{
	if (m_bUTM)
	{
		// find absolute extents in UTM coordinates
		m_area.left = min(m_Corners_utm[0].x, m_Corners_utm[1].x);
		m_area.right = max(m_Corners_utm[2].x, m_Corners_utm[3].x);
		m_area.bottom = min(m_Corners_utm[0].y, m_Corners_utm[3].y);
		m_area.top = max(m_Corners_utm[1].y, m_Corners_utm[2].y);
	}
	else
	{
		// find absolute extents in degrees
		m_area.left = min(m_Corners_lat[0].x, m_Corners_lat[1].x);
		m_area.right = max(m_Corners_lat[2].x, m_Corners_lat[3].x);
		m_area.bottom = min(m_Corners_lat[0].y, m_Corners_lat[3].y);
		m_area.top = max(m_Corners_lat[1].y, m_Corners_lat[2].y);
	}
}


void ElevationGrid::ComputeCornersFromExtents()
{
	if (m_bUTM)
	{
		m_Corners_utm[0].x = (int)m_area.left;
		m_Corners_utm[0].y = (int)m_area.bottom;
		m_Corners_utm[1].x = (int)m_area.left;
		m_Corners_utm[1].y = (int)m_area.top;
		m_Corners_utm[2].x = (int)m_area.right;
		m_Corners_utm[2].y = (int)m_area.top;
		m_Corners_utm[3].x = (int)m_area.right;
		m_Corners_utm[3].y = (int)m_area.bottom;
	}
	else
	{
		m_Corners_lat[0].x = m_area.left;
		m_Corners_lat[0].y = m_area.bottom;
		m_Corners_lat[1].x = m_area.left;
		m_Corners_lat[1].y = m_area.top;
		m_Corners_lat[2].x = m_area.right;
		m_Corners_lat[2].y = m_area.top;
		m_Corners_lat[3].x = m_area.right;
		m_Corners_lat[3].y = m_area.bottom;
	}
}


//allocates data for array accordingly.
void ElevationGrid::AllocateArray()
{
	if (m_bFloatMode) {
		m_pData = NULL;
		m_pFData = (float *)calloc(m_iColumns, m_iRows*sizeof(float));
	} else {
		m_pData = (short *)calloc(m_iColumns, m_iRows*sizeof(short));
		m_pFData = NULL;
	}

	// Initially no data
	int i, j;
	if (m_bFloatMode)
	{
		for (i = 0; i < m_iColumns; i++)
			for (j = 0; j < m_iRows; j++)
				SetFValue(i, j, INVALID_ELEVATION);
	}
	else
	{
		for (i = 0; i < m_iColumns; i++)
			for (j = 0; j < m_iRows; j++)
				SetValue(i, j, INVALID_ELEVATION);
	}
}
