//
// ElevationGrid
//

#ifndef ELEVATIONGRIDH
#define ELEVATIONGRIDH

#include "dem/tformat_types.h"

typedef struct tagDRECT
{
    double    left;
    double    top;
    double    right;
    double    bottom;
} DRECT;

typedef struct tagFRECT
{
    float    left;
    float    top;
    float    right;
    float    bottom;
} FRECT;

#define INVALID_ELEVATION	-30000

//
//	stores the data points of the elevation grid.
//		data points are stored as either floats or shorts, depending on 
//		the accuracy of the initial data.  if initial data from DEM is in
//		meters, then values are stored as shorts.  if DEM data is in feet,
//		then height data will be stored in float, to preserve the precision
//		of the original data.
class ElevationGrid
{
public:
	ElevationGrid();
	ElevationGrid(DRECT area, int iColumns, int iRows, bool bFloatMode);
	ElevationGrid(ElevationGrid *pOld, int iNewZone);	// create by converting from another UTM zone
	~ElevationGrid();

	bool LoadFromDEM(const char *szFileName, void progress_callback(int) = NULL);
	bool LoadFromCDF(const char *szFileName, void progress_callback(int) = NULL);
	bool LoadFromDTED(const char *szFileName, void progress_callback(int) = NULL);
	bool LoadFromGTOPO30(const char *szFileName, void progress_callback(int) = NULL);
	bool LoadFromBT(const char *szFileName);
	void SaveToBTFile(FILE *fp);

	void FindHeightExtents();
	void GetDimensions(int &nColumns, int &nRows);
	void ConvertGeoToUTM(ElevationGrid *pOld, int iZone,
		void progress_callback(int) = NULL);

	bool FloatMode() {return m_bFloatMode;}

	void SetFValue(int i, int j, float value);
	void SetValue(int i, int j, short value);

	//returns stored value as a integer
	int GetValue(int i, int j);
	//returns stored value as a float
	float GetFValue(int i, int j);

//	short *GetData() { return m_pData; }
	char *GetName() { return m_szDemName; }

	void ConvertCornersToUTM(int iToZone);
	void ConvertCornersToLatlon();
	void ComputeExtentsFromCorners();
	void ComputeCornersFromExtents();
	// public properties
	bool	m_bUTM;
	short	m_iUTMZone;

	// bounds in the original data space
	DRECT	m_area;

	double	m_fGRes;			// Ground Resolution
	double	m_fVRes;			// Vertical Resolution

	DPoint2	m_Corners_lat[4];	// data corners, in lat-lon
	IPoint2	m_Corners_utm[4];	// data corners, in UTM

	// minimum and maximum height values for the whole grid
	float	m_fMinHeight, m_fMaxHeight;

protected:
	void AllocateArray();

	int		m_iColumns;
	int		m_iRows;
	char 	m_szDemName[80];
	bool	m_bFloatMode;
	short	*m_pData;
	float	*m_pFData;
};

//////////////////////////////////////////////

//	Header structure for a GTOPO30 DEM header file
typedef struct  
{
	char			ByteOrder[30];		//	Byte order in which image pixel values are stored 
										//		M = Motorola byte order (MSB first)
	char			Layout[30];			//	Organization of the bands in the file
										//		BIL = Band interleaved by line (DEM = single band image)
	unsigned long	NumRows;			//	Number of rows in the image
	unsigned long	NumCols;			//	Number of columns in the image
	char			Bands[30];			//	Number of spectral bands in the image (1 for DEM)
	char			Bits[30];			//	Number of bits per pixel (16 for DEM)
	char			BandRowBytes[30];	//	Number of bytes per band per row 
										//		(twice the number of columns for 16 bit DEM)
	char			TotalRowBytes[30];	//	Total number of bytes of data per row
										//		(twice the number of columns for 16 bit DEM)
	char			BandGapBytes[30];	//	Number of bytes between bands in a BSQ format
										//		(0 for DEM)
	short			NoData;				//	Value used for no data or masking purposes
	double			ULXMap;				//	Longitude of the center of the upper left pixel in decimal degrees.		
	double			ULYMap;				//	Latitude of the center of the upper left pixel in decimal degrees.		
	double			XDim;				//	x dimension of a pixel in geographic units (decimal degrees).		
	double			YDim;				//	y dimension of a pixel in geographic units (decimal degrees).
} GTOPOHeader;

#endif