/*
    Copyright (C) 1997, 1998, 1999, 200 by Alex Pfaffe, Klaus Hartmann
	(Digital Dawn Graphics Inc)
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	Written by Klaus Hartmann.
*/
#include "ext/ddgeco.h"
#include "math/ddgnoise.h"

// These are our ecosystem definitions. The list is traversed from top to bottom, and
// the first matching ecosystem is used. Hence, order is important. If no matching
// ecosystem can be found then a magenta emergency-color will be displayed.
static ddgEcosystem     g_ecosystems[] =
{
// LEGEND:
//
// USER -> user-defined constant (doesn't need to be unique)
// RGB  -> red, green, blue
// EL   -> elevation line
// MIS  -> minimum slope
// MAS  -> maximum slope
// ESK  -> elevation skew
// ESKA -> elevation skew azimuth
//
//   USER  R    G    B     EL    MIS   MAS   ESK    ESKA   DETAIL-TEXTURE
//#define ECODEBUG
#ifdef ECODEBUG
    {  0,  70, 121,  34, 0.01f, 0.0f, 90.0f, 0.0f, 180.0f, "texture/beach1.tga", },  // BEACH
    {  1,  70, 121,  34, 0.02f, 0.0f, 90.0f, 0.0f, 180.0f, "texture/dirt1.tga", },  // DIRT
    {  2,  70, 121,  34, 0.03f, 0.0f, 90.0f, 0.0f, 180.0f, "texture/dirt2.tga", },  // DIRT
    {  3,  70, 121,  34,  0.2f, 0.0f, 90.0f, 0.0f, 180.0f, "texture/grass1.tga",},  // GRASS
    {  4,  72, 101,   4,  0.3f, 0.0f, 90.0f, 0.0f, 180.0f, "texture/grass2.tga",},  // GRASS
    {  5,  75,  81,   4,  0.4f, 0.0f, 90.0f, 0.0f, 180.0f, "texture/grass3.tga",},  // GRASS
    {  6, 120, 118, 107,  0.5f, 0.0f, 90.0f, 0.0f, 180.0f, "texture/rock1.tga", },  // ROCK
    {  7, 120, 118, 107,  0.6f, 0.0f, 90.0f, 0.0f, 180.0f, "texture/rock2.tga", },  // ROCK
    {  8, 120, 118, 107,  0.7f, 0.0f, 90.0f, 0.0f, 180.0f, "texture/rock3.tga", },  // ROCK
    {  9, 249, 249, 249, 0.72f, 0.0f, 90.0f, 0.0f,   0.0f, "texture/snow1.tga", },  // SNOW
    { 10, 249, 249, 249,  1.0f, 0.0f, 90.0f, 0.0f,   0.0f, "texture/snow2.tga", },  // SNOW
#else
    {  0,  70, 121,  34, 0.02f, 0.0f, 90.0f, 0.0f,   0.0f, "texture/beach1.tga", },  // BEACH
    {  1,  70, 121,  34, 0.03f, 0.0f, 90.0f, 0.2f, 180.0f, "texture/dirt1.tga", },  // DIRT
    {  2,  70, 121,  34, 0.05f, 0.0f, 90.0f, 0.2f, 180.0f, "texture/dirt2.tga", },  // DIRT
    {  3,  70, 121,  34,  0.2f, 0.0f, 90.0f, 0.1f, 180.0f, "texture/grass1.tga",},  // GRASS
    {  4,  72, 101,   4,  0.3f, 0.0f, 90.0f, 0.2f, 180.0f, "texture/grass2.tga",},  // GRASS
    {  5,  75,  81,   4,  0.4f, 0.0f, 90.0f, 0.2f, 180.0f, "texture/grass3.tga",},  // GRASS
    {  6, 120, 118, 107,  0.5f, 0.0f, 90.0f, 0.2f, 180.0f, "texture/rock1.tga", },  // ROCK
    {  7, 120, 118, 107,  0.6f, 0.0f, 90.0f, 0.2f, 180.0f, "texture/rock2.tga", },  // ROCK
    {  8, 120, 118, 107,  0.7f, 0.0f, 90.0f, 0.2f, 180.0f, "texture/rock3.tga", },  // ROCK
    {  9, 249, 249, 249, 0.72f, 0.0f, 90.0f, 0.0f,   0.0f, "texture/snow1.tga", },  // SNOW
    { 10, 249, 249, 249,  1.0f, 0.0f, 90.0f, 0.0f,   0.0f, "texture/snow2.tga", },  // SNOW
#endif
};

// These are our directional light sources.
ddgEcoLightSource g_lights[] =
{
//    DIRX  DIRY  DIRZ   R     G     B
    { 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f }, // SUN LIGHT
    {-1.0f, 1.0f,-1.0f, 0.3f, 0.3f, 0.3f }, // ANTI-SUN LIGHT
};

///////////////////////////////////////////////////////////////////////////////
// ddgEco
///////////////////////////////////////////////////////////////////////////////

//-----------------------------------------------------------------------------
// Name: ddgEco::ddgEco
//
// Desc:
//-----------------------------------------------------------------------------

ddgEco::ddgEco()
{
}

//-----------------------------------------------------------------------------
// Name: ddgEco::~ddgEco
//
// Desc:
//-----------------------------------------------------------------------------

ddgEco::~ddgEco()
{
    m_synthesizer.Destroy();
}

//-----------------------------------------------------------------------------
// Name: ddgEco::Synthesize
//
// Desc:
//-----------------------------------------------------------------------------

bool ddgEco::Synthesize(ddgEcoParams& params)
{
    if (!params.mp_lights)
	{
		params.mp_lights   = g_lights;
		params.m_numLights = sizeof(g_lights) / sizeof(g_lights[0]);
	}
	if (!params.mp_ecosystems)
	{
		params.mp_ecosystems   = g_ecosystems;
		params.m_numEcosystems = sizeof(g_ecosystems) / sizeof(g_ecosystems[0]);
	}
    if (!m_synthesizer.Initialize(params)) return ddgFailure;

    try
    {
        m_synthesizer.Synthesize(params.m_randomAdditions);

        if (params.m_filter) m_synthesizer.FilterTexture();

        m_synthesizer.Destroy();
    }
    catch(...)
    {
        m_synthesizer.Destroy();
        return ddgFailure;
    }

    return ddgSuccess;
}

///////////////////////////////////////////////////////////////////////////////
// ddgEcoSynthesizer
///////////////////////////////////////////////////////////////////////////////

//-----------------------------------------------------------------------------
// Name: ddgEcoSynthesizer::ddgEcoSynthesizer
//
// Desc:
//-----------------------------------------------------------------------------

ddgEcoSynthesizer::ddgEcoSynthesizer() :
    mp_ecosystems(NULL)
{
}

//-----------------------------------------------------------------------------
// Name: ddgEcoSynthesizer::~ddgEcoSynthesizer
//
// Desc:
//-----------------------------------------------------------------------------

ddgEcoSynthesizer::~ddgEcoSynthesizer()
{
    Destroy();
}

//-----------------------------------------------------------------------------
// Name: ddgEcoSynthesizer::Destroy
//
// Desc:
//-----------------------------------------------------------------------------

void ddgEcoSynthesizer::Destroy()
{
    if (mp_ecosystems != NULL)
    {
        for (long n = 0; n < m_numEcosystems; n++)
        {
            if (mp_ecosystems[n].mp_texture != NULL)
            {
                delete mp_ecosystems[n].mp_texture;
            }
        }

        delete[] mp_ecosystems;
        mp_ecosystems = NULL;
    }
}

//-----------------------------------------------------------------------------
// Name: ddgEcoSynthesizer::Initialize
//
// Desc:
//-----------------------------------------------------------------------------

bool ddgEcoSynthesizer::Initialize(ddgEcoParams& params)
{
    float t;

    Destroy();

    mp_heightMap = params.mp_heightMap;
    if (mp_heightMap == NULL) return false;

    mp_baseTexture = params.mp_baseTexture;
    if (mp_baseTexture == NULL) return false;

    mp_ecoMap = params.mp_ecoMap;
    if (mp_ecoMap != NULL)
    {
        mp_ecoMap->allocate(mp_heightMap->rows(), mp_heightMap->cols(), 1);
    }

    mp_lights = params.mp_lights;
    if (mp_lights != NULL) m_numLights = params.m_numLights;
    else m_numLights = 0;

    for (int i = 0; i < m_numLights; i++)
    {
        float denom;

        denom = 1.0f / ::sqrtf(mp_lights[i].x * mp_lights[i].x +
                               mp_lights[i].y * mp_lights[i].y +
                               mp_lights[i].z * mp_lights[i].z);
        mp_lights[i].x *= denom;
        mp_lights[i].y *= denom;
        mp_lights[i].z *= denom;
    }

    m_numEcosystems = params.m_numEcosystems;
    mp_ecosystems   = new ddgEcoListEntry[m_numEcosystems];
	ddgMemorySet(ddgEcoListEntry,m_numEcosystems);

    for (long n = 0; n < m_numEcosystems; n++)
    {
        ddgEcosystem    * p_srcEntry = &params.mp_ecosystems[n];
        ddgEcoListEntry * p_dstEntry = &mp_ecosystems[n];

        p_dstEntry->m_elevLine = p_srcEntry->m_elevLine;

        p_dstEntry->m_user = p_srcEntry->m_user;

        p_dstEntry->m_colorR = p_srcEntry->m_colorR;
        p_dstEntry->m_colorG = p_srcEntry->m_colorG;
        p_dstEntry->m_colorB = p_srcEntry->m_colorB;

        t = p_srcEntry->m_minSlope;
        if (t <= 0.0f) t = 1.0f;
        else if (t >= 90.0f) t = 0.0f;
        else t = ::cosf(ddgAngle::degtorad(t));
        p_dstEntry->m_minSlope = t;

        t = p_srcEntry->m_maxSlope;
        if (t <= 0.0f) t = 1.0f;
        else if (t >= 90.0f) t = 0.0f;
        else t = ::cosf(ddgAngle::degtorad(t));
        p_dstEntry->m_maxSlope = t;

        p_dstEntry->m_elevSkew = p_srcEntry->m_elevSkew;

        t = p_srcEntry->m_elevSkewAzimuth;
        t = ::fmodf(t - 90.0f, 360.0f);
        if (t < 0) t += 360.0f;
        p_dstEntry->m_elevSkewAzimuth = ::cosf(ddgAngle::degtorad(t));

        p_dstEntry->m_elevSkewX = ::cosf(ddgAngle::degtorad(t));
        p_dstEntry->m_elevSkewZ = ::sinf(ddgAngle::degtorad(t));

        p_dstEntry->mp_texture = new ddgImage();
		ddgMemorySet(ddgImage,1);
        p_dstEntry->mp_texture->readTGA(p_srcEntry->mp_texture);
    }

    return true;
}

//-----------------------------------------------------------------------------
// Name: ddgEcoSynthesizer::BiLerp
//
// Desc: Bilinear interpolation.
//-----------------------------------------------------------------------------

float ddgEcoSynthesizer::BiLerp(
    float nwElev,
    float neElev,
    float seElev,
    float swElev,
    float tHorz,
    float tVert)
{
    float   elev, thv;

    thv = tHorz * tVert;

    elev  = nwElev * (1.0f - tVert - tHorz + thv);
    elev += swElev * (tVert - thv);
    elev += neElev * (tHorz - thv);
    elev += seElev * thv;

    return elev;
}

//-----------------------------------------------------------------------------
// Name: ddgEcoSynthesizer::VertexNormal
//
// Desc: Computes the vertex normals at the specified location in the height
//       field. Elevations that lie outside the height field are extrapolated.
//-----------------------------------------------------------------------------

void ddgEcoSynthesizer::VertexNormal(ddgVector3& normal, long row, long col, float gridSpacing)
{
    mp_heightMap->vertexNormal(normal, row, col, gridSpacing);
    return;

    float denom, nx, nz;

    // Compute x-component of normal
    if (col > 0 && col < mp_heightMap->cols() - 1)
    {
        nx = mp_heightMap->getf(row, col - 1) - mp_heightMap->getf(row, col + 1);
    }
    else if (col > 0) 
    {
        nx = 2.0f * (mp_heightMap->getf(row, col - 1) - mp_heightMap->getf(row, col));
    }
    else 
    {
        nx = 2.0f * (mp_heightMap->getf(row, col) - mp_heightMap->getf(row, col + 1));
    }

    // Compute z-component of normal
    if (row > 0 && row < mp_heightMap->rows() - 1)
    {
        nz = mp_heightMap->getf(row - 1, col) - mp_heightMap->getf(row + 1, col);
    }
    else if (row > 0)
    {
        nz = 2.0f * (mp_heightMap->getf(row - 1, col) - mp_heightMap->getf(row, col));
    }
    else 
    {
        nz = 2.0f * (mp_heightMap->getf(row, col) - mp_heightMap->getf(row + 1, col));
    }

    // Normalize the normal
    gridSpacing *= 2.0f;
    denom = 1.0f / ::sqrtf(nx * nx + gridSpacing * gridSpacing + nz * nz);
    normal.v[0] = nx * denom;
    normal.v[1] = gridSpacing * denom;
    normal.v[2] = nz * denom;
}

//-----------------------------------------------------------------------------
// Name: ddgEcoSynthesizer::Illuminate
//
// Desc:
//-----------------------------------------------------------------------------

void ddgEcoSynthesizer::Illuminate(ddgVector3& intensity, ddgVector3& normal)
{
    float I;

    intensity.v[0] = 0.1f;
    intensity.v[1] = 0.1f;
    intensity.v[2] = 0.1f;

    for (long j = 0; j < m_numLights; j++)
    {
        I = mp_lights[j].x * normal.v[0] + mp_lights[j].y * normal.v[1] + mp_lights[j].z * normal.v[2];
        if (I > 0)
        {
            intensity.v[0] += I * mp_lights[j].Ir;
            intensity.v[1] += I * mp_lights[j].Ig;
            intensity.v[2] += I * mp_lights[j].Ib;
        }
    }
}

//-----------------------------------------------------------------------------
// Name: ddgEcoSynthesizer::GetDetailColor
//
// Desc:
//-----------------------------------------------------------------------------

void ddgEcoSynthesizer::GetDetailColor(
    unsigned char * p_color,
    ddgEcoListEntry * p_entry, 
    long texX,
    long texY)
{
    if (p_entry != NULL)
    {
        if (p_entry->mp_texture != NULL)
        {
            ddgImage * p_texture = p_entry->mp_texture;

            texY = texY % p_texture->rows();
            texX = texX % p_texture->cols();
            p_texture->get(texY, texX, &p_color[0], &p_color[1], &p_color[2]);
        }
        else
        {
            p_color[0] = p_entry->m_colorR;
            p_color[1] = p_entry->m_colorG;
            p_color[2] = p_entry->m_colorB;
        }
    }
    else
    {
        p_color[0] = 255;
        p_color[1] = 0;
        p_color[2] = 255;
    }
}

//-----------------------------------------------------------------------------
// Name: ddgEcoSynthesizer::Synthesize
//
// Desc: 
//-----------------------------------------------------------------------------

void ddgEcoSynthesizer::Synthesize(bool randomAdditions)
{
    unsigned long hflWidth  = mp_heightMap->cols();
    unsigned long hflHeight = mp_heightMap->rows();
    unsigned long hflPitch  = mp_heightMap->cols();

    unsigned long texX, texY, texX0, texX1, texY0, texY1;
    float tx, ty, r, g, b;
    unsigned char nw[3], ne[3], se[3], sw[3], index;

    ddgEcoListEntry * p_neEntry, *p_nwEntry, *p_swEntry, *p_seEntry;

    ddgVector3 nwNormal, neNormal, swNormal, seNormal, normal;
    ddgVector3 intensity;

    float nwElev, neElev, swElev, seElev;

    long row, col;

    float minf = mp_heightMap->minf();
    float diff = 1.0f / ::fabsf(mp_heightMap->maxf() - minf);


    double pt[3];


    // For each elevation in the height field
    for (row = 0; row < mp_heightMap->rows() - 1; row++)
    {
        for (col = 0; col < mp_heightMap->cols() - 1; col++)
        {
            // Get the elevations at the four corners of the cell
            nwElev = mp_heightMap->getf(row    , col    );
            neElev = mp_heightMap->getf(row    , col + 1);
            seElev = mp_heightMap->getf(row + 1, col + 1);
            swElev = mp_heightMap->getf(row + 1, col    );

            // Map the elevations into range [0;1]
            nwElev = (nwElev - minf) * diff;
            neElev = (neElev - minf) * diff;
            seElev = (seElev - minf) * diff;
            swElev = (swElev - minf) * diff;

            // Add some random to the elevation
            if (randomAdditions)
            {
                pt[0] = row; pt[1] = col;
                nwElev += ddgNoise::fBm(pt, 1.0f, 1.3f, 3) * 0.1f;

                pt[0] = row; pt[1] = col + 1;
                neElev += ddgNoise::fBm(pt, 1.0f, 1.3f, 3) * 0.1f;

                pt[0] = row + 1; pt[1] = col + 1;
                seElev += ddgNoise::fBm(pt, 1.0f, 1.3f, 3) * 0.1f;

                pt[0] = row + 1; pt[1] = col;
                swElev += ddgNoise::fBm(pt, 1.0f, 1.3f, 3) * 0.1f;
            }

            // Clamp elevations
            if (nwElev > 1) nwElev = 1;
            if (neElev > 1) neElev = 1;
            if (seElev > 1) seElev = 1;
            if (swElev > 1) swElev = 1;

            // Compute the vertex normal for each of the four corner
            // vertices.
            VertexNormal(nwNormal, row,     col    , 1.0f);
            VertexNormal(neNormal, row,     col + 1, 1.0f);
            VertexNormal(seNormal, row + 1, col + 1, 1.0f);
            VertexNormal(swNormal, row + 1, col    , 1.0f);

            // Find the matching ecosystem for each of the four elevations
            p_nwEntry = Classify(nwElev, nwNormal, index);
            p_neEntry = Classify(neElev, neNormal, index);
            p_seEntry = Classify(seElev, seNormal, index);
            p_swEntry = Classify(swElev, swNormal, index);

            // Fill the ecomap, if requested
            if (mp_ecoMap != NULL)
            {
                mp_ecoMap->set(row,     col,     (p_nwEntry != NULL) ? p_nwEntry->m_user : 0xFF);
                mp_ecoMap->set(row,     col + 1, (p_neEntry != NULL) ? p_neEntry->m_user : 0xFF);
                mp_ecoMap->set(row + 1, col + 1, (p_seEntry != NULL) ? p_seEntry->m_user : 0xFF);
                mp_ecoMap->set(row + 1, col,     (p_swEntry != NULL) ? p_swEntry->m_user : 0xFF);
            }

            // Map the cell from height field coordinates to
            // texture map coordinates
            texX0 = (col * mp_baseTexture->cols()) / hflWidth;
            texY0 = (row * mp_baseTexture->rows()) / hflHeight;

            texX1 = ((col + 1) * mp_baseTexture->cols()) / hflWidth;
            texY1 = ((row + 1) * mp_baseTexture->rows()) / hflHeight;

            // For each texel in the cell
            for (texY = texY0; texY < texY1; texY++)
            {
                for (texX = texX0; texX < texX1; texX++)
                {
                    // Compute the position in range [0;1] of the texel inside
                    // the cell.
                    tx = (texX - texX0) / float(texX1 - texX0);
                    ty = (texY - texY0) / float(texY1 - texY0);

                    // Extract the texels from the four involved detail textures
                    GetDetailColor(nw, p_nwEntry, texX, texY);
                    GetDetailColor(ne, p_neEntry, texX, texY);
                    GetDetailColor(se, p_seEntry, texX, texY);
                    GetDetailColor(sw, p_swEntry, texX, texY);

                    // Interpolate the texel's base color from the
                    // four detail texels and illuminate it.
                    r = BiLerp(nw[0], ne[0], se[0], sw[0], tx, ty);
                    g = BiLerp(nw[1], ne[1], se[1], sw[1], tx, ty);
                    b = BiLerp(nw[2], ne[2], se[2], sw[2], tx, ty);

                    // Burn light into texture, if the user specified light
                    // any sources
                    if (m_numLights > 0)
                    {
                        // Interpolate vertex normal for the current texel. This
                        // allows us to do Phong shading.
                        normal.v[0] = BiLerp(nwNormal[0], neNormal[0], seNormal[0], swNormal[0], tx, ty);
                        normal.v[1] = BiLerp(nwNormal[1], neNormal[1], seNormal[1], swNormal[1], tx, ty);
                        normal.v[2] = BiLerp(nwNormal[2], neNormal[2], seNormal[2], swNormal[2], tx, ty);
                        normal.normalize();

                        // Compute the illumination factor
                        Illuminate(intensity, normal);

                        // Illuminate the texel and clamp it into range [0; 255]
                        r *= intensity.v[0];
                        g *= intensity.v[1];
                        b *= intensity.v[2];

                        r = ddgUtil::clamp(r, 0.0f, 255.0f);
                        g = ddgUtil::clamp(g, 0.0f, 255.0f);
                        b = ddgUtil::clamp(b, 0.0f, 255.0f);
                    }

                    // Write the texel into the texture map
                    mp_baseTexture->set(texX, texY, r, g, b, 0);
                }
            }
        }

		ddgConsoleMessage::progress("Synthesizing texture", 1 + row,mp_heightMap->rows());
    }
	ddgConsoleMessage::progress("Synthesizing texture", 1 ,1);
}

//-----------------------------------------------------------------------------
// Name: ddgEcoSynthesizer::FilterTexture
//
// Desc: Applies a simple box filter to the whole texture after it has been
//       generated. 
//
// TODO: This should be replaced with a better filter (maybe a filter matrix).
//-----------------------------------------------------------------------------

void ddgEcoSynthesizer::FilterTexture()
{
    long texX, texY;
    unsigned char r, g, b;
    unsigned long avgR, avgG, avgB;

    for (texY = 1; texY < mp_baseTexture->rows() - 1; texY++)
    {
        for (texX = 1; texX < mp_baseTexture->cols() - 1; texX++)
        {
            mp_baseTexture->get(texY, texX - 1, &r, &g, &b);

            avgR = r; avgG = g; avgB = b;

            mp_baseTexture->get(texY, texX + 1, &r, &g, &b);
            avgR += r; avgG += g; avgB += b;

            mp_baseTexture->get(texY - 1, texX, &r, &g, &b);
            avgR += r; avgG += g; avgB += b;

            mp_baseTexture->get(texY + 1, texX, &r, &g, &b);
            avgR += r; avgG += g; avgB += b;

            avgR >>= 2; avgG >>= 2; avgB >>= 2;

            mp_baseTexture->set(texY, texX, avgR, avgG, avgB);
        }
    }
}

//-----------------------------------------------------------------------------
// Name: ddgEcoSynthesizer::Classify
//
// Desc: 
//-----------------------------------------------------------------------------

ddgEcoListEntry * ddgEcoSynthesizer::Classify(float elev, ddgVector3& normal, unsigned char& index)
{
    long    n = m_numEcosystems;
    float   elevLine, skewAmount, skewDenom, slope;
    bool    skew;


    // The y-component of the normal is a measure for the slope
    slope = normal[1];

    // Compute the factor for normalization
    skewDenom = ::sqrtf(normal[0] * normal[0] + normal[2] * normal[2]);
    if (skewDenom > 0.0001f) 
    {
        skewDenom = 1.0f / skewDenom;
        skew      = true;
    }
    else skew = false;

    // Walk through the list of ecosystems and return the first match
    ddgEcoListEntry * p_eco;

    for (n = 0; n < m_numEcosystems; n++)
    {
        p_eco = &mp_ecosystems[n];

        // Get the elevation limit from the ecosystem
        elevLine = p_eco->m_elevLine;

        // Skew the elevation line
        if (skew)
        {
            skewAmount = p_eco->m_elevSkewX * normal[0] + p_eco->m_elevSkewZ * normal[2];
            elevLine += p_eco->m_elevSkew * skewAmount * skewDenom;
        }

        // Test the input elevation against the elevation line
        if (elev > elevLine) continue;

        // Test whether the input slope falls into the range defined
        // by the ecosystem.
        if (slope <= p_eco->m_minSlope && slope >= p_eco->m_maxSlope)
        {
            index = n;
            // We found a match so return it.
            return p_eco;
        }
    }
   
    // No matching ecosystem
    index = 0xFF;
    return NULL;
}

