///////////////////////////////////////////////////////////////////////////////
// HeightField.cpp
//-----------------------------------------------------------------------------
//
// This file contains the implementation of the height field class.
//-----------------------------------------------------------------------------
//
// Project....: Terrain Generator Demo
// Compiler...: Microsoft Visual C++ 6.0
// Platform...: Microsoft Windows 95/98/2000
///////////////////////////////////////////////////////////////////////////////
#include "ext/ddgsptrr.h"
#include "ext/ddghfld.h"
#include "ext/ddgspnse.h"
#include "ext/ddgfft.h"



///////////////////////////////////////////////////////////////////////////////
// Construction & Destruction
///////////////////////////////////////////////////////////////////////////////

CHeightField::CHeightField() :
    mp_field(NULL)
{
}

CHeightField::~CHeightField()
{
    Destroy();
}

///////////////////////////////////////////////////////////////////////////////
// Height Field Generators
///////////////////////////////////////////////////////////////////////////////

//-----------------------------------------------------------------------------
// Name: Destroy
// Desc: Destroys the height field by freeing allocated memory.
//-----------------------------------------------------------------------------
bool CHeightField::Destroy()
{
    if (mp_field != NULL) delete[] mp_field;
    mp_field = NULL;

    return false;
}

//-----------------------------------------------------------------------------
// Name: Create
// Desc: Creates a height field of the given size and initializes all heights
//       with the specified initial height.
//-----------------------------------------------------------------------------
bool CHeightField::Create(int width, int height, float initialHeight)
{
    Destroy();

    m_fieldSize = width * height;
    if (m_fieldSize == NULL) return Destroy();

    m_fieldWidth  = width;
    m_fieldHeight = height;

    mp_field = new float[m_fieldSize];
    if (mp_field == NULL) return Destroy();

    for (int i = 0; i < m_fieldSize; i++) mp_field[i] = initialHeight;

    m_minHeight = initialHeight;
    m_maxHeight = initialHeight;

    // Success
    return true;
}

//-----------------------------------------------------------------------------
// Name: CreateRidged
// Desc: Creates a ridged multifractal height field.
//-----------------------------------------------------------------------------
bool CHeightField::CreateRidged(
    int width,
    int height,
    float H,
    float lacunarity,
    float octaves,
    float offset,
    float gain)
{
    if (!Create(width, height, 0)) return Destroy();

    CPerlinNoise    noise;

    // Precompute and store spectral weights
    float * p_exponents = new float[int(octaves + 1)];
    if (p_exponents == NULL) return Destroy();

    for (int i = 0; i <= int(octaves); i++)
    {
        p_exponents[i] = ::powf(lacunarity, -float(i) * H);
    }

    for (int row = 0; row < width; row++)
    {
        for (int col = 0; col < height; col++)
        {
            float x = row / float(m_fieldWidth) + float(1);
            float y = col / float(m_fieldHeight) + float(1);

            float signal = noise.Noise2(x, y);

            if (signal < float(0)) signal = -signal;
            signal  = offset - signal;
            signal *= signal;

            float height = signal;
            float weight = float(1);

            for (int i = 1; weight > float(0.001) && i < int(octaves); i++)
            {
                x *= lacunarity;
                y *= lacunarity;

                weight = signal * gain;
                if (weight > float(1)) weight = float(1);
                if (height < float(0)) weight = float(0);

                signal = noise.Noise2(x, y);
                if (signal < float(0)) signal = -signal;
                signal  = offset - signal;
                signal *= weight;

                height += signal * p_exponents[i];
            }

            mp_field[col + row * m_fieldWidth] = height;
        }
    }

    // Delete exponents array
    delete[] p_exponents;

    // Find minimum and maximum heights
    FindMinMaxHeights();

    // Success
    return true;
}

//-----------------------------------------------------------------------------
// Name: CreateSpectral
// Desc: Spectral synthesis of a mountain. The size of the height field is
//       (2^level) + 1 by (2^level) + 1.
//-----------------------------------------------------------------------------
#ifndef PI
    #define PI      3.1415926535897932384626433832795028841971693993751
#endif

bool CHeightField::CreateSpectral(int level, float smoothness)
{
    if (level < 2) return false;

    if (!Create((1 << level) + 1, (1 << level) + 1, 0)) return Destroy();
    
    // Allocate an array of m_fieldSize complex numbers
    float * p_complex = new float[m_fieldSize * 2];
    if (p_complex == NULL) return Destroy();

    // Some locals - mainly used to do some pre-calculations
    float   radTimesSinPhase, radTimesCosPhase, rad, phase;
    int     x, y, x0, y0, ysqr, yofs, y0ofs, index;

    int     fieldWidth = m_fieldWidth - 1;
    int     fieldPitch = fieldWidth << 1;
    float   hDiv2      = -(smoothness + float(1.0)) / float(2.0);
    int     widthHalf  = fieldWidth >> 1;

    // Initialize White Gaussian Noise
    CWhiteGaussianNoise noise;

    for (y = 0; y <= widthHalf; y++)
    {
        y0    = (y != 0) ? fieldWidth - y : 0;
        ysqr  = y * y;
        yofs  = y * fieldPitch;
        y0ofs = y0 * fieldPitch;

        for (x = 0; x <= widthHalf; x++)
        {
            // Compute Fourier coefficients
            phase = float(PI) * float(2) * (rand() / float(RAND_MAX));

            rad = float(0);
            if (x != 0 || y != 0)
            {
                rad = ::powf(float(x * x + ysqr), hDiv2) * noise.Noise();
            }

            radTimesSinPhase = rad * ::sinf(phase);
            radTimesCosPhase = rad * ::cosf(phase);

            // Fill half the array with Fourier coefficients
            p_complex[(x << 1) + yofs + 0] = radTimesCosPhase;
            p_complex[(x << 1) + yofs + 1] = radTimesSinPhase;

            // Fill other half of the array with Fourier coefficents
            x0 = (x != 0) ? fieldWidth - x : 0;

            p_complex[(x0 << 1) + y0ofs + 0] =  radTimesCosPhase;
            p_complex[(x0 << 1) + y0ofs + 1] = -radTimesSinPhase;
        }
    }

    // Imag(fieldWidth / 2, 0) = 0
    p_complex[fieldWidth + 1] = float(0);

    // Imag(0, fieldWidth / 2) = 0
    p_complex[widthHalf * fieldPitch + 1] = float(0);

    // Imag(fieldWidth / 2, fieldWidth / 2) = 0
    p_complex[fieldWidth + widthHalf * fieldPitch + 1] = float(0);

    for (y = 1; y < widthHalf; y++)
    {
        ysqr = y * y;

        for (x = 1; x < widthHalf; x++)
        {
            // Compute Fourier coefficients
            phase = float(PI) * float(2) * (rand() / float(RAND_MAX));
            rad   = ::powf(float(x * x + ysqr), hDiv2) * noise.Noise();

            radTimesSinPhase = rad * ::sinf(phase);
            radTimesCosPhase = rad * ::cosf(phase);

            // Real(x, fieldWidth - y) = radTimesCosPhase
            // Imag(x, fieldWidth - y) = radTimesSinPhase
            index = (x << 1) + (fieldWidth - y) * fieldPitch;
            p_complex[index + 0] = radTimesCosPhase;
            p_complex[index + 1] = radTimesSinPhase;

            // Real(fieldWidth - x, y) = radTimesCosPhase
            // Imag(fieldWidth - x, y) = -radTimesSinPhase
            index = ((fieldWidth - x) << 1) + y * fieldPitch;
            p_complex[index + 0] = radTimesCosPhase;
            p_complex[index + 1] = -radTimesSinPhase;
        }
    }

    // Perform inverse FFT in order to transform from frequency
    // domain to spatial domain.
    int dimensions[2];
    dimensions[0] = fieldWidth;
    dimensions[1] = fieldWidth;
    ::FFT(p_complex - 1, dimensions, 2, true);

    // Copy real parts into the height field and determine the minimum
    // and maximum heights in the height field.
    float minHeight = *p_complex;
    float maxHeight = minHeight;
    float height;

    unsigned int levelMask = (1 << level) - 1;

    for (y = 0; y < m_fieldWidth; y++)
    {
        int yt = fieldWidth - y;
        for (x = 0; x < m_fieldWidth; x++)
        {
            x0 = x & levelMask;
            y0 = y & levelMask;
            height = p_complex[(x0 * 2 + y0 * fieldPitch)];

            mp_field[x + yt * m_fieldWidth] = height;

            if (height < minHeight) minHeight = height;
            if (height > maxHeight) maxHeight = height;
        }
    }

    m_minHeight = minHeight;
    m_maxHeight = maxHeight;

    // Delete the array of complex numbers
    delete[] p_complex;

    // Success
    return true;
}

//-----------------------------------------------------------------------------
// Name: CreateMidpoint
// Desc: Generates a height field by applying midpoint displacement.
//-----------------------------------------------------------------------------
bool CHeightField::CreateMidpoint(int level, float sigma, float H, bool randomAdditions, bool island)
{
    if (!Create((1 << level) + 1, (1 << level) + 1, 0)) return Destroy();

    if (!island) return CreateMDPMountain(level, sigma, H, randomAdditions);
    return CreateMDPIsland(level, sigma, H, randomAdditions);
}

bool CHeightField::CreateMDPMountain(int level, float sigma, float H, bool randomAdditions)
{
    #define F3(x0, x1, x2)          (((x0) + (x1) + (x2)) / 3 + delta * noise.Noise())
    #define F4(x0, x1, x2, x3)      (((x0) + (x1) + (x2) +(x3)) / 4 + delta * noise.Noise())
    #define HEIGHT(xi, yi)          mp_field[(xi)+(yi)*m_fieldWidth]

    int     n, D, d, x, y, stage;
    float   delta;


    // Initialize White Gaussian Noise
    CWhiteGaussianNoise noise;

    n = 1 << level;

    delta = sigma;

    // grid type I: set initial corner heights
    HEIGHT(0,0) = delta * noise.Noise();
    HEIGHT(0,n) = delta * noise.Noise();
    HEIGHT(n,0) = delta * noise.Noise();
    HEIGHT(n,n) = delta * noise.Noise();

    // prepare for first stage
    D = n;
    d = n/2;
    for (stage = 0;stage < level; stage++)
    {
        // going from grid type I to type II
        delta *= ::powf(float(0.5), float(0.5) * H);

        // interpolate and offset points
        for (x = d; x <= n - d; x += D)
        {
            for (y = d; y <= n - d; y += D) 
            {
                HEIGHT(x,y)=F4(HEIGHT(x+d,y+d),HEIGHT(x+d,y-d),HEIGHT(x-d,y+d),HEIGHT(x-d,y-d));
            }
        }

        // displace other points also if needed
        if (randomAdditions)
        {
            for (x = 0; x <= n; x += D)
            {
                for (y = 0; y <= n; y += D) 
                {
                    HEIGHT(x,y)=HEIGHT(x,y)+delta*noise.Noise();
                }
            }
        }

        // going from grid type II to type I
        delta *= ::powf(float(0.5), float(0.5) * H);

        // interpolate and offset boundary grid points
        for (x = d; x <= n - d; x += D)
        {
            HEIGHT(x,0)=F3(HEIGHT(x+d,0),HEIGHT(x-d,0),HEIGHT(x,d));
            HEIGHT(x,n)=F3(HEIGHT(x+d,n),HEIGHT(x-d,n),HEIGHT(x,n-d));
            HEIGHT(0,x)=F3(HEIGHT(0,x+d),HEIGHT(0,x-d),HEIGHT(d,x));
            HEIGHT(n,x)=F3(HEIGHT(n,x+d),HEIGHT(n,x-d),HEIGHT(n-d,x));
        }

        // interpolate and offset interior grid points
        for (x = d;x <= n - d; x += D)
        {
            for (y = D; y <= n - d; y += D) 
            {
                HEIGHT(x,y)=F4(HEIGHT(x,y+d),HEIGHT(x,y-d),HEIGHT(x+d,y),HEIGHT(x-d,y));
            }
        }

        for (x = D; x <= n - d; x += D)
        {
            for (y = d; y <= n - d; y += D) 
            {
                HEIGHT(x,y)=F4(HEIGHT(x,y+d),HEIGHT(x,y-d),HEIGHT(x+d,y),HEIGHT(x-d,y));
            }
        }

        // displace other points also if needed
        if (randomAdditions)
        {
            for (x = 0; x <= n; x += D)
            {
                for (y = 0; y <= n; y += D) HEIGHT(x,y)=HEIGHT(x,y)+delta*noise.Noise();
            }

            for (x = d; x <= n - d; x += D)
            {
                for (y = d; y <= n - d; y += D) HEIGHT(x,y)=HEIGHT(x,y)+delta*noise.Noise();
            }
        }

        // prepare for next stage
        D /= 2;
        d /= 2;
    }

    FindMinMaxHeights();

    #undef F3
    #undef F4
    #undef HEIGHT

    // Success
    return true;
}

bool CHeightField::CreateMDPIsland(int level, float sigma, float H, bool randomAdditions)
{
    #define F3(x0, x1, x2)          (((x0) + (x1) + (x2)) / 3 + delta * noise.Noise())
    #define F4(x0, x1, x2, x3)      (((x0) + (x1) + (x2) +(x3)) / 4 + delta * noise.Noise())
    #define HEIGHT(xi, yi)          mp_field[(xi)+(yi)*m_fieldWidth]

    int     n, D, d, x, y, stage;
    float   delta;


    // Initialize White Gaussian Noise
    CWhiteGaussianNoise noise;

    n = 1 << level;

    delta = sigma;

    // grid type I: set initial corner heights
    HEIGHT(0,0) = float(0);
    HEIGHT(0,n) = float(0);
    HEIGHT(n,0) = float(0);
    HEIGHT(n,n) = float(0);

    // prepare for first stage
    D = n;
    d = n/2;
    for (stage = 0;stage < level; stage++)
    {
        // going from grid type I to type II
        delta *= ::powf(float(0.5), float(0.5) * H);

        // interpolate and offset points
        for (x = d; x <= n - d; x += D)
        {
            for (y = d; y <= n - d; y += D) 
            {
                HEIGHT(x,y)=F4(HEIGHT(x+d,y+d),HEIGHT(x+d,y-d),HEIGHT(x-d,y+d),HEIGHT(x-d,y-d));
            }
        }

        // displace other points also if needed
        if (randomAdditions)
        {
            for (x = D; x <= n - D; x += D)
            {
                for (y = D; y <= n - D; y += D) 
                {
                    HEIGHT(x,y)=HEIGHT(x,y)+delta*noise.Noise();
                }
            }
        }

        // going from grid type II to type I
        delta *= ::powf(float(0.5), float(0.5) * H);

        // interpolate and offset boundary grid points
        for (x = d; x <= n - d; x += D)
        {
            HEIGHT(x,0) = float(0);
            HEIGHT(x,n) = float(0);
            HEIGHT(0,x) = float(0);
            HEIGHT(n,x) = float(0);
        }

        // interpolate and offset interior grid points
        for (x = d;x <= n - d; x += D)
        {
            for (y = D; y <= n - d; y += D) 
            {
                HEIGHT(x,y)=F4(HEIGHT(x,y+d),HEIGHT(x,y-d),HEIGHT(x+d,y),HEIGHT(x-d,y));
            }
        }

        for (x = D; x <= n - d; x += D)
        {
            for (y = d; y <= n - d; y += D) 
            {
                HEIGHT(x,y)=F4(HEIGHT(x,y+d),HEIGHT(x,y-d),HEIGHT(x+d,y),HEIGHT(x-d,y));
            }
        }

        // displace other points also if needed
        if (randomAdditions)
        {
            for (x = D; x <= n - D; x += D)
            {
                for (y = D; y <= n - D; y += D) HEIGHT(x,y)=HEIGHT(x,y)+delta*noise.Noise();
            }

            for (x = d; x <= n - d; x += D)
            {
                for (y = d; y <= n - d; y += D) HEIGHT(x,y)=HEIGHT(x,y)+delta*noise.Noise();
            }
        }

        // prepare for next stage
        D /= 2;
        d /= 2;
    }

    FindMinMaxHeights();

    #undef F3
    #undef F4
    #undef HEIGHT

    // Success
    return true;
}

///////////////////////////////////////////////////////////////////////////////
// Transformations
///////////////////////////////////////////////////////////////////////////////

//-----------------------------------------------------------------------------
// Name: Scale
// Desc: Scales all heights in the height field by the specified scaling
//       factor.
//-----------------------------------------------------------------------------
void CHeightField::Scale(float scaleFactor)
{
    if (mp_field == NULL) return;

    float h;

    m_minHeight = mp_field[0] * scaleFactor;
    m_maxHeight = m_minHeight;

    for (int i = 0; i < m_fieldSize; i++)
    {
        h = mp_field[i] * scaleFactor;
        mp_field[i] = h;

        if (h < m_minHeight) m_minHeight = h;
        else if (h > m_maxHeight) m_maxHeight = h;
    }
}

//-----------------------------------------------------------------------------
// Name: PowerScale
// Desc: Power-scales all heights in the height field by the specified scaling
//       factor.
//-----------------------------------------------------------------------------
void CHeightField::PowerScale(float exponent)
{
    if (mp_field == NULL) return;
    if (exponent == float(0)) return;

    float h;

    h = mp_field[0];
    if (h < 0) h = -::powf(-h, exponent);
    else h = ::powf(h, exponent);

    m_minHeight = h;
    m_maxHeight = h;

    for (int i = 0; i < m_fieldSize; i++)
    {
        h = mp_field[i];

        if (h < 0) h = -::powf(-h, exponent);
        else h = ::powf(h, exponent);

        mp_field[i] = h;

        if (h < m_minHeight) m_minHeight = h;
        else if (h > m_maxHeight) m_maxHeight = h;
    }
}

///////////////////////////////////////////////////////////////////////////////
// Miscellaneous
///////////////////////////////////////////////////////////////////////////////

//-----------------------------------------------------------------------------
// Name: FindMinMaxHeights
// Desc: Finds the minimum and maximum heights in the height field.
//-----------------------------------------------------------------------------
void CHeightField::FindMinMaxHeights()
{
    if (mp_field == NULL) return;

    float h;

    m_minHeight = mp_field[0];
    m_maxHeight = m_minHeight;

    for (int i = 1; i < m_fieldSize; i++)
    {
        h = mp_field[i];
        if (h < m_minHeight) m_minHeight = h;
        else if (h > m_maxHeight) m_maxHeight = h;
    }
}