///////////////////////////////////////////////////////////////////////////////
// Noise.cpp
//-----------------------------------------------------------------------------
//
// This file contains source code for White Gaussian Noise, and Perlin Noise
// over 1, 2, and 3 dimensions.
//-----------------------------------------------------------------------------
//
// Project....: Terrain Generator Demo
// Compiler...: Microsoft Visual C++ 6.0
// Platform...: Microsoft Windows 95/98/2000
///////////////////////////////////////////////////////////////////////////////
#include "ext/ddgsptrr.h"
#include "ext/ddgspnse.h"



#define RECIP_RAND_MAX      float(1) / float(RAND_MAX)

///////////////////////////////////////////////////////////////////////////////
// White Gaussian Noise
///////////////////////////////////////////////////////////////////////////////

CWhiteGaussianNoise::CWhiteGaussianNoise()
{
    m_numSamples  = 4;
    m_gaussAdd    = ::sqrtf(float(3 * m_numSamples));
    m_gaussFactor = float(2) * m_gaussAdd / float(m_numSamples);
}

CWhiteGaussianNoise::CWhiteGaussianNoise(unsigned int numSamples)
{
    m_numSamples  = numSamples;
    m_gaussAdd    = ::sqrtf(float(3 * m_numSamples));
    m_gaussFactor = float(2) * m_gaussAdd / float(m_numSamples);
}

float CWhiteGaussianNoise::Noise()
{
    float   sum   = float(0);
    unsigned int    count = m_numSamples;

    do
    {
        sum += float(rand()) * RECIP_RAND_MAX;
    } while (--count != 0);

    return m_gaussFactor * sum - m_gaussAdd;
}

///////////////////////////////////////////////////////////////////////////////
// Perlin Solid Noise over 1, 2, and 3 dimensions
///////////////////////////////////////////////////////////////////////////////

#define PERLIN_BITMASK       0xFF
#define PERLIN_B             0x100
#define PERLIN_N             0x1000


CPerlinNoise::CPerlinNoise() :
    mp_indices(NULL),
    mp_dim1(NULL),
    mp_dim2(NULL),
    mp_dim3(NULL)
{}

CPerlinNoise::~CPerlinNoise()
{
    if (mp_indices != NULL) delete[] mp_indices;
    if (mp_dim1 != NULL) delete[] mp_dim1;
    if (mp_dim2 != NULL) delete[] mp_dim2;
    if (mp_dim3 != NULL) delete[] mp_dim3;
}

float CPerlinNoise::Noise1(float x)
{
    int       bx0, bx1;
    float    rx0, rx1;
    float    sx, u, v;

    if (mp_indices == NULL) Initialize();

    Setup(x, bx0, bx1, rx0, rx1);

    sx = Spline(rx0);
    u  = rx0 * mp_dim1[mp_indices[bx0]];
    v  = rx1 * mp_dim1[mp_indices[bx1]];

    return Lerp(sx, u, v);
}

float CPerlinNoise::Noise2(float x, float y)
{
    int      bx0, bx1, by0, by1, b00, b01, b10, b11;
    float    rx0, rx1, ry0, ry1;
    float    sx, sy, u, v, a, b;

    if (mp_indices == NULL) Initialize();

    Setup(x, bx0, bx1, rx0, rx1);
    Setup(y, by0, by1, ry0, ry1);

    b00 = mp_indices[mp_indices[bx0] + by0];
    b10 = mp_indices[mp_indices[bx1] + by0];
    b01 = mp_indices[mp_indices[bx0] + by1];
    b11 = mp_indices[mp_indices[bx1] + by1];

    sx = Spline(rx0);
    sy = Spline(ry0);

    u = rx0 * mp_dim2[b00].v[0] + ry0 * mp_dim2[b00].v[1];
    v = rx1 * mp_dim2[b10].v[0] + ry0 * mp_dim2[b10].v[1];
    a = Lerp(sx, u, v);

    u = rx0 * mp_dim2[b01].v[0] + ry1 * mp_dim2[b01].v[1];
    v = rx1 * mp_dim2[b11].v[0] + ry1 * mp_dim2[b11].v[1];
    b = Lerp(sx, u, v);

    return Lerp(sy, a, b);
}

float CPerlinNoise::Noise3(float x, float y, float z)
{
    int       bx0, bx1, by0, by1, bz0, bz1, b00, b01, b10, b11;
    float    rx0, rx1, ry0, ry1, rz0, rz1;
    float    sx, sy, sz, u, v, a, b, c, d;
    float *  p;

    if (mp_indices == NULL) Initialize();

    Setup(x, bx0, bx1, rx0, rx1);
    Setup(y, by0, by1, ry0, ry1);
    Setup(z, bz0, bz1, rz0, rz1);

    b00 = mp_indices[mp_indices[bx0] + by0];
    b10 = mp_indices[mp_indices[bx1] + by0];
    b01 = mp_indices[mp_indices[bx0] + by1];
    b11 = mp_indices[mp_indices[bx1] + by1];

    sx = Spline(rx0);
    sy = Spline(ry0);
    sz = Spline(rz0);

    p = &(mp_dim3[b00 + bz0].v[0]);
    u = rx0 * p[0] + ry0 * p[1] + rz0 * p[2];
    p = &(mp_dim3[b10 + bz0].v[0]);
    v = rx1 * p[0] + ry0 * p[1] + rz0 * p[2];
    a = Lerp(sx, u, v);

    p = &(mp_dim3[b01 + bz0].v[0]);
    u = rx0 * p[0] + ry1 * p[1] + rz0 * p[2];
    p = &(mp_dim3[b11 + bz0].v[0]);
    v = rx1 * p[0] + ry1 * p[1] + rz0 * p[2];
    b = Lerp(sx, u, v);

    c = Lerp(sy, a, b);

    p = &(mp_dim3[b00 + bz1].v[0]);
    u = rx0 * p[0] + ry0 * p[1] + rz1 * p[2];
    p = &(mp_dim3[b10 + bz1].v[0]);
    v = rx1 * p[0] + ry0 * p[1] + rz1 * p[2];
    a = Lerp(sx, u, v);

    p = &(mp_dim3[b01 + bz1].v[0]);
    u = rx0 * p[0] + ry1 * p[1] + rz1 * p[2];
    p = &(mp_dim3[b11 + bz1].v[0]);
    v = rx1 * p[0] + ry1 * p[1] + rz1 * p[2];
    b = Lerp(sx, u, v);

    d = Lerp(sz, a, b);

    return Lerp(sz, c, d);
}

bool CPerlinNoise::Initialize()
{
    int   i, j, k;

    mp_indices = new int[PERLIN_B + PERLIN_B + 2];
    mp_dim1    = new OiDim1[PERLIN_B + PERLIN_B + 2];
    mp_dim2    = new OiDim2[PERLIN_B + PERLIN_B + 2];
    mp_dim3    = new OiDim3[PERLIN_B + PERLIN_B + 2];

    // Create random vectors with all components in range [-1; 1], and
    // normalize them afterwards.
    for (i = 0; i < PERLIN_B; i++)
    {
        // Setup index array
        mp_indices[i] = i;

        // 1-D vectors (cannot be normalized)
        mp_dim1[i] = float(rand()) * RECIP_RAND_MAX * float(2) - float(1);

        // 2-D vectors
        for (j = 0; j < 2; j++) mp_dim2[i].v[j] = float(rand()) * RECIP_RAND_MAX * float(2) - float(1);
        Normalize2D(&(mp_dim2[i].v[0]));

        // 3-D vectors
        for (j = 0; j < 3; j++) mp_dim3[i].v[j] = float(rand()) * RECIP_RAND_MAX * float(2) - float(1);
        Normalize3D(&(mp_dim3[i].v[0]));
    }

    // Shuffle the index array
    while (--i)
    {
        j = rand() % PERLIN_B;
        k = mp_indices[i];
        mp_indices[i] = mp_indices[j];
        mp_indices[j] = k;
    }

    for (i = 0; i < PERLIN_B + 2; i++)
    {
        mp_indices[PERLIN_B + i] = mp_indices[i];

        mp_dim1[PERLIN_B + i] = mp_dim1[i];
        for (j = 0; j < 2; j++) mp_dim2[PERLIN_B + i].v[j] = mp_dim2[i].v[j];
        for (j = 0; j < 3; j++) mp_dim3[PERLIN_B + i].v[j] = mp_dim3[i].v[j];
    }

    // Success
    return true;
}

void CPerlinNoise::Normalize2D(float * p_v)
{
    float denom = ::sqrtf(p_v[0] * p_v[0] + p_v[1] * p_v[1]);
    if (denom > 0.0001f) denom = float(1) / denom;

    p_v[0] *= denom;
    p_v[1] *= denom;
}

void CPerlinNoise::Normalize3D(float * p_v)
{
    float denom = ::sqrtf(p_v[0] * p_v[0] + p_v[1] * p_v[1] + p_v[2] * p_v[2]);
    if (denom > 0.0001f) denom = float(1) / denom;

    p_v[0] *= denom;
    p_v[1] *= denom;
    p_v[2] *= denom;
}

void CPerlinNoise::Setup(float v, int& b0, int& b1, float& r0, float& r1)
{
    float t = v + PERLIN_N;
    
    b0 = int(t) & PERLIN_BITMASK;
    b1 = (b0 + 1) & PERLIN_BITMASK;
    r0 = t - int(t);
    r1 = r0 - float(1);
}