/*
    Copyright (C) 1997, 1998, 1999, 2000 by Alex Pfaffe
	(Digital Dawn Graphics Inc)
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#ifndef _ddgCamera_Class
#define _ddgCamera_Class

#include "struct/ddgcntxt.h"
#include "render/ddgscene.h"

/**
 * Camera object to manage a scene and connect to a viewport.
 *
 * Class which supports OpenGL camera capabilities.
 * Contains parameters to control various settings
 * Also registers keys for camera motion control.
 *
 * Should define a set of activities which camera can perform.
 * These activities can then be mapped to events.  ddgVector of
 * events is then mapped to activities to allow alternate key mappings.
 */
class WEXP ddgCamera
{
    /// Constant used for ssdistance, calculate once per frame.
    float   _kk;
    /// The scene which this camera will render.
    ddgScene*			_scene;
	/// The context for this camera.
	ddgContext			*_ctx;
	/// Screen space 2D clipping rectangle.
	int				_vp[4];
	/// Is this camera in ortho graphic mode.
	bool			_ortho;
	/// Size of orthographic window.
	unsigned int	_width,
					_height;
  public:

    /**
	 *  Create a camera object.
     */
    ddgCamera( ddgContext *ctx );
    /// Set the scene which this camera will render.
    void scene( ddgScene* s ) { _scene = s; }
    /// Callback which is called to initialize the camera.
    bool init(void);
    /// Render the scene maintained by this camera.
    void draw(void);
    /// Update the camera for the next frame.
    void update(void);

	/**
	 *  Map a 3D world coordinate to view space.
	 *  vin is in world space, vout is in view space,
	 *  Returns false if point is not on view and
	 *  vout is left unchanged.  (Uses OpenGL)
     */
	bool map(ddgVector3 vin, ddgVector3 *vout);
	/**
	 *  Map a view coordinate to world space.
	 *  vin is in view space, vout is in world space,
	 *  Returns false if point is not in view and
	 *  vout is left unchanged. (Uses OpenGL)
     */
	bool unmap(ddgVector3 vin, ddgVector3 *vout);
	/// Set up an ortho graphic camera.
	void ortho( unsigned int w, unsigned int h);
	/// Return the window size.	vp returns the viewport dimensions in pixels.
	void viewport( int vp[4] ) { vp[0] = _vp[0]; vp[1] = _vp[1]; vp[2] = _vp[2]; vp[3] = _vp[3]; }
	/// Return the context for this camera.
	ddgContext *context(void) { return _ctx; }
}; 


#endif
