/*
    Copyright (C) 1997, 1998, 1999, 2000 by Alex Pfaffe
	(Digital Dawn Graphics Inc)
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#ifndef _ddgFacet_Class_
#define _ddgFacet_Class_

#include "math/ddgvec.h"
#include "struct/ddgnode.h"
#include "util/ddgerror.h"

// This class stores static information needed to perform
// spherical classification and uses this to perform culling
// operations.
// Algorithm to optimize backface culling of large polygon sets.
// Define a static array of the normal vectors (sN) of an icosahedron.
// This represents and evenly spaced set of normals (20) in R3.
// Enumerate sN in such a way that vectors i and (i+10)%20 = <0,0,0> holds
// true. Two vector pointing in opposite directions.
//
// Preprocessing step for each triangle in the polyset:
// Calculate the normal tN and find out in which sector sN the triangle should
// be associated with based on the minimum angle between tN and sN.
// Add triangle t into the appropriate category.
//
// Memory overhead 20 arrays with a combined storage space of n integer
// indices into the triangle array.
//
// At runtime determine which sector most faces the camera and render the
// fixed set of arrays which are associated with that sector.
// For icosahedron
#include "render/ddgvbuf.h"
/**
 * Class to represent a facet on an icosahedron or
 * a regular subtriangulation thereof.
 */
class WEXP ddgFacetSet : public ddgDLNode {
    typedef ddgDLNode super;
    friend class ddgClippedFacetSet;
	/// Statically initialized.
	static ddgVector3 _icovertices[12];
	/// Statically initialized.
	static int     _icopolygons[20][3];
	/// Number of polygons,
	unsigned int   _pcount;
	/// Split a facet into three subfacets.  Return true on error.
	bool subdivide( ddgVector3 v1, ddgVector3 v2, ddgVector3 v3, long depth);
	/// Number of subdivision to perform.
	unsigned int	_levels;
	/// A set of vertices/normals for corners of the faces.
	ddgVector3       *_vertices;
public:
	/// Class constructor.
	ddgFacetSet(unsigned int l = 0) : _vertices(0),_levels(l),_pcount(0) {	}
	/// Class destructor.
	~ddgFacetSet();
	/**
	* Initialize the data and allocate memory.
	* returns true on error.
	*/
	bool init(ddgContext *ctx);
	/// Draw the object.
	bool draw(ddgContext *ctx);
	/// Return the number of facets.
	unsigned int count(void) { return _pcount; }
};

/**
 * Class to represent a facetset whose facets are on or off depending
 * on whether they are visible from the given camera.
 */
class WEXP ddgClippedFacetSet : public ddgFacetSet {
    typedef ddgFacetSet super;
	/// Indicates which sectors are on.
	bool		*_on;
	/// Centre point angles.
	ddgVector3		*_centre;
	/// The angle of a single facet.
	float		_faceAngle;
	/// The maximum angle for which clipVisible returns true. default 90.
	float		_clipAngle;
	/// The angle within which a facet will be visible.
	float		_visAngle;

public:
	/// Class constructor.
	ddgClippedFacetSet(unsigned int l = 0) : ddgFacetSet(l), _on(0),_centre(0) { }
	/// Class destructor.
	~ddgClippedFacetSet();
	/**
	* Initialize the data and allocate memory.
	* returns true on error.
	*/
	bool init(ddgContext *ctx);
	/// Draw the object.
	bool draw(ddgContext *ctx);
	/// Return true if a sector is active.
	bool active(int i)  { return _on[i]; }
	/// Given a vector v, return the class of the nearest icosahedron normal. Preprocessing step.
	unsigned char classify( ddgVector3 v);
	/**
	* Calculate the visible set of iso polygons and turn those on
	* w.r.t given vector v.  Usually vector v is forward facing
	* vector of the camera.
	* if vector is null vector and reset is true, all facets are
	* set to active.
	* If reset if false, only facets which are currently
	* active are retested.
	* Returns false if no changes were needed to current state.
	*/
	bool clipVisibleSet(ddgContext *ctx );
	/// Set status of specific facet.
	void setFacet( unsigned int i, bool a )
		{ ddgAssert(i<count()); _on[i] = a; }
};
#endif
