/*
    Copyright (C) 1997, 1998, 1999, 2000 by Alex Pfaffe
	(Digital Dawn Graphics Inc)
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#ifndef _ddgLightModel_Class_
#define _ddgLightModel_Class_

#include "struct/ddgcolor.h"
#include "struct/ddggroup.h"

/**
 * A lightmodel object in a light model.
 * Add this object to the scene twice, once to turn
 * lighting on and once to turn it off again.
 */ 
class WEXP ddgLightModel: public ddgGroup {
    typedef ddgGroup super;
	/// If true lighting is active.
	bool _state;
    /// Ambient light level
    ddgColor4  _lightmodel_ambient;
	/// Enable two sided lighting.
	bool _twosided;
    /// Local or Parallel light.
    bool _local;
    /// Local or Parallel light.
    bool _colorMaterial;
public:
	/**
	 * Create a light model and set the ambient light.
	 * By default the light model is on, and lighting is
	 * not twosided and not local, and not coloredMaterial (faster).
	 */
	ddgLightModel( ddgColor4 *a = NULL, bool t = false, bool l = false, bool c = false )
	{ _state = true; _twosided = t; _local = l; _colorMaterial = c;
	_lightmodel_ambient.set( a ? a : (ddgColor4*)ddgColor4(100,100,100,255));
	}
	/// Draw all elements of this group lighted.
	bool draw( ddgContext *ctx );
    /// Set the ambient light level.
    void ambient( ddgColor4 *col) { _lightmodel_ambient.set(col); }
    /// Set the ambient light level.
    void ambient( ddgColor3 *col, unsigned char a = 255)
	{ _lightmodel_ambient.set(col->v[0],col->v[1],col->v[2],a); }
    /// Set the ambient light level.
    void ambient( unsigned char r, unsigned char g, unsigned char b, unsigned char a = 255)
	{ _lightmodel_ambient.set(r,g,b,a); }
    /// Set lighting calculation to perform angle math.
    void local( bool l) { _local = l; }
    /// Set lighting calculation to perform angle math.
    void twosided( bool t) { _twosided = t; }
    /// Mix current colour with material/texture color.
    void colormaterial( bool t) { _colorMaterial = t; }
    /// Set lighting on or off.
    void state( bool s) { _state = s; }
};

/**
 * A light object in a light model.  OpenGL supports 8.
 * Add this object to the scene twice, once to turn
 * the light on and once to turn it off again.
 */ 
class WEXP ddgLight: public ddgNode {
    typedef ddgNode super;
	/// Global identified for this light.
	static unsigned int lightid;
	/// Identifier for this light.
	unsigned int _id;
	/// Spotlight.angle. 0 is off >0 - 90 is half of cone width.
	float       _spotlight;
    /// Position and distance of the light.
	/// If _pos[3] is set to 0.0 it is infinite distance light with parallel rays.
    ddgVector4     _pos;
    /// Orientation of the light, for spotlights.
    ddgVector3     _direction;
	/// Ambient color of the light.
	ddgColor4      _colour_ambient;
	/// Diffuse color of the light.
	ddgColor4      _colour_diffuse;
	/// Specular color of the light.
	ddgColor4      _colour_specular;
	/// Is the light fixed in space. If not it is relative to the camera.
	bool        _fixed:1;
	/// Flags to indicate which parameters to set;
	bool		_fspotlight:1;
	bool		_fdirection:1;
	bool		_fpos:1;
	bool		_fambient:1;
	bool		_fdiffuse:1;
	bool		_fspecular:1;
public:
	/// Constructor.
	ddgLight(void);
	/// Recalculate the light position.
//	int update(void);
    /// Activate a single light.
//	void activate(void);
    /// Deactivate a single light.
//	void deactivate(void);
    /// Activate the light for rendering.
	bool  draw( ddgContext *ctx);

    // Set/Get methods.

    /// Set the ambient light colour.
    void ambient( ddgColor4 *c )
	{ _colour_ambient.set(c); _fambient = true;}
    void ambient( unsigned char r, unsigned char g, unsigned char b )
	{ _colour_ambient.set(r,g,b,255); _fambient = true; }
    /// Get the ambient light colour.
    ddgColor4 *ambient(void) { return &_colour_ambient; }
    /// Set the diffuse light colour.
    void diffuse( ddgColor4 *c )
	{ _colour_diffuse.set(c); _fdiffuse = true; }
    void diffuse( float r, float g, float b )
	{ _colour_diffuse.set(r,g,b,255); _fdiffuse = true; }
    /// Get the diffuse light colour.
    ddgColor4 *diffuse(void)
	{ return &_colour_diffuse; }
    /// Set the specular light colour.
    void specular( ddgColor4 *c )
	{ _colour_specular.set(c); _fspecular = true; }
    void specular( unsigned char r, unsigned char g, unsigned char b )
	{ _colour_specular.set(r,g,b,255); _fspecular = true; }
    /// Get the specular light colour.
    ddgColor4 *specular(void)
	{ return &_colour_specular; }
    /// Set the light position, without affecting the distance.
    void pos( ddgVector3 *p )
	{ _pos.set(p->v[0],p->v[1],p->v[2],_pos[3]); _fpos = true; }
    void pos( float i, float j, float k ) { _pos.set(i,j,k,_pos[3]); _fpos = true; }
    /// Get the light position.
    ddgVector4 *pos( void ) { return &_pos; }
    /// Set the orientation of the light as a vector.
    void direction( ddgVector3 *d )
	{ _direction.set(d); _fdirection = true; }
    void direction( float i, float j, float k  )
	{ _direction.set(i,j,k); _fdirection = true; }
    /// Set the orientation of the light as a heading and azimuth.
    void direction( float theta, float phi );
    /// Get the light direction.
    ddgVector3 *direction(void )
	{ return &_direction; }
    /// Set the distance of the light.
    void distance( float d )
	{ _pos.v[3] = d; _fpos = true; }
    /// Get the distance of the light from the position.
    float distance(void)
	{ return _pos[3]; }
    /// Set the light to be fixed or not.
    void fixed(bool f)
	{ _fixed = f; }
    /// Get the light's  to be fixed or not.
    bool fixed(void)
	{ return _fixed; }
    /// Set the size of the spotlight, if zero it is an infinitely strong light.
    void spotlight(float size)
	{ _spotlight = size; if (size) _fspotlight = true;}
    /// Get the size of the spotlight.
    float spotlight(void)
	{ return _spotlight; }

};
#endif
