/*
    Copyright (C) 1997, 1998, 1999, 2000 by Alex Pfaffe
	(Digital Dawn Graphics Inc)
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#ifndef _ddgCamera_Class_
#define _ddgCamera_Class_

#include "struct/ddgcolor.h"
#include "struct/ddgnode.h"
#include "render/ddgcamera.h"
#include "render/ddglight.h"


/**
 * Draw a camera object in the scene.
 */
class WEXP ddgCameraObj: public ddgNode {
    typedef ddgNode super;
	/**
	 * The real camera which this object represents.
	 * Orientation and position are linked to the camera
	 * object.
	 */
	ddgCamera* _camera;
	/// The color with which to draw the camera.
	ddgColor3*  _color;
	/// The size of the camera.
	float	_scale;
public:
	/// Constructor.
	ddgCameraObj(ddgCamera *c, ddgColor3 *cl, float s = 1.0 ) 
		: _camera(c), _color(cl), _scale(s) {}
	/// Render the camera object.
	bool draw( ddgContext *ctx);
	/// Set the size of the frustrum.
	void scale ( float s) { _scale = s; }
};

/**
 * Draw a light object in the scene.
 */
class WEXP ddgLightObj: public ddgNode {
    typedef ddgNode super;
	/**
	 * The real light which this object represents.
	 * Orientation and position are linked to the light
	 * object.
	 */
	ddgLight*   _light;
	/// The color with which to draw the camera.
	ddgColor3*  _color;
	/// The size of the camera.
	float	_scale;
public:
	/// Constructor.
	ddgLightObj(ddgLight *l, ddgColor3 *cl = 0, float s = 1.0 ) 
        : _color(cl), _scale(s), _light(l) { }
	/// Render the camera object.
	bool draw( ddgContext *ctx);
	/// Set the size of the frustrum.
	void scale ( float s) { _scale = s; }
};

/**
 * A yellow object located at the origin.
 * This class is used as a base class by other classes.
 */
class WEXP ddgGeneralObj: public ddgNode {
    typedef ddgNode super;
public:
	/// Color of object.
	ddgColor3 color;
	/// Location of object.
	ddgVector3 orig;
	/// Orientation of object.
	enum direction { X = 0, Y, Z };
	/// Constructor of object.
	ddgGeneralObj( float x = 0, float y = 0, float z = 0)
    { orig.set(x,y,z); color.set(255,255,0); }
};

/**
 * Render an X,Y,Z axis somewhere in the scene.
 */
class WEXP ddgAxis: public ddgGeneralObj {
    typedef ddgGeneralObj super;
public:
	/// Size of axis object.
	float d;
	/// Create a default axis object.
	ddgAxis(void) { d = 10; }
	/// Render the axis in the scene.
	bool draw( ddgContext *ctx );
};

/// A Two dimension grid oriented in the X,Y or Z direction.
class WEXP ddgGrid: public ddgGeneralObj {
    typedef ddgGeneralObj super;
public:
	/// Grid will be orthogonal to X,Y, or Z axis.
	direction dir;
	/// Grid width.
	float w;
	/// Grid height.
	float h;
	/// Number of gridlines along w axis.
	float nw;
	/// Number of gridlines along h axis.
	float nh;
	/// Create a default grid.
	ddgGrid(void) { nw = nh = 10; w = h = 100; dir = Y; }
	/** Create a grid of dimension ww and hh, with grid lines
	 * every dw,dh. Orient orthogonal to X,Y or Z axis.
	 */
	ddgGrid(float ww, float hh, float dw, float dh, direction d = Y)
	{ w = ww; h = hh; nw = dw; nh = dh; dir = d; }
	/// Render the lattice.
	bool draw( ddgContext *ctx);
};

/**
 * A Three dimensional gridded cube wherein only the inside facing polygons
 * are drawn.  Ie. you can always look into the cube. 
 */
class WEXP ddgLattice: public ddgGrid {
    typedef ddgGrid super;
public:
  /// Size of the lattice.
  float d;
  /// Size of grid overlay.
  float nd;
  /// Render the lattice.
  bool draw( ddgContext *ctx);
  /// Constructor.
  ddgLattice( void ) { d = 100; nd = 10; }
};



/**
* A one dimensional line in 3-space
*/
class WEXP ddgLineObj: public ddgGeneralObj {
	typedef ddgGeneralObj super;
public:
	/// start of line
	ddgVector3 *origin;
	/// end of line
	ddgVector3 *direction;
	/// Constructor.
	ddgLineObj(void)
	{
		origin = new ddgVector3(0, 0, 0);
		direction = new ddgVector3(1, 1, 1);
	}
	/// Constructor.
	ddgLineObj(ddgVector3 *o, ddgVector3 *d) : origin(o), direction(d) {}
	/// Render the axis in the scene.
	~ddgLineObj(void) {delete origin; delete direction;}
	/// Render the object.
	bool draw( ddgContext *ctx );
};

#endif
