/*
    Copyright (C) 1997, 1998, 1999, 2000 by Alex Pfaffe
	(Digital Dawn Graphics Inc)
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#ifndef _ddgTexture_Class_
#define _ddgTexture_Class_

#include "math/ddgvec.h"
#include "struct/ddgimage.h"
#include "struct/ddgcolor.h"

/**
 * A convenience function to support textures.
 * ddgTextures require an ddgImage to hold the texture data.
 * By default textures are initialized so that texture coordinates
 * can be specified from 0.0 to 1.0.  If you prefer to specify
 * texture coordinates in pixels, call scale with (1,1).
 * If you want to specify some arbitrary coordinate system from
 * 0-n, 0-m scale with (image->cols/n,image->rows/m);
 */
class WEXP ddgTexture {
public:
	/// Mode with which to apply the texture to the frame buffer.
	enum Mode{ DECAL = 0, MODULATE, BLEND };
private:
	/// Data part of texture.
	ddgImage		*_image;
	/// The height represented by this texture.
	float			_scalewidth;
	/// The width represented by this texture.
	float			_scaleheight;
	/// Filtered.
    bool			_linear:1;
	/// Repeated texture.
	bool            _repeat:1;
	/// Texture application mode.
    Mode			_mode;
    /// Color for blend or modulate mode.
    ddgVector4      _color;
    /// Unique texture id.
    unsigned int	_textureID;
public:
	ddgTexture( ddgImage *i = 0, bool l = false, bool r = true, Mode m = DECAL)
		: _mode(m), _linear(l), _repeat(r), _image(i), _color(1,1,1,1),
		_scalewidth(i->cols()), _scaleheight(i->rows()) { }
	/// Initialize the texture.  Must be called before rendering.
	bool init(void);
	/// Set linear filtering.
	void linear( bool l ) { _linear = l; }
	/// Set linear filtering.
	void repeat( bool r ) { _repeat = r; }
	/// Activate this texture.  Called before sending geometry.
	bool activate(void);
	/// Deactivate this texture.  Called after sending geometry.
	bool deactivate(void);
	/**
	 *  Set the size this texture represents.
	 *  independent of the actual image size.
	 *  By default a texture is scaled to have a texture
	 *  coordinate range from 0 to 1.
	 */
	void scale( float w, float h)
	{ _scalewidth = w; _scaleheight = h; }
	/**
	 * Set the texture mode and the color to use for BLEND or MODULATE.
	 * Note for MODULATE this is only a default color.  You can emit
	 * any colour during rendering.
	 */
    void mode( Mode m, ddgColor4 *c = NULL)
    {
	_mode = m;
    if (c)
        {
        _color.v[0] = c->v[0]/255.0;
        _color.v[1] = c->v[1]/255.0;
        _color.v[2] = c->v[2]/255.0;
        _color.v[3] = c->v[3]/255.0;
        } 
    }
    /// Return the image assigned to this texture.
    ddgImage *image(void) { return _image; }
	/// Return the base color of the texture.
	ddgVector4 *color(void) { return &_color; }
};

#endif
