/*
    Copyright (C) 1997, 1998, 1999, 2000 by Alex Pfaffe
	(Digital Dawn Graphics Inc)
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#ifndef _ddgViewPort_Class_
#define _ddgViewPort_Class_

#include "util/ddgutil.h"
#include "struct/ddgevent.h"
#include "render/ddgcamera.h"
#include "render/ddguiobj.h"



/**
 * This class supports rendering a view in a subregion of the main display.
 * One viewport should be defined as the root display.
 */
class WEXP ddgViewport : public ddgListNode
{
	friend class ddgViewportSet;
   /// Unique id used by internal callbacks.
    int		_id;
	/// If viewport is enabled.
    bool	_enabled:1;
	/// If true draw an edge around it.
    bool	_edge:1;
	/// If true automatically update if root is updated.
	bool    _slave:1;
	/// If true the viewport accepts keyboard input.
	bool	_control:1;
	/// TODO Camera should just register its controls to the viewport.
	/**
	 * Camera associated with the viewport.
	 * This can be NULL.
	 */
    ddgCamera *_camera;
	/// User defined update function for idle calls.
	void (*_update)(void);
    /**
	 * A set of UI Objects which are rendered at the after the 
     * scene.  Shown like a Heads Up Display.
     */
    ddgUIGroup		_uigroup;
	/// The default event handler.
	ddgEventHandler	*_eventHandler;

public:
    /** Viewport positions and sizes.
     * If this is the root viewport these are pixel coordinates.
     * If this is a sub viewport these are percentages coordinates.
	 */
    unsigned short int _px, _py, _dx, _dy;
    /** Create a sub-viewport associated with a camera
	 * at a given canvas position and a given size.
	 */
    ddgViewport( ddgCamera *c, unsigned short x = 5, unsigned short y = 70, unsigned short w = 25, unsigned short h = 25);

	/// Destroy the viewport.
	~ddgViewport()
	{
	}
    /// Return (and set)  the X position of the viewport.
    unsigned short int px( int i = -1) { return (i>-1 ? _px = i: _px); }
    /// Return (and set)  the Y position of the viewport.
    unsigned short int py( int i = -1) { return (i>-1 ? _py = i: _py); }
    /// Return (and set) the width of the viewport.
    unsigned short int dx( int i = 0) { return (i ? _dx = i : _dx ); }
    /// Return (and set) the width of the viewport.
    unsigned short int dy( int i = 0) { return (i ? _dy = i : _dy ); }
    /// Return the viewports id.
    int id( int i = 0) { return (i ? _id = i : _id); }

    /// Parameter access functions.
	void enabled( bool e ) { _enabled = e; }
	/// Return if viewport is active.
	bool enabled(void) { return _enabled; }
	/// Indicate if the viewport should have an edge drawn around it.
    void edge(bool edge) { _edge = edge; }
    ///
    bool edge(void) { return _edge; }
	/// Set and get the slave settings.
	void slave( bool s ) { _slave = s; }
	///
	bool slave( void )   { return _slave; }
	/// Set and get the slave settings.
	void control( bool c ) { _control = c; }
	/// Is this viewport a controller.
	bool control( void )   { return _control; }

	/// Set the camera.
	void camera( ddgCamera *c ) { _camera = c; }
    /// Get the camera for this viewport.
    ddgCamera* camera(void) { return _camera; }
	/// Get the uigroup.
	ddgUIGroup *uigroup(void) { return &_uigroup; }
	/// Set the update function.
	void update( void (*u)(void)) { _update = u; }
	/// Check if an update function is set.
	bool update( void ) { return _update ? true : false; }
	/// Call the update function to allow camera to perform some work.
	void doUpdate (void ) { _update(); }
	/// Return the event handler object.
	ddgEventHandler *eventHandler(void) { return _eventHandler; }
	/// Set the event handler object.
	void eventHandler(ddgEventHandler *eh) { _eventHandler = eh; }
};

/**
 * The set of viewports connected to the window.
 * The first viewport added must be the root window.
 */
class WEXP ddgViewportSet : public ddgList
{
    typedef ddgList super;
	/// Name of the root window.
	ddgStr    _name;
	/// Internal variable to indicate the current viewport.
	ddgViewport *_current;
	/// Global viewport set.
	static ddgViewportSet *_viewportset;
	/// Determine which sub viewport exists at this location.
	ddgViewport *hitTest(int x, int y);
	/// Function to call when viewportset exits.
	bool (*_shutdown)(void);
    /// Fullscreen. (WARNING: Use at own risk.)
    bool    _fullscreen;
	/// Fullscreen mode. eg 1 = "640x480:16@60"
	unsigned int _fsmode;
public:
	/**
	 * Create a viewport set with a title, choose a screen mode,
	 * 0 windowed
     * 1 640x480:16@60
	 * 2 800x600:16@60
	 * 3 1024x768:16@60
	 * 4 1152x870:16@60
	 * 5 1280x1024:16@60
	 * 6 1600x1200:16@60
     * 7 640x480:32@60
	 * 8 800x600:32@60
	 * 9 1024x768:32@60
	 * 10 1152x870:32@60
	 * 11 1280x1024:32@60
	 * 12 1600x1200:32@60
	 */
	ddgViewportSet( char *n = 0, unsigned int fs = 0) 
		: _current(0), _shutdown(0), _fsmode(fs)
	{ _name.assign(n ? n : "DDG"); _fullscreen = fs ? true:false;
	}

	/// Set the name. (Should affect window title.
	void name( char *n );
	/// Set or return the name.
	char *name( void ) { return _name; }
	// Initialize all viewports;
	/// Returns 0 on success, viewport id of problem viewport otherwise.
	int init( int argc, char **argv );
	/// Start rendering called once from main to enter the event loop.
	void enable(void);
	/// Callback functions
	static void draw(void);
	///
	static void reshape( int w, int h );
	///
	static void specialKey( int key, int x, int y );
	///
	static void key( unsigned char key, int x, int y );
	///
	static void mouse( int button, int state, int x, int y );
	///
	static void mouseMotion( int x, int y );
	///
	static void idle( void );
	///
	static void visible( int vis );
	/// Return the viewport with the given id or else return 0.
	ddgViewport *find( int id )
	{
		ddgViewport *v = (ddgViewport*) head();
		while (v && v->id() != id) v = (ddgViewport *) v->next();
		return v;
	}
	// This function will be called at shutdown time to
	// allow the user to clean up data.
	// Returns true if shutdown should be aborted.
	/// Set the update function.
	void shutdown( bool (*s)(void)) { _shutdown = s; }
	/// Test if the user defined shutdown function is set.
	bool shutdown( void ) { return _shutdown ? true : false; }
	/// Shutdown, dall the user defined shutdown function.  Pass in an optional error.
	void doShutdown ( ddgError *e );
	/// Check if we are initialization or rendering caused an error.
	static bool checkForError(void);
	/// Return a mode descriptor.
	static char *mode (unsigned int m, char *buf, unsigned int bufsize );
	/// Default key handler.
	static bool smouse(void *obj, int button, int state, int x, int y);

	/// Default Event handler, returns true if a change occurred.
	static bool smouseMotion(void *obj, int x, int y);

	/// Default Event handler, returns true if a change occurred.
	static bool sspecialKey(void *obj, int k, int x, int y);

	/// Default Event handler, returns true if a change occurred.
	static bool skey(void *obj, unsigned char k, int x, int y);
};

class WEXP ddgViewportEventHandler : public ddgEventHandler
{
public:
	/// Default constructor.
	ddgViewportEventHandler(void);
};

#endif
