/*
    Copyright (C) 1997, 1998, 1999, 2000 by Alex Pfaffe
	(Digital Dawn Graphics Inc)
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "util/ddgerror.h"
#include "struct/ddggroup.h"
// ----------------------------------------------------------------------

const unsigned long ddgGroup::_block = LIST_BLOCK_SIZE;

ddgGroup::ddgGroup( ) :
  _size(0), _buffer(0), _nodes(0){ }



ddgGroup::~ddgGroup( void )
{
	unsigned long i;
	for ( i = 0; i < _size ; i++ )
	{
		delete (_nodes[i]);
		_nodes[i] = 0;
	}
	if (_nodes) free(_nodes);
	_nodes = 0;
}

// =======================================================================
//
// Adds an entity to the end of the list.
// If the list is ordered, the item is inserted in order of its name.
// If the name already exists it is added at the end of the names.
// If the list is unique, adding a duplicate name returns an error.
// If the entity being added is null the method returns an error.
// Returns FALSE (0) on error, Entity pointer on success.
//
// =======================================================================

ddgNode *ddgGroup::add( ddgNode* entity )
{
	ddgAssert(entity);
	// Grow the list if needed.
	if (_buffer <= _size ) {
		_buffer += _block;	  // Add blocks of entries 
		if (!_nodes) _nodes = (ddgNode **) malloc ((size_t) _buffer * (sizeof(ddgNode *)));
		else _nodes =            (ddgNode **) realloc( (void*)_nodes, (size_t) _buffer * (sizeof(ddgNode *)));

		if (!_nodes) return 0;   // Failed to alloc memory
	} // End if

	long index = _size;

	_nodes[index] = entity;
	_size++;

	return entity;
}


// =======================================================================
//
// Remove an entity from the list if the entity is found in the list.
// If depth is not zero, the entity is also destroyed.
// Returns FALSE on error, TRUE on success.
// Reasons for the error may be:
// 1: Item was 0
// 2: Item not found, or
// 3: Error collapsing list.
//
// =======================================================================

bool ddgGroup::remove( ddgNode* entity, bool kill)
{
	unsigned long i = 0;
	if (!entity) return false;
	while (i < _size ) {
		if ( entity == _nodes[i] ) { // Element found in list
			if (kill) { // Destroy entity
				delete entity;
			}
			while( i++ < _size-1 ) _nodes[i-1] = _nodes[i];
			_size--;
			if (_buffer - _block > _size ) {
				// Collapse list
				_buffer -= _block;
				if (!_nodes) _nodes = (ddgNode **) malloc((size_t) _buffer *(sizeof(ddgNode *)));
				else _nodes = (ddgNode **) realloc( (void*)_nodes,(size_t)_buffer *(sizeof(ddgNode *)));
				if (!_nodes) return false;
				if (!_buffer) _nodes = 0;	// Empty list.
			} // End if
			return true;
		} // End if
		else i++;
	} // End while
	return false;
}



// =======================================================================
//
// Find if node is a member of the list and if so, return the index #
// If found, this methods returns a number from 0-(_size-1)
// If not found, the method returns LIST_NOT_MEMBER. 
//
// =======================================================================

long ddgGroup::which( ddgNode *e ) const
{
	long n = _size;
	while( n && _nodes[--n]!=e); 
	if (_nodes[n]==e) return n;
	return LIST_NOT_MEMBER;
}

// =======================================================================
//
// Render all nodes in the group. 
//
// =======================================================================

bool ddgGroup::draw( ddgContext *c )
{
	unsigned long n = 0;
	bool result = false;
	while( n < _size && !result) 
	{
		if (_nodes[n]->mode.flags.disabled == false)
			result |= ((_nodes[n])->draw(c));
		n++;
	}
	return result;
}

// =======================================================================
//
// Initialize all nodes in the group. 
// Each camera will call init.
//
// =======================================================================

bool ddgGroup::init( ddgContext *c )
{
	unsigned long n = 0;
    if (super::init(c))
        return ddgFailure;
    bool error = ddgSuccess;
	while( n < _size && ddgSuccess == error) 
	{
		if ((_nodes[n])->init(c) && (_nodes[n]->mode.flags.init == false))
            error = ddgFailure;
		n++;
	}
	return error;
}

// =======================================================================
// OBJECT:  DIterator
// =======================================================================
// METHODS:
//
// =======================================================================
ddgIterator tmp;

ddgIterator& ddgIterator::operator++( void )    { if (_list && (_index < _list->size())) _index++; return *this;}
ddgIterator& ddgIterator::operator++( int )     { tmp = *this; if (_list && (_index < _list->size())) _index++; return tmp;}
ddgIterator& ddgIterator::operator--( void )    { if (_list && (_index >= 0))            _index--; return *this;}
ddgIterator& ddgIterator::operator--( int )     { tmp = *this; if (_list && (_index >= 0))            _index--; return tmp;}
