/*
    Copyright (C) 1997, 1998, 1999, 2000 by Alex Pfaffe
	(Digital Dawn Graphics Inc)
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#ifndef _ddgGroup_Class_
#define _ddgGroup_Class_

#include "struct/ddgnode.h"

#define LIST_BLOCK_SIZE	10
#define LIST_NOT_MEMBER	-1
/**
 * This object is used to store an ordered collection of entities.  The
 * Entities are added the end of the list or sorted by their name. 
 * DGroups automatically allocate additional blocks of space as needed and
 * free blocks of space when no longer needed.  DGroups can also be searched
 * for entities.  DGroups may also be unique and only allow one entity of
 * each name.
 * An empty DGroup will not reserve any space.
 */
class WEXP ddgGroup : public ddgNode
{
    typedef ddgNode super;
private:
    /// Block size for growing list
    static const unsigned long	_block;
    /// Buffer size
    unsigned long      			_buffer;
    /// Current list size
    unsigned long		     		_size;
    // A list of entities
    ddgNode              	      **_nodes;
public:
    /// Constructor
    ddgGroup( );
    /// Delete this group list and any of its children.		 
    virtual	~ddgGroup( void );
    /// Add an entity to the list.
    ddgNode*	 add( ddgNode* n );
    /**
    * Remove an node from the list.
    * The kill flag indicates the the node should also be delete.
    */
    bool  	 remove( ddgNode* n, bool k = false);
    /**
    * Find the nth node in the group.  Return 0 if out of range.
    */
    ddgNode*	 index( unsigned long n ) const  { return ( n>=0 && n <_size) ? _nodes[n] : 0; }
    /**
    * Determine which position the entity has in the list.
    * Return NULL if the entity is not in the list.
    */
    long	 	 which( ddgNode *n ) const;
    /**
    * Return the current number of entities in the list.
    */
    long	 	 size( void ) const     { return _size; }
    /**
    * Initialize all nodes in the group.
    * If any member returns true, further initialization is aborted.
    */
    bool init (ddgContext *ctx);
    /**
    * Render all the members of the group.
    * If any member returns true, further rendering is aborted.
    */
    bool draw( ddgContext *c );
};

/**
 * This object is used to traverse a DGroup object.  It remebers its current
 * location in the object it is traversing.  Multiple Iterators can operate
 * on the same DGroup.
 */
class ddgIterator
{
    /// DGroup over which to iterate.
    ddgGroup *	_list;
    /// Current entry.
    long 	_index;
public:
	/// Default Constructor.
	ddgIterator( void )  : _list(0) , _index(0) {}
    /// Constuctor initialized with a ddgGroup.
	ddgIterator( ddgGroup *l )   { _list = l; _index = 0; }
    ///
    void 	 initFirst( ddgGroup * l ) { _list = l; _index = 0; }
    ///
    void 	 initLast ( ddgGroup * l ) { _list = l; _index = _list->size()-1; }
    /// Return the element in the list to which the iterator is currently pointing.
    ddgNode*	 operator*( void ) const { return ( _list && (_index < _list->size()))  ? _list->index(_index) : 0; }
    /**
    * Inlining of the following causes problems in some cases on the SGI.
    * The following while expression will fail:
    * Iterator it( list ), it2 ( list2 );
    * while ( *it++ && *it2++) { do something }
    */
    ddgIterator&	 operator++( void ); ///   { if (_list && (_index < _list->size())) _index++; return *this;}
    ddgIterator&	 operator++( int );  ///   { DIterator tmp = *this; if (_list && (_index < _list->size())) _index++; return tmp;}
    ddgIterator&	 operator--( void ); ///   { if (_list && (_index >= 0))            _index--; return *this;}
    ddgIterator&	 operator--( int );  ///   { DIterator tmp = *this; if (_list && (_index >= 0))            _index--; return tmp;}
};

#endif
