/*
    Copyright (C) 1997, 1998, 1999, 2000 by Alex Pfaffe
	(Digital Dawn Graphics Inc)
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
// For timing info.
#ifdef WIN32
#include <sys/timeb.h>
#include <windows.h>
#include <winbase.h>
static LARGE_INTEGER qpcfrequency;
bool	useqpc = false;
#else
#include <sys/types.h>
#include <sys/times.h>
#endif
#include <time.h>
#include <fstream.h>

#include "util/ddgerror.h"
#include "util/ddgclock.h"
#include "util/ddgutil.h"

// ----------------------------------------------------------------------
ddgClock::ddgClock( float init)
{
    _time = init;
    _elapsed = 0;
    _sysfactor = 1.0 / 1000.0;
	_simfactor = 600.0; // Each system second is 10 minutes in simulation time.
 	_simstart = 0.0;    // The start time int seconds.
	_frame = 0;         // 1st frame yet to be shown.
	_prev = 0;
	_totalFrame = 0;
	_totalTime = 0;
	_minTime = 0xFFFF;
	_maxTime = 0x0000;
#ifdef WIN32
	if (QueryPerformanceFrequency ( &qpcfrequency))
		useqpc = true;
	else
		useqpc = false;	// Failed.
#endif

	_fpsidx = 0;
	_fpstotal = 0;
	
	for (_fpsidx = ddgFPSBUFSIZE; _fpsidx > 0; _fpsidx--)
		_fpsbuf[_fpsidx-1] = 0;
}

void ddgClock::timeStamp( ddgStr *timestamp )
{
	// Create a time stamp of the form: YYYYMMDDHHMMSS
#define TIMESTMP 14
	struct tm *stm;
	long ltime;
	time(&ltime);
	stm = localtime(&ltime);
	char * tmstmpbuf = new char[TIMESTMP+1];
#ifdef DDGSTREAM
	ostrstream tstmp(tmstmpbuf,TIMESTMP);
	tstmp.fill('0');
	tstmp.width(4);	tstmp << ((stm->tm_year < 200) ? stm->tm_year +1900 : stm->tm_year);
	tstmp.width(2);	tstmp << stm->tm_mon+1;
	tstmp.width(2);	tstmp << stm->tm_mday;
	tstmp.width(2);	tstmp << stm->tm_hour;
	tstmp.width(2);	tstmp << stm->tm_min;
	tstmp.width(2);	tstmp << stm->tm_sec;
#endif
	tmstmpbuf[TIMESTMP] = '\0';
	timestamp->assign(tmstmpbuf);
}

bool ddgClock::perfDump( ddgStr *filename, char *s)
{
	ddgStr tstmp;
	ddgClock::timeStamp(&tstmp);

	if (filename->length()>0)
	{
		// Create a filename with that time stamp.
		int l = tstmp.length() + 4;
		int fl = filename->length();
		char * filebuf = new char[fl+l+1];
#ifdef DDGSTREAM
		ostrstream fn(filebuf,fl+l);
		fn << filename->s << tstmp.s << ".txt" << '\0';
		filebuf[fl+l] = 0;
		char *fullName = ddgFileHelper::getFullFileName(filebuf);
		ofstream fout(fullName);
		if (!fout.good())
		{
			ddgErrorSetStr(FileWrite,"Could not open perf log " << ((char *) (fullName ? fullName : "(null)")));
			ddgError::report();
			return true;
		}

		fout << "Time " << _sysfactor*_totalTime << " seconds representing frames " << _totalFrame << endl;
		fout << "Avg " << (float)((_totalTime>0)?((float)_totalFrame/(_sysfactor*_totalTime)):0) << endl;
		fout << "Max " << (1.0/((float)_minTime*_sysfactor)) << "(" << (_sysfactor*_minTime) << "s/f)" << endl;
		fout << "Min " << (1.0/((float)_maxTime*_sysfactor)) << "(" << (_sysfactor*_maxTime) << "s/f)" << endl;
		fout << "Date " << tstmp.s << endl;
		if (s) fout << s;
		fout << endl;
		fout.close();
#endif
	}
	else
	{
		cerr << "Time " << _sysfactor*_totalTime << " seconds representing frames " << _totalFrame << endl;
		cerr << "Avg " << (float)((_totalTime>0)?((float)_totalFrame/(_sysfactor*_totalTime)):0) << endl;
		cerr << "Date " << tstmp.s << endl;
		if (s) cerr << s;
		cerr << endl;
	}
	return ddgSuccess;
}
float ddgClock::getSimTime(float *h, float *m, float *s, float *r)
{
   float simSeconds = _time * _sysfactor * _simfactor + _simstart;
   float simDays = simSeconds / ( 24 * 60 * 60 );
   if (h)
     {
     *h = simDays;
     simDays = (int)simDays;
     *h = 24.0 * (*h - simDays);
     }
   if (m)
     {
     *m = *h;
     *h = (int) *h;
     *m = 60.0 * ( *m - *h);
     }
   if (s)
     {
     *s = *m;
     *m = (int) *m;
     *s = 60.0 * ( *s - *m);
     }
   if (r)
     {
     *r = *s;
     *s = (int) *s;
     *r = *r - *s;
     }
   return simDays;
}

// This function only take samples every avgFrame calls.
float ddgClock::update(void)
{
    unsigned long now = 0;
#ifdef WIN32
	LARGE_INTEGER p;
	if (useqpc && QueryPerformanceCounter(&p))
	{	
		now = (p.QuadPart*1000L)/qpcfrequency.QuadPart;
	}
	else
	{
		// Fall back to ftime.
		timeb tp;
		ftime (&tp);
		now = (unsigned long) tp.time*1000+tp.millitm;
	}
#else
	struct tms buf;
	now = times (&buf) * 1000 / CLK_TCK;
#endif
	if (_prev)
	{
		_elapsed = (now-_prev);
		_time += _elapsed;

		if (_totalFrame && _elapsed )
		{
			_totalTime += _elapsed;
		}
		if (_frame > 2 )
		{
			if (_elapsed < _minTime)
				_minTime = _elapsed;
			else if (_elapsed > _maxTime)
				_maxTime = _elapsed;
		}
	}
	_prev = now;
	_frame++;

	return _elapsed * _sysfactor;
}

float ddgClock::getFrameRate(void)
{
    return (float)(_elapsed>0.0)? 1.0/(_sysfactor * _elapsed):0.0;
}

float ddgClock::getAvgFrameRate(void)
{
	float fps = (float)(_elapsed>0.0)? 1.0/(_sysfactor * _elapsed):0.0;
	// Remove old value.
	_fpstotal -= _fpsbuf[_fpsidx];
	// Set new value.
	_fpsbuf[_fpsidx] = fps;
	// Add New value.
	_fpstotal += fps;
	// Calc avg value.
	fps = _fpstotal / ddgFPSBUFSIZE;
	// Advance to next slot.
	_fpsidx++;
	// Wrap around if past the end.
	if (_fpsidx == ddgFPSBUFSIZE)
		_fpsidx = 0;
	return fps;
}

float ddgClock::getDayLight(void)
{
    // 1.0 = 100% light (noon).
	float d = 0, h;
	d = getSimTime(&h);
    if      (h <= 5 || h >=19 )
		d = 0.0;		// Night from 7pm to 5am.
    else if (h >= 7 && h < 17)
		d = 1.0f;		// Day from 7am to 5 pm.
    else if (h < 7)
		d = (h-5.0f)/2.0f;	// Dawn from 5am to 7am.
    else
		d = 1.0 - (h-17.0f)/2.0f;		// Sunset from 5pm to 7pm.
    return d;
}

#ifdef ASMCLOCK
#ifdef _DEBUG
#define TIMING_LATENCY 160
#else
#define TIMING_LATENCY 91
#endif

class timing_c
{
private:
	__int64 start;
	__int64 end;
	__int64 count;
	__int64 total;

	int reset;
public:
timing_c()
{
	count = 0;
	total = 0;
	reset = 256;
}

timing_c(int res)

{
	count = 0;
	total = 0;
	res ? res : 1;
	reset = res;
}

void Start()
{
	const __int64 *s = &start;
	__asm
	{
		push eax
		push ebx
		push edx

		rdtsc
		mov ebx, s
		mov [ebx], eax
		mov [ebx + 4], edx

		pop edx
		pop ebx
		pop eax
	}
}

int End()
{
	const __int64 *e = &end;
	__int64 time;
	__asm
	{
		push eax
		push ebx
		push edx

		rdtsc
		mov ebx, e
		mov [ebx], eax
		mov [ebx + 4], edx

		pop edx
		pop ebx
		pop eax
	}
	time = end - start - TIMING_LATENCY;
	if (time < 0)
	{
		time = 0;
	}
	count++;
	total += time;

	return((int)time);
}

int GetAverage()
{
	int time = 0;

	if(count)
	{
		time = (int)(total / count);
		if(count > reset)
		{
			count = 0;
			total = 0;
		}
	}

	return(time);
}

int GetCount()
{
	return((int)count);
}

int GetTotal(bool adjusted)
{
	if (adjusted)
	{
		return((int)(total / 400000));
	}
	else
	{
		return total;
	}
}

};

#endif