/*
    Copyright (C) 1997, 1998, 1999, 2000 by Alex Pfaffe
	(Digital Dawn Graphics Inc)
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#ifndef _ddgClock_Class_
#define _ddgClock_Class_

#include "util/ddgstr.h"

#define ddgFPSBUFSIZE	16

/// General purpose real-time clock object.
class WEXP ddgClock
{
    /// In system units.
    float  _time;			
	/// In system units.
    float  _elapsed;
	/// Conversion unit from sys units to seconds
    float  _sysfactor;
	/// Conversion unit from sys seconds to sim seconds
    float  _simfactor;
	/// The start time for the simulation in sim seconds.
    float  _simstart;
	/// Working variable.
	unsigned long _prev;
    /// Working variable.
    unsigned int _frame;
	/// Define the number of frames to record in perf log.
	unsigned int _totalFrame;
	/// Total elapsed time.
	unsigned int _totalTime;
	/// Fastest frame.
	unsigned int _minTime;
	/// Slowest frame.
	unsigned int _maxTime;
	/// Buffer for averaging FPS measurements.
	float _fpsbuf[ddgFPSBUFSIZE];
	/// Circular buffer index.
	int _fpsidx;
	/// Current total for all frames in the buffer.
	float _fpstotal;
public:
	/// Create a clock and initialize its time.
    ddgClock( float init = 0.0);
	/// Returns elapsed in milli seconds time since last call.
    float update(void);
	/// Set the clock to some time in seconds.
    void set(int t) { _time = t / _sysfactor; }
	/// Get time in seconds.
    float getSysTime(void) { return _time * _sysfactor; }
	/// Get elapsed time in seconds.
    float getElapsed(void) { return _elapsed * _sysfactor; }
    /// Calculate a frame rate based on time since last call.
    float getFrameRate(void);
	/// Calculate an averaged frame rate. Across ddgFPSBUFSIZE frames.
	float getAvgFrameRate(void);

    // Return the simulated time where one system second is _simfactor seconds.
    // The return values is the fractional number of days elapsed.
    // If h is given, h returns the number of hours and returned days are rounded.
    // If m is given, m returns the number of minutes and h is rounded.
    // If s is given, s returns the number of seconds and m is rounded.
    /// If r is given, r returns the number of partial seconds and s is rounded.
    float getSimTime(float *h = 0, float *m = 0, float *s = 0, float *r = 0);
	/// Set the start time in sim seconds.
	void start( float t) { _simstart = t; }
	/// Set the conversion factor for sys seconds to sim seconds..
	void simfactor( float f ) { _simfactor = f; }
	/// Return the light intensity.  0 at night, 1 during day, inbetween at dusk and dawn.
	float getDayLight(void);

	/// Set perf on or off.
	void perf( unsigned int tf ) { _totalFrame = tf-1; }
	/// Write collected performance information to a logfile or standard out.  Return true on error.
	bool perfDump( ddgStr *filename = NULL, char *s = NULL );
	/// Generate a timestamp string.
	static void timeStamp( ddgStr *timestamp );
};
#endif
