/*
    Copyright (C) 1997, 1998, 1999, 2000 by Alex Pfaffe
	(Digital Dawn Graphics Inc)
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include <fstream.h>

#include "util/ddgerror.h"
#include "util/ddgclock.h"
#include "util/ddgutil.h"
// ----------------------------------------------------------------------

// Global error object to ensure gError class gets initialized.
ddgError *ddgError::gError = NULL;

/// Function to call when viewportset exits.
void (*ddgError::_errorHandler)(ddgError*) = NULL;

#ifdef DDGSTREAM
#define DDG_ERRORMSGBUFLENTH	256
/// Small buffer to set error messages to.
char _errorBuf[DDG_ERRORMSGBUFLENTH] = "";
/// Stream object access.
ostrstream	_errorStr(_errorBuf,DDG_ERRORMSGBUFLENTH);
ofstream    *fileStream = NULL;
#endif



ddgError::~ddgError(void)
{
	ddgLogFile::logClose();
}

ddgError::ddgError(void)
{
	gError = this;
	/// Global Error code info.
	_error = Unknown;
	// The error code as a string.
	_code.assign("-");
	/// Line number at which error occured.
	_line = 0;
	/// File in which error occured.
	_file.assign("-");
	/// Error message.
	_msg.assign("-");
}

#ifdef DDGSTREAM
/// Return a stream object.
ostrstream &ddgError::errorStream(void)
{
	return _errorStr;
}
#endif

bool ddgError::isError(void)
{
	if (!gError || gError->_error == ddgSuccess)
		return false;
	return true;
}

void ddgError::exitHandler( void )
{
	if (!gError)
		gError = new ddgError;
	if (_errorHandler)
	{
		_errorHandler(gError);
	}
	else
	{
		exit (report());
	}
}

void ddgError::errorHandler( void (*s)(ddgError*))
{
	_errorHandler = s;
}

int ddgError::report(void)
{
	if (gError->_error == ddgSuccess)
		return ddgSuccess;

	if (!gError) gError = new ddgError;
#ifdef DDGSTREAM
	if (gError)
		cerr << "ERROR:"
			 << gError->_file.s << "(" 
			 << gError->_line << ")"
			 << gError->_error << ":"
			 << gError->_code.s << endl 
			 << gError->_msg.s << endl;
#endif
	// If the log file is open, close it now.
	ddgLogFile::logClose();

	return gError->_error;
}

bool ddgError::set(ddgErrorCode error, const char *msg, const char *filename, int line, const char *code )
{
	if (!gError) gError = new ddgError;
	if (gError)
	{
		gError->_error = error;
		gError->_line = line;
		gError->_file.assign(filename);
		gError->_msg.assign(msg ? msg : _errorBuf);
		gError->_code.assign(code);
	}
	return ddgFailure;
}

#ifdef _DEBUG
#define ddgMEMALLOCBUFSIZE	6000
struct _memAllocBuf
{
	unsigned int line;
	ddgStr		file[256];
	ddgStr		type[20];
	long size;
	long count;
} memAllocBuf[ddgMEMALLOCBUFSIZE];
#endif
unsigned long memAllocBufsize = 0;
unsigned long memTotalInUse = 0;
unsigned long memMaxInUse = 0;
void ddgError::memory( const char * filename, unsigned int line, long size, long count, const char *type)
{
#ifdef _DEBUG
	if (memAllocBufsize < ddgMEMALLOCBUFSIZE && (count * size))
	{
		memAllocBuf[memAllocBufsize].file->assign(filename);
		memAllocBuf[memAllocBufsize].type->assign(type);
		memAllocBuf[memAllocBufsize].line = line;
		memAllocBuf[memAllocBufsize].count = count;
		memAllocBuf[memAllocBufsize].size = size;
		memAllocBufsize++;
	}
#endif
	memTotalInUse += count * size;
	if (memTotalInUse > memMaxInUse)
		memMaxInUse = memTotalInUse;
}

unsigned long ddgError::reportTotalInUse( void )
{
	return memTotalInUse;
}

unsigned long ddgError::reportMaxInUse( void )
{
	return memMaxInUse;
}

bool ddgError::reportMemory( ddgStr *filename, char *s)
{
#ifdef _DEBUG
	unsigned int i = 0;
	if (filename)
	{
#ifdef DDGSTREAM
		ddgStr tstmp;
		ddgClock::timeStamp(&tstmp);
		int l = tstmp.length() + 4;
		int fl = filename->length();
		char * filebuf = new char[fl+l+1];
		ostrstream fn(filebuf,fl+l);
		fn << filename->s << tstmp.s << ".txt";
		filebuf[fl+l] = 0;
		char *fullName = ddgFileHelper::getFullFileName(filebuf);
		ofstream fout(fullName);
		if (!fout.good())
		{
			ddgErrorSet(FileRead,(char *) (filename->s ? filename->s : "(null)"));
			ddgError::report();
			return ddgFailure;
		}
		while (i < memAllocBufsize)
		{
			fout
				 << ":" << memAllocBuf[i].type->s 
				 << "\t[" << memAllocBuf[i].size << "]x " << memAllocBuf[i].count
				 << "=" << (memAllocBuf[i].size * memAllocBuf[i].count)
				 << "\t:" << memAllocBuf[i].file->s <<":" <<memAllocBuf[i].line
				 << endl;
			i++;
		}
		fout << "Total memory in use: " << memTotalInUse << endl;
		fout << "Max memory in use: " << memMaxInUse << endl;
		if (s) fout << s;
		fout << endl;
		fout.close();
#endif
	}
	else
	{
		while (i < memAllocBufsize)
		{
			cerr 
				 << ":" << memAllocBuf[i].type->s 
				 << "[" << memAllocBuf[i].size << "]x " << memAllocBuf[i].count
				 << "=" << (memAllocBuf[i].size * memAllocBuf[i].count)
				 << "\t:" << memAllocBuf[i].file->s <<":" <<memAllocBuf[i].line
				 << endl;
			i++;
		}
		ddgConsole("Total memory in use: " << memTotalInUse << endl);
		if (s) cerr << s;

	}
#endif
	return ddgSuccess;
}

/* ==================================================================================== */


ddgLogFile *ddgLogFile::gLogFile = NULL;
char _logBuf[DDG_ERRORMSGBUFLENTH] = "";
ostrstream	_logStr(_logBuf,DDG_ERRORMSGBUFLENTH);



ddgLogFile::ddgLogFile(void)
{
	gLogFile = this;
	_enabled = false;
	_logMessage = _logBuf;
}

ddgLogFile::~ddgLogFile()
{
	ddgLogFile::logClose();
}

#ifdef DDGSTREAM
/// Return a stream object.
ostrstream &ddgLogFile::logStream(void)
{
	return _logStr;
}
#endif

bool ddgLogFile::log(const char *msg)
{
#ifdef DDGSTREAM
	// Open the file on the first request.
	if (!fileStream)
	{
		// Create a filename with that time stamp.
		ddgStr tstmp;
		ddgStr filename("log");
		ddgClock::timeStamp(tstmp);
		int l = tstmp.length() + 4;
		int fl = filename.length();
		char * filebuf = new char[fl+l+1];

		ostrstream fn(filebuf,fl+l);
		fn << filename.s << tstmp.s << ".txt" << '\0';
		filebuf[fl+l] = 0;
		char *fullName = ddgFileHelper::getFullFileName(filebuf);
		fileStream = new ofstream(fullName);
		if (!fileStream->good())
		{
			ddgErrorSetStr(FileWrite,"Could not open perf log " << ((char *) (fullName ? fullName : "(null)")));
			ddgError::report();
			return ddgFailure;
		}
	}
	// Send the message to the stream.
	*fileStream << msg;
#endif
	return ddgSuccess;
}

bool ddgLogFile::logClose(void)
{
#ifdef DDGSTREAM
	if (fileStream)
	{
		fileStream->close();
		delete fileStream;
		fileStream = NULL;

	}
	enabled(false);
#endif
	return ddgSuccess;
}

/// Set file logging state.
void ddgLogFile::enabled(bool active)
{
	if (!gLogFile) gLogFile = new ddgLogFile;

	gLogFile->_enabled = active;
}
/// Get file loggin state.
bool ddgLogFile::enabled(void)
{
	if (!gLogFile) gLogFile = new ddgLogFile;

	return gLogFile->_enabled;
}
/// Log message.
char *ddgLogFile::logMessage(void)
{
	if (!gLogFile) gLogFile = new ddgLogFile;

	return gLogFile->_logMessage;
}
