/*
    Copyright (C) 1997, 1998, 1999, 2000 by Alex Pfaffe
	(Digital Dawn Graphics Inc)
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include <iomanip.h>

#include "math/ddgnoise.h"
#include "render/ddgcloud.h"

// ----------------------------------------------------------------------
ddgCloud::ddgCloud( ddgImage* i, float size ) : super(ddgVArray::triangle)
{
	_image = i;
	_octaves = 0;
	_dome.sphere(size,20);
}


ddgCloud::ddgCloud( unsigned short textureSize, unsigned int o, float size ): super(ddgVArray::triangle)
{
	// Set the colours that make up the horizon and so on.
	_octaves = o;
	_dome.sphere(size,20);
	// Luminance = alpha.
	_image = new ddgImage(1);
	ddgMemorySet(ddgImage,1);
	_image->allocate(textureSize,textureSize);

	// Indexes in the texture map.
	int r,c;
	// Position in noise space.
	double v[3];
	v[0] = 0; v[1] = 0; v[2] = 0;

	// Sample value in noise space at position v.
	double t;
	// Working variables.
	double deg = 0.0,
	// Due to problem with noise function only 80% of range
	// is useful so we need to scale the input.
	s = 0.8,
	sp = 360.0/(float)(_image->cols()),
	sz = s*2.0*ddgPi/(float)(_image->rows());
	// Calculate a cylinder, this produces a scrollable
	// cloud layer.
	if (_octaves)
	{
		for (c=0; c < _image->cols(); c++)
		{
			if (c%20==0)
				ddgConsoleMessage::progress("Building clouds data",c,_image->cols());
			// Map rows 0..imgheight -> degrees
			v[0] = 0.05+ddgAngle::sin(deg)*s; v[1] = 0.05+ddgAngle::cos(deg)*s; v[2] = 0;
			deg = deg + sp;
			// Calc all values for this angle.
			for (r=0; r < _image->rows(); r++)
			{
				v[2] = v[2] + sz;
				t = (ddgNoise::fBm(v,0.5,2.0,_octaves)+1.0)/2.0;
				// t is from 0.0 to 1.0
				if (_image->channels() == 1)
					_image->set(r,c,(unsigned char)t*255);
				else
				{
					// Set cloud color.
					ddgColor4 cl;
                    ddgColor4 clear(0,0,0,0);
                    ddgColor4 solid(200,200,255,255);
					// Use exponential curve, which looks more realisitic.
					cl.einterp(clear,solid,t);
					_image->set(r,c,
						(unsigned char)255,
						(unsigned char)255,
						(unsigned char)255,
						(unsigned char)cl[3]);
				}
			}
		}
    }
	ddgConsoleMessage::progress("Building clouds data",1,1);
}

bool ddgCloud::init(ddgContext *ctx )
{
    if (super::init(ctx)) return ddgFailure;

	_bright = 0.6;
	_cover =1.0;
	_cday.set(255,255,255,255);
	_ctwilight.set(255,125,0,255);
	_cnight.set(50,50,50,255);
	drift(0,1);

    _texture = new ddgTexture(_image,true);
	ddgMemorySet(ddgTexture,1);
	_texture->scale(_image->rows()*2,_image->cols()*2);
    _texture->init();
	_drift.set(0,0);
	varray.renderMode(true,false,false);
	_dome.init(&varray, true,false);

	return ddgSuccess;
}



// Render a sky patch from -10 to 10 degrees in each direction.
// with 0,0 directly over head.
//
// TODO:
//	  Determine which grid cells are visible and need to be rendered.


bool ddgCloud::draw(ddgContext *ctx )
{
	if (ctx->quality() == ddgContext::DEBUG)
		return ddgFailure;

	float t = ctx->clock()->getSysTime() / 100.0f;
	float day = ctx->clock()->getDayLight();
	static ddgColor4 cnow;
	if (day < 0.5)
		cnow.linterp(_cnight,_ctwilight,day*2);
	else
		cnow.linterp(_ctwilight,_cday,(day-0.5)*2);

	// Adjust this to go to full cover.
	cnow.v[3] = _cover*255;

	_texture->mode(ddgTexture::MODULATE,cnow);
	_texture->activate();
	drift(t,0);

	if (_drift[0] || _drift[1])
	{
		glMatrixMode(GL_TEXTURE);
		glPushMatrix();
		glTranslatef(_drift[0], _drift[1],0);
		glMatrixMode(GL_MODELVIEW);
	}

	glEnable(GL_BLEND);
	glShadeModel(GL_SMOOTH);
	glDepthFunc(GL_LEQUAL);
	// I would like to use GL_DST_ALPHA to keep sun active and stars dimmer, but it doesnt work.
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	super::draw(ctx);
	glDepthFunc(GL_LESS);
	texture()->deactivate();
	if (_drift[0] || _drift[1])
	{
		glMatrixMode(GL_TEXTURE);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
	}

    return ddgSuccess;
}



