/*
    Copyright (C) 1997, 1998, 1999, 2000 by Alex Pfaffe
	(Digital Dawn Graphics Inc)
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
// ======================================================================
//
// Object:  Culling
// Name:    Culling
// Description:   Culling functionality to reduce the size of the scene
//                graph before rendering.
//
// ----------------------------------------------------------------------
#include <string.h>

#include "struct/ddgoctree.h"
#include "render/ddgculling.h"
#include "render/ddgscene.h"

// Create a global instance with rendering parameters.
ddgCulling cullingInfo;
// Texture mapping functions for the HOM.
#define HOMnumber     9
// ------------ Screen coordinates for HOM maps.
static int mc[HOMnumber][3] = 
   {{256,0,0},  {128,0,96},{64,128,160},{32,128,128},
    {16,128,112},{8,128,104},{4,128,100},{2,128,98}, {1,128,96}};

static unsigned long  _occ[1];

// ----------------------------------------------------------------------
// ddgCulling:
//   Initialize the culling infomation.
ddgCulling::ddgCulling(int w, int h)
{
  scale = 0;
  viewCulling = 0;
  occlusionCulling = 0;
  HOMLimit = 10; // The number of leaves to render for occlusion culling.
  HOMWidth = w;
  HOMHeight = h;

  // Image size is a power of 2 plus 1.
  // If window is resizable this should be reallocated. TODO
  _tw = (int)pow(2.0,(int)(log(HOMWidth)/log(2.0)+1.0));
  _th = (int)pow(2.0,(int)(log(HOMHeight)/log(2.0)+1.0));
  _mapBuffer = (void*) malloc(_tw*_th*4*sizeof(GLubyte));
 
}

// ----------------------------------------------------------------------
// ddgCulling:
//   Free the culling buffer.
ddgCulling::~ddgCulling()
{
  free(_mapBuffer);
}

// ----------------------------------------------------------------------
// scaleBox:
//   Rescale the bbox's area from screen coordinates to
//   HOM pixel coordinates.
void ddgCulling::scaleBox( double w, double h)
{
  _minx *= HOMWidth/w;
  _maxx *= HOMWidth/w;
  _miny *= HOMHeight/h;
  _maxy *= HOMHeight/h;
}
// ----------------------------------------------------------------------
// _HOMtest:
//   Recursively test the given screen area against the HOM maps.
//
bool ddgCulling::HOMvisible(void)
{
  bool visible;
//  renderInfo.setWindow(2);
  visible = _HOMtest(0,0,HOMWidth,HOMHeight,9,0,0);
//  renderInfo.setWindow(1);
  return visible;

}
// ----------------------------------------------------------------------
// _HOMtest:
//   Recursively test the given screen area against the HOM maps.
// sx,sy:
//   Starting x and y screen location.
// dx,dy:
//   Delta x and y (width and height) of the area to test.
// level:
//   The current HOM level.
// px,py:
//   The x and y pixel position in the HOM.
//
// Returns true if visible, false if culled.
//
bool ddgCulling::_HOMtest( float sx, float sy, float dx, float dy, int level, int px, int py)
{
  int i,j;
  
  if (level==2) return true;  // Break out of the recursion. 
  // TODO adjust to be sqrt(largest HOM).
  // printf("Testing %d %d %d:%3.1f %3.1f - %3.1f %3.1f\n",level,px,py,sx,sy,sx+dx,sy+dy);
  // Otherwise analyse the 4 sub squares.
  // If black, the area is not occluded => render.
  // If not white, the area is partly occluded, analyse sub region.
  // If white, it is occluded, skip it.
  dx /= 2.0;
  dy /= 2.0;
  level--;

  for (i=0;i<2;i++)
    for (j=0;j<2;j++)
      {
      // Check that area to be tested is in the bbox,
      // if not do not test.
      if ((_maxx>sx+i*dx)&&(_minx<sx+dx+i*dx)
        &&(_maxy>sy+j*dy)&&(_miny<sy+dy+j*dy))
        {
        glReadPixels(mc[level][1]+px+i, mc[level][2]+py+j,
           1, 1, GL_RGBA, GL_UNSIGNED_BYTE,(GLubyte*)_occ);
	//	        printf("%8x:%d %d %3.1f %3.1f - %3.1f %3.1f\n", _occ[0],i,j, sx+i*dx, sy+j*dy, sx+dx+i*dx, sy+dy+j*dy);
        if       (_occ[0] == 0xFF000000) return true; // Black
        else if ((_occ[0] != 0xFFFFFFFF)     // Grey
          && _HOMtest(sx+i*dx,sy+j*dy,dx,dy,
                  level,2*(px+i),2*(py+j))) return true;
        // Else white, can ignore, blocked.
        }
      }

  // If all 4 subzone are occluded, do not render.
  return false;
}

// ----------------------------------------------------------------------
// copybuftotexture:
//   Copy the current HOM from the window buffer into a texture.
//   This may not be needed if the HOM was rendered directly into
//   texture memory. TODO
void ddgCulling::copybufToTexture(int sx, int sy, int dx, int dy )
{
  // Get the original image from the backbuffer.
  glReadPixels(sx,sy,dx,dy,GL_RGBA,GL_UNSIGNED_BYTE,_mapBuffer);
}

// ----------------------------------------------------------------------
// setupTexture:
//   Initialize the texture buffer.  setupTexture(mc[0][0],mc[0][0]);

//
void ddgCulling::setupTexture(int dx, int dy)
{
  static bool init = false;
  glMatrixMode(GL_TEXTURE);
  glLoadIdentity();
  // Set the scaling for the texture map.
  glScalef((GLfloat)1.0/(dx-1),(GLfloat)1.0/(dy-1),1.0);
  if (!init)
  {
    // Set the texture mode.
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
    // Set up the filters used to minify and magnify the texture when
    // mapping to the surface.
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    init = true;
  }
   
  // Set up the two-dimensional texture map (values) as loaded
  // from the back buffer.
  glTexImage2D(GL_TEXTURE_2D, 0, 4, dx, dy, 0,
           GL_RGBA, GL_UNSIGNED_BYTE, _mapBuffer);
  glMatrixMode(GL_MODELVIEW);
  glEnable(GL_TEXTURE_2D);
}

void ddgCulling::finishTexture(void)
{
  glDisable(GL_TEXTURE_GEN_S);
  glDisable(GL_TEXTURE_GEN_T);
  glDisable(GL_TEXTURE_2D);  
}


// ----------------------------------------------------------------------
// buildHOMMap:
//   Construct Hierarchical Occlusion maps at all resolutions down to a
//   two by two pixel.
//
void ddgCulling::buildHOMMap(void)
{
  int i;

  // Draw data into HOM.
  // Turnoff lighting etc.
  //  renderInfo.setWindow(1);
  // Get the screen dimensions.
  static GLint vp[4] = {0,0,0,0};
  glGetIntegerv(GL_VIEWPORT,vp);
  glReadBuffer(GL_BACK);
  // Grab it.
  copybufToTexture(mc[0][1],mc[0][2],mc[0][0],mc[0][0]);

  // HOM2 is the source (black and white) for all the other maps.
//  renderInfo.setWindow(2);
  glClear(GL_COLOR_BUFFER_BIT);
  glPolygonMode(GL_FRONT,GL_FILL);
  // Create the 1st HOM.
  setupTexture(mc[0][0],mc[0][0]);
  glBegin(GL_QUADS);
    glTexCoord2i(vp[0],vp[1]);
    glVertex2i(mc[1][1],          mc[1][2]);
    glTexCoord2i(vp[0],vp[3]-1);
    glVertex2i(mc[1][1],          mc[1][2]+mc[1][0]);
    glTexCoord2i(vp[2]-1,vp[3]-1);
    glVertex2i(mc[1][1]+mc[1][0], mc[1][2]+mc[1][0]);
    glTexCoord2i(vp[2]-1,vp[1]);
    glVertex2i(mc[1][1]+mc[1][0], mc[1][2]);
  glEnd();

  // Create all the sub images.
  for (i = 1; i < HOMnumber - 1; i++)
    {
    copybufToTexture(mc[i][1],mc[i][2],mc[i][0],mc[i][0]);
    setupTexture(mc[i][0],mc[i][0]);
    glBegin(GL_QUADS);
      // Scale down. 
      glTexCoord2i(0,0);
      glVertex2i(mc[i+1][1],            mc[i+1][2]);
      glTexCoord2i(0,mc[i][0]-1);
      glVertex2i(mc[i+1][1],            mc[i+1][2]+mc[i+1][0]);
      glTexCoord2i(mc[i][0]-1,mc[i][0]-1);
      glVertex2i(mc[i+1][1]+mc[i+1][0], mc[i+1][2]+mc[i+1][0]);
      glTexCoord2i(mc[i][0]-1,0);
      glVertex2i(mc[i+1][1]+mc[i+1][0], mc[i+1][2]);
    glEnd();
    }
  finishTexture(); 
  // Switch rendering back to starting HOM.
//  renderInfo.setWindow(1);
}





