/*
    Copyright (C) 1997, 1998, 1999, 2000 by Alex Pfaffe
	(Digital Dawn Graphics Inc)
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

// This class stores static information needed to perform
// normal classification and to use this to perform culling
// operations.
// Algorithm to optimize backface culling of large polygon sets.
// Define a static array of the normal vectors (sN) of an icosahedron.
// This represents and evenly spaced set of normals (20) in R3.
// Enumerate sN in such a way that vectors i and (i+10)%20 = <0,0,0> holds
// true. Two vector pointing in opposite directions.
//
// Preprocessing step for each triangle in the polyset:
// Calculate the normal tN and find out in which sector sN the triangle should
// be associated with based on the minimum angle between tN and sN.
// Add triangle t into the appropriate category.
//
// Memory overhead 20/80/320 arrays with a combined storage space of n integer
// indices into the triangle array.
//
// At runtime determine which sector most faces the camera and render the
// fixed set of arrays which are associated with that sector.
//
// In start field with 18000+ stars and 80 sectors approx 200/400 stars/sector
// 300x300 FPS 7.7/8.3     with clipping FPS 11.1/12.5   47.5% gain in speed
//
#include <stdlib.h>

#include "render/ddgfacet.h"
#include "render/ddgcamera.h"
// ----------------------------------------------------------------------

#define X 0.525731112119133606
#define Z 0.850650808352039932
// Facial radius angle across face of one triangle.
#define faceangle                   41.810314895778596167

ddgVector3 ddgFacetSet::_icovertices[12];

// Initialize with the initial icosahedron polygons.
int ddgFacetSet::_icopolygons[20][3] = {
    {0,4,1},{0,9,4},{9,5,4},{4,5,8},{4,8,1},
    {8,10,1},{8,3,10},{5,3,8},{5,2,3},{2,7,3},
    {7,10,3},{7,6,10},{7,11,6},{11,0,6},{0,1,6},
    {6,1,10},{9,0,11},{9,11,2},{9,2,5},{7,2,11}};

ddgFacetSet::~ddgFacetSet(void)
{
	delete [] _vertices;
}

bool ddgFacetSet::subdivide( ddgVector3 v1, ddgVector3 v2, ddgVector3 v3, long depth)
{
	if (--depth)
	{
		ddgVector3 v12(v1+v2), v23(v2+v3), v31(v3+v1);
		v12.normalize();
		v23.normalize();
		v31.normalize();
		subdivide(v1,v12,v31,depth);
		subdivide(v2,v23,v12,depth);
		subdivide(v3,v31,v23,depth);
		subdivide(v12,v23,v31,depth);
	}
	else
	{
		_vertices[_pcount*3+0] = v1;
		_vertices[_pcount*3+1] = v2;
		_vertices[_pcount*3+2] = v3;
		_pcount++;
	}
	return ddgSuccess;
}

bool ddgFacetSet::init(ddgContext *c) 
{ 
	if (super::init(c))
		return ddgFailure;
	// Level 1 = 20, 2 = 80, 3 = 320 facets.
	ddgAssert(_levels > 0 && _levels < 4 );
	// From redbook (56) example 
	unsigned int i;
	unsigned int psz[] = {20,80,320,960};
	// Initialize with initial icosahedron vertices.
	_icovertices[0].set(-X,0.0,Z);
	_icovertices[1].set(X,0.0,Z);
	_icovertices[2].set(-X,0.0,-Z);
	_icovertices[3].set(X,0.0,-Z);
	_icovertices[4].set(0.0,Z,X);
	_icovertices[5].set(0.0,Z,-X);
	_icovertices[6].set(0.0,-Z,X);
	_icovertices[7].set(0.0,-Z,-X);
	_icovertices[8].set(Z,X,0.0);
	_icovertices[9].set(-Z,X,0.0);
	_icovertices[10].set(Z,-X,0.0);
	_icovertices[11].set(-Z,-X,0.0);
	_vertices = new ddgVector3[psz[_levels-1]*3];
	ddgAsserts(_vertices,"Failed to Allocate memory");
	ddgMemorySet(ddgVector3,psz[_levels-1]*3);

	// For each starting triangle.
	for (i = 0; i < 20; ++i) {
    subdivide(
		_icovertices[_icopolygons[i][0]],
		_icovertices[_icopolygons[i][1]],
		_icovertices[_icopolygons[i][2]],
		_levels);
	}

	ddgAssert (_pcount == psz[_levels-1]);
	return ddgSuccess;
}

bool ddgFacetSet::draw(ddgContext *)
{
	unsigned int i;
	glBegin(GL_TRIANGLES);
	for (i = 0; i < _pcount; i++)
    {
		glNormal3fv(_vertices[i*3+0]); glVertex3fv(_vertices[i*3+0]);
		glNormal3fv(_vertices[i*3+1]); glVertex3fv(_vertices[i*3+1]);
		glNormal3fv(_vertices[i*3+2]); glVertex3fv(_vertices[i*3+2]);
	}
	glEnd();
	return ddgSuccess;
}

bool ddgClippedFacetSet::init( ddgContext *ctx )
{
	if (super::init(ctx)) return ddgFailure;

	ddgAssert(_vertices);
	if (!_on)
	{
		_on = new bool[count()];
		ddgMemorySet(bool,count());
	}
	if (!_centre)
	{
		_centre = new ddgVector3[count()];
		ddgMemorySet(ddgVector3,count());
		unsigned int i;
		// Calculate the centre point of each polygon.
		for (i = 0; i < count(); i++)
		{
			_centre[i] = _vertices[i*3+0] + _vertices[i*3+1] + _vertices[i*3+2];
			_centre[i].divide(3.0);
		}
	}

	// Take 2 neighbouring normals to calculate the face angle.
	_faceAngle = _vertices[0].angle(&(_vertices[2]));
	_clipAngle = 0;
	_visAngle = 0;
	return ddgSuccess;
}

// Calculate the visible set of iso polygons and turn those on.
bool ddgClippedFacetSet::clipVisibleSet( ddgContext *ctx )
{
	unsigned int i;
	if (!ctx->dirty())
		return false;

	if (_clipAngle == 0)
	{
		/*
		* Set clip angle in degrees.
		* When using camera FOV, use the formula:
		*   atan(sqrt(1+aspect*aspect) * tan(fov/2)).
		* this is to account for the fact that the camera fov is
		* measured in width and we need the fov at its widest point
		* which will be a diagonal.
		*/
		_clipAngle = ddgAngle::radtodeg(atan(sqrt(1.0+ctx->aspect()*ctx->aspect())
						 *tan(ddgAngle::degtorad(ctx->fov()/2.0)))
									);
		_visAngle = _clipAngle + _faceAngle/2.0;
	}
	ddgVector3 *v = ctx->forward();
	// If a NULL vector is given, everything is visible.
	bool noclip = v->null();

	float angle;
	for (i=0; i< count(); i++)
    {
		angle = v->angle(&(_centre[i]));
		_on[i] = noclip ? true 
			:((angle < _visAngle)
			? true : false);
    }
	return true;
}
// Given a vector return the sector index of the nearest normal.
unsigned char ddgClippedFacetSet::classify( ddgVector3 v)
{
	v.normalize();
	unsigned int i, nc = 0;  // Assume sector 0.
	float a, angle = v.angle(&(_centre[nc]));
	for (i=1; i< count(); i++)
	{
		if ((a = v.angle(&(_centre[i]))) < angle) // We have a better sector.
		{
			nc = i;
			angle = a;
		// if (angle < _faceAngle) break;  // As good as it will get. fA is not right
		}
	}
	return nc;
}

// For debugging purposes the visible surface set can be displayed.
bool ddgClippedFacetSet::draw(ddgContext *)
{
	unsigned int i;
	glColor4ub(0,0,255,125);
	glShadeModel(GL_FLAT);
	glBegin(GL_TRIANGLES);
	ddgVector3 v1,v2,v3;
	for (i = 0; i < count(); i++)
    {
		if (_on[i])
	    {
			glNormal3fv(_vertices[i*3+0]); glVertex3fv(_vertices[i*3+0]);
			glNormal3fv(_vertices[i*3+1]); glVertex3fv(_vertices[i*3+1]);
			glNormal3fv(_vertices[i*3+2]); glVertex3fv(_vertices[i*3+2]);
		}
	}
	glEnd();
	glColor3ub(0,0,50);
	for (i = 0; i < count(); i++)
    {
		if (_on[i])
		{
			glBegin(GL_LINE_LOOP);
			glVertex3fv(_vertices[i*3+0]);
			glVertex3fv(_vertices[i*3+1]);
			glVertex3fv(_vertices[i*3+2]);
			glEnd();
		}
    }
	return ddgSuccess;
}

ddgClippedFacetSet::~ddgClippedFacetSet(void)
{
	delete [] _on;
	delete [] _centre;
}