/*
    Copyright (C) 1997, 1998, 1999 by Alex Pfaffe
	(Digital Dawn Graphics Inc)
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

// GL includes
#include <GL/glut.h>
//
#include "render/ddgmeshrender.h"
#include "struct/ddgbtree.h"
#include "render/ddgcamera.h"
#include "ext/ddgtxtsyn.h"

// ----------------------------------------------------------------------

ddgMeshRender::ddgMeshRender( ddgTBinMesh *mesh, ddgHeightMap *height, ddgStr *tfile, ddgStr *efile, int size, int scale, bool meshcolor)
{
    _mesh = mesh;
	_texture = NULL;
	_height = height;
	if (tfile)
		_tfile.assign(tfile);
	if (efile)
		_efile.assign(efile);
	_size = size;
	_scale = scale;
	_curvature = 0.001;
    // Make room for vertices/normals/color tuples
    varray.size((mesh->absMaxDetail()*3*11)/10,(mesh->absMaxDetail()*3*11)/10);

    mode.flags.solid = true;
	mode.flags.texture = tfile ? true : false;
	mode.flags.normal = true;
	mode.flags.wire = true;
	mode.flags.smooth = true;

	_meshcolor = meshcolor;

	selected = 0xFFFFFFF;
}

ddgMeshRender::~ddgMeshRender( )
{
	unsigned int i;
	if (_scale > 0)
	{
		for (i = 0; i < _mesh->getBinTreeNo()/2; i++)
			delete _texture[i];
	}
	else if (_size)
		delete _texture[0];
	delete[] _texture;
}

bool ddgMeshRender::init( ddgContext *ctx)
{
	unsigned int i, r = 0, c = 0, k;
	// Avoid being called twice (once for each camera).
	if (super::init(ctx))
		goto error;

	if (_size && _tfile.length()>0)
	{
		ddgStr	tbasename(_tfile);
		bool loadImage = true;
		ddgImage *image = new ddgImage();
		unsigned int	width = 0;
		unsigned int height = 0;
		unsigned int channels = 0;
		// See if the split blocks are already available.
		if (_scale )
		{
			ddgStr	tblock(32);
#ifdef DDGSTREAM
			ostrstream tblocks(tblock.s,32);
			char *t = tbasename.findChar('.');
			ddgAssert(t);
			t[0] = 0;
			tblocks << tbasename.s << ((int)0) << ".tga" << '\0';
#endif
			// See if we already have block textures, if so we do not need to load the mother image.
			if (image->readFile(tblock) == ddgSuccess )
			{
				channels = image->channels();
				width = _scale * _size;
				height = _scale * _size;
				delete image;
				image = NULL;
				loadImage = false;
			}
		}

		if (loadImage)
		{
			// Try to find the image, if it doesn't exists or it is of the wrong size
			// regenerate it.
			if (image->readFile(_tfile) == ddgFailure)
			{	
				// Terrain texture was not found, should we synthesize?
				if (_scale)
				{
					ddgImage	*ecomap = new ddgImage();
					ddgVector2	vSize(_scale,_scale);
					if (ddgTextureSynthesizer::synthesizeSurfaceTexture( _height, image, vSize, ecomap ) == ddgFailure)
						return ddgFailure;
					// Save texture to disk for future reuse.
					image->writeTGA(_tfile);
					ecomap->writeTGA(_efile);
					delete ecomap;
				}
				else  // Texture does not exist and no scale was specified.
				{
					ddgErrorSetStr(Failure,"Texture " << _tfile.s << 
						"does not exist and I can not generate a texture if no scale is specified.");
					goto error;
				}
			}
			else // The mother image loaded fine.
			{
				// If image is too small to span terrain, generate error.
				if (_scale && (image->cols()-1 < _scale*_size))
				{
					ddgErrorSetStr(Failure,"Texture " << _tfile.s << 
						"is too small to cover the terrain.");
					goto error;
				}
				else // Set scale.
					_scale = (image->cols()/_size);
			}
			channels = image->channels();
			width = image->cols()-1;
			height = image->rows()-1;

		}
		// Construct the texture array.
		_texture = new ddgTexture*[_scale ? _mesh->getBinTreeNo()/2 : 1];
		ddgMemorySet(ddgTexture*,_scale ? _mesh->getBinTreeNo()/2 : 1);

		if (_scale)
		{
			// Input texture is size 2^n+1 x 2^m+1
			// Assumption:
			// bintree pairs each form a square region on the mother texture.
			// Eg. 0/1  is bottom left n-1/n is top right of mother texture.
			// each bintree pair can be mapped left to right, top to bottom
			// onto the mother texture.
			k = width * height;
			// Calculate size of split textures.
			k = 2*k / _mesh->getBinTreeNo();
			k = sqrt(k);
			// $TODO: Deal with edge overlap.
			char fileNameBuf[32];
			for (i = 0; i < _mesh->getBinTreeNo()/2; i++)
			{
				ddgConsoleMessage::progress("Texture blocking",i,_mesh->getBinTreeNo()/2);
#ifdef DDGSTREAM
				ostrstream msg(fileNameBuf,32);
				msg << tbasename.s << i << ".tga" << '\0';
#endif
				ddgStr sname;
				sname.assign(fileNameBuf);
				ddgImage *simg = new ddgImage();
				// Load the file if it exists, otherwise create if from the texture.
				if (simg->readFile(sname)== ddgFailure)
				{
					ddgAsserts(image,"No base image to create block textures from");
					simg->allocate(k,k,channels);
					simg->import(image,c,r,0,0,k,k);
					simg->writeTGA(fileNameBuf);
				}
				_texture[i] = new ddgTexture(simg);
				_texture[i]->linear(true);
				_texture[i]->repeat(false);
				_texture[i]->scale( simg->rows(), simg->cols());
				_texture[i]->mode(_meshcolor ? ddgTexture::MODULATE : ddgTexture::DECAL);
				_texture[i]->init();
				c = c + k;
				if (c + 1 >= width)
				{
					r += k;
					c = 0;
				}
			}
			ddgConsoleMessage::progress("Texture blocking",1,1);

			// Release mother image because we no longer need it.
			delete image;
			image = NULL;
		}
		else 
		{
			k = image->cols() / ddgTBinMesh_size;
			_texture[0] = new ddgTexture(image);
			_texture[0]->scale(k,k);
			_texture[0]->linear(true);
			_texture[0]->mode(_meshcolor ? ddgTexture::MODULATE : ddgTexture::DECAL);
			_texture[0]->init();
		}
	}

    if ( _mesh->init(ctx))
    {
        mode.flags.init = false;
        ddgErrorSetStr(Failure,"Mesh renderer failed to initialize!");
        goto error;
    }

	return ddgSuccess;

error:
	return ddgFailure;
}


// ---------------------------------------------------------------------
// Render the mesh.
//
bool ddgMeshRender::draw(ddgContext *ctx)
{
	static bool modified = true;

	if (ctx->quality() != ddgContext::DEBUG)
	{
		// Optimize the mesh w.r.t. the current viewing location.
		modified = _mesh->calculate(ctx);
	}

	//
	// A triangle is defined by the point that it
	// splits. To render it, you need to render its
	// father, and two other grand/great-grand parents.
	ddgNode::RenderFlags rm = mode;
	if (ctx->quality() == ddgContext::DEBUG)
	{
		mode.all = 0;
		mode.flags.wire = true;
	}
	glDisable(GL_ALPHA_TEST);
	glDisable(GL_BLEND);
	glDepthFunc(GL_LEQUAL);
	if (mode.flags.smooth)
		glShadeModel(GL_SMOOTH);
	else
		glShadeModel(GL_FLAT);

	ddgVector2 pos(ctx->control()->position()->v[0],ctx->control()->position()->v[2]);

	if (rm.flags.solid && ctx->quality() != ddgContext::DEBUG)
	{
		mode.flags.wire = false;

		unsigned int i = 0, s = 0, d = 0;
		ddgTBinTree *bt;

		if (modified)
		{
			ddgVBuffer::reset();
			// Update the vertex buffers.
			ddgCacheIndex ci = 0;
			while (i < _mesh->getBinTreeNo())
			{
				bt = _mesh->getBinTree(i);
				unsigned int v = 0;
				if (bt && bt->vis() != ddgOUT)
				{
					// Render each triangle.
					ci = bt->chain();
					// Render each triangle.
					while (ci)
					{
						ddgTNode *tn = (ddgTNode*) _mesh->tcache()->get(ci);
 						if (drawTriangle(bt, tn->tindex()) == ddgSuccess)
							v++;
						ci = tn->next();
					}
				}
				bt->visTriangle(v);
				i++;
			}

		}
		// Render the vertex buffer piece by piece.
		if (!mode.flags.texture)
			glColor3ub(128,160,128);

		i = 0;
		while (i < _mesh->getBinTreeNo())
		{
			d = _mesh->getBinTree(i)->visTriangle() + _mesh->getBinTree(i+1)->visTriangle();
			if (d > 0)
			{
				if (mode.flags.texture && _texture && _texture[i/2])
					_texture[i/2]->activate();
				// Render this bintree.
				range(s,d);

				super::draw(ctx);
				s = s+d;
			}
			i = i+2;
		}
	}
	if (mode.flags.texture && _texture )
		_texture[0]->deactivate();

	// Wire overlay
	if (rm.flags.wire)
	{
		bool lighting = false;
		glShadeModel(GL_FLAT);

		if (glIsEnabled(GL_LIGHTING) == GL_TRUE)
		{
			lighting = true;
			glDisable(GL_LIGHTING);
		}
		mode.flags.solid = false;
		mode.flags.wire = true;

		// Update the vertex buffers.
		ddgCacheIndex ci = 0;
		unsigned int i = 0;
		ddgTBinTree *bt;
		while (i < _mesh->getBinTreeNo())
		{
			if (bt = _mesh->getBinTree(i))
			{
				// Render each triangle.
				ci = bt->chain();
				// Render each triangle.
				while (ci)
				{
					ddgTNode *tn = (ddgTNode*) _mesh->tcache()->get(ci);
 					drawTriangleWire(bt, tn->tindex());
					ci = tn->next();
				}
			}
			i++;
		}

		if (lighting == true)
			glEnable(GL_LIGHTING);
	}

	mode = rm;
	return ddgSuccess;
}


/**
 *  Retrieve info for a single triangle.
 *  Returns true if triangle should be rendered at all.
 */
bool ddgMeshRender::drawTriangleWire( ddgTBinTree *bt, ddgVBIndex tvc )
{
    ddgVector3 p[3];
    ddgTriIndex tv[3] = {bt->parent(tvc), _mesh->v1(tvc), _mesh->v0(tvc)};

	bt->vertex(tv[0],&p[0]);
	bt->vertex(tv[1],&p[1]);
	bt->vertex(tv[2],&p[2]);
	static ddgColor3 c[3];
	unsigned int pr = bt->priority(tvc);
	if (pr < 0) pr = 0;
	if (pr > 100) pr = 100;
	float pp = ((float)pr) / 100.0;
	if (bt->qmcacheIndex(tvc))
		c[0].set(255,255,255);
	else if (!bt->visible(tvc) )
		c[0].set(255,0,0);
	else if (pr == 0)
		c[0].set(0,150,0);
	else if (bt->fullyvisible(tvc))
		c[0].set(0,0,50+200*pp);
	else // partially visible
		c[0].set(125+125*pp,125+125*pp,125+125*pp);
	// Color the selected triangle yellow.
	if (selected == tvc)
		c[0].set(255, 255, 0 );

    glColor3ubv(c[0]);
	glNormal3f(0,0,0);
	glBegin(GL_LINE_LOOP);
	glVertex3fv(p[0]);
	glVertex3fv(p[1]);
	glVertex3fv(p[2]);
	glEnd();
    return ddgSuccess;
}

/**
 *  Retrieve info for a single triangle.
 *  Returns true if triangle should be rendered at all.
 */
bool ddgMeshRender::drawTriangle( ddgTBinTree *bt, ddgVBIndex tvc )
{
	if ( !bt->visible(tvc))
		return ddgFailure;

    ddgVector3 p[3];
    ddgTriIndex tv[3] = {bt->parent(tvc), _mesh->v1(tvc), _mesh->v0(tvc)};
	static ddgVector2 t[3];
	unsigned int bufindex[3] = {0, 0, 0};
	unsigned int i;
	for (i = 0; i < 3; i++)
	{
// $TODO: Since the parent, v1 and v0 are not in the mesh we can not use
// them to store the buffer indexes, this would require an additional field
// to always be present.
//		bufindex[i] = bt->vbufferIndex(tv[i]);
//		ddgCacheIndex ci = bt->tcacheId(tv[i]);
//		if (!bufindex[i])
		{
			bt->vertex(tv[i],&p[i]);
			if (varray.textureOn())
				bt->textureC(tv[i],&(t[i]));

			bufindex[i] = varray.pushVT(&p[i],&t[i]);
		}
		// Record that these vertices are in the buffer.
//		bt->vbufferIndex(tv[i],bufindex[i]);
 	}
    varray.pushTriangle(bufindex[0],bufindex[1],bufindex[2]);

    return ddgSuccess;
}

