/*
    Copyright (C) 1997, 1998, 1999, 2000 by Alex Pfaffe
	(Digital Dawn Graphics Inc)
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include <fstream.h>

#include "render/ddgstars.h"
#include "render/ddgfacet.h"
// ----------------------------------------------------------------------

int ddgStar::compare(const void *s1, const void *s2)
{
	return ((ddgStar *)s1)->facet - ((ddgStar*)s2)->facet;
}

ddgStarClass::ddgStarClass(ddgStr *filename, unsigned int l): scale(10.0), _group(0), _starCount(0),
		ddgClippedFacetSet(l),minMagnitude(0)
{
#ifdef _DEBUG
    cerr << "Loading star data";
#endif
	_facets = NULL;
	char *fullName = ddgFileHelper::getFullFileName(*filename);
	ifstream fin(fullName);
	if (!fin.good())
	{
		ddgErrorSetStr(FileRead,"Could not load stars data file " << ((char *) (fullName ? fullName : "(null)")));
		ddgError::report();
        mode.flags.init = false;
	}
	else
	{
		float b,a,d;
		ddgVector3 c,p;

		_starCount = 0;

		while ( !fin.eof() && _starCount<ddgStarClass_max) {
			fin >> a >> d >> b;
			if (b > minMagnitude)
			{
	//			fout << a << " " << d << " " << b << endl;
				_stars[_starCount].brightness =
					(unsigned char)ddgUtil::clamp(255*sqrt(b*100)/10,0,255);
				//    PITCH       YAW         ROLL
				p.set(d,a,0);
				p.angletovector(&c,0,0);
				c.normalize();
				c.assign(_stars[_starCount].pos);
				_starCount++;
	#ifdef _DEBUG
				if (_starCount %1000 == 0) cerr << '.';
	#endif
			}
		}
		fin.close();
		_group = new unsigned int[_starCount];
		ddgMemorySet(unsigned int,_starCount);
	}
//	fout.close();
#ifdef _DEBUG
	cerr << '\n';
#endif
	if (!_group)
	{
		ddgErrorSet(FileAccess,"No stars were read in");
        mode.flags.init = false;
	}
}

ddgStarClass::~ddgStarClass()
{
	unsigned int i = count();
	while (i)
		delete _facets[--i];
	delete [] _facets;
	delete [] _group;
}

// brightness range max = 328 min = 0
bool ddgStarClass::init( ddgContext *ctx)
{
	if (super::init(ctx) == ddgFailure)
		return ddgFailure;

	unsigned int i = 0;
	ddgAssert(_stars);
	// Determine which facet each star is nearest to.
	while ( i < _starCount)
	{
		_stars[i].facet = classify(_stars[i].pos);
		i++;
	}
	// Sort them by facet.
	::qsort(_stars,_starCount,sizeof(ddgStar),ddgStar::compare);
	// Find the starting point of each facet.
	for (i = 0; i < _starCount; i++)
	{
		_group[_stars[i].facet] = i;
	}
	// Define a red and blue color.
//	ddgColor3 cs(255,0,0), ce(0,0,255);
	// Create a spectrum of colors.
//	_colorset.spread(&cs,&ce,0,_starCount,_starCount);

	unsigned int j,k;
	_facets = new ddgVBuffer*[count()];
	ddgMemorySet(ddgVBuffer*,count());
	// For each group build a vertex buffer
	for (k = 0; k < count(); k++)
	{
		int s = 2+_group[k] - (k > 0 ? _group[k-1]:0);
		ddgAssert(s>0);
		{
			_facets[k] = new ddgVBuffer(ddgVArray::point);
			_facets[k]->varray.size(s,s);
			_facets[k]->varray.renderMode(false,false,true);
			_facets[k]->init(ctx);
			_facets[k]->reset();

			// Render all the stars in this group.
			for (j = k > 0 ? _group[k-1] : 0; j < _group[k]; j++)
			{
				unsigned char b = _stars[j].brightness*0.5;
				ddgColor4 c(b,b,b,0);
				_facets[k]->varray.pushPoint(_facets[k]->varray.pushVC(&(scale*_stars[j].pos),c));
			}
		}
	}
	ddgMemorySet(ddgVBuffer,count());

	return ddgSuccess;
}

// Forward facing vector and a field of view.
// Could create a display list of each facet, and
// call it as needed.
bool ddgStarClass::draw( ddgContext *ctx )
{
	bool modified = false;
	// Adjust star brightness according to time.
	float br = 1.0 - (ctx->clock() ? ctx->clock()->getDayLight(): 0.0);
	// No stars during the day.
//	if (br == 0.0)
//		return ddgSuccess;

	static float oldbr = 0.0;
	
	unsigned int j = 0, k = 0;
	if (ctx->quality() != ddgContext::DEBUG)
	{
		modified = clipVisibleSet( ctx );
	}
	if (modified || oldbr != br)
	{
		listInvalid();
	}
	if (!listValid())
	{
		oldbr = br;
		listStart();
		glPushMatrix();
		// Scale them some distance out from the camera.
		if (ctx->quality() == ddgContext::DEBUG)
			glScalef(0.2*scale,0.2*scale,0.2*scale);
//		else
//			glScalef(scale,scale,scale);

		glShadeModel(GL_FLAT);

		glDepthFunc(GL_LEQUAL);
		if (_smooth)
		{
			glEnable(GL_POINT_SMOOTH);
			glPointSize(1.0);
		}
		if (ctx->quality() != ddgContext::DEBUG)
		{
			// For each group.
			for (k = 0; k < count(); k++)
			{
				// If group is in view.
				if (active(k))
				{
					// Render all the stars in this group.
					_facets[k]->draw(ctx);
/*
					glBegin(GL_POINTS);
					for (j = k > 0 ? _group[k-1] : 0; j < _group[k]; j++)
					{
						unsigned char b = _stars[j].brightness*br;
						glColor3ub(b,b,b);
						glVertex3fv(_stars[j].pos);
					}
					glEnd();
*/
				}
			}
		}

		if (_smooth)
			glDisable(GL_POINT_SMOOTH);
		glDepthFunc(GL_LESS);
		glPopMatrix();
		listEnd();
	}
	listRender();

	return ddgSuccess;
}
