/*
    Copyright (C) 1997, 1998, 1999, 2000 by Alex Pfaffe
	(Digital Dawn Graphics Inc)
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "render/ddgsun.h"


// ----------------------------------------------------------------------
ddgSun::~ddgSun()
{
	delete _image;
}

ddgSun::ddgSun( ddgVector3 p, ddgCamera *c, unsigned short size)
{
	_image = new ddgImage(4);
	ddgMemorySet(ddgImage,1);
	_image->allocate(size,size);
	_texture = NULL;
	_cnight.set(200,120,120,0);		// Night color.
	_ctwilight.set(200,120,120,0);	// Twilight color.
	_cday.set(255,255,200,0);		// Daytime color.
	_pos =p;
	_scale = 40; // Sun texture will affect 40% of screen width.
	_camera = c;
}

bool ddgSun::init(ddgContext *ctx)
{
    if (super::init(ctx))
        return ddgFailure;

	// Create a sun image.
	float t;
	unsigned int v, tr;
	int r,c, cx = _image->cols()/2-1, cy = _image->rows()/2-1;
	// Scale factor.
	float radius = 2.0/(float)_image->cols();
	// Only need to calculate 1/8th and reuse the rest.
	for (r=0; r < _image->rows()/2; r++)
	{
		for (c=0; c <= r; c++)
		{
			t = sqrt(r*r+c*c)*radius;
		  
			// t is from 0.0 to 1.0
			// Use exponential curve, which looks more realisitic.
			t = pow(2.0,8.0*(1.0-t));

			v =  (unsigned char) ddgUtil::clamp(t,0,255);
			tr = ddgUtil::clamp(300 - v,0,255);
//			tr = 255;
			_image->set(cx-r,cy-c,    v,v,v,tr);
			_image->set(cx-r,cy+c+1,  v,v,v,tr);
			_image->set(cx+r+1,cy-c,  v,v,v,tr);
			_image->set(cx+r+1,cy+c+1,v,v,v,tr);
			_image->set(cx-c,cy-r,    v,v,v,tr);
			_image->set(cx-c,cy+r+1,  v,v,v,tr);
			_image->set(cx+c+1,cy-r,  v,v,v,tr);
			_image->set(cx+c+1,cy+r+1,v,v,v,tr);
		}
	}
	_texture = new ddgTexture(_image,true);
	ddgMemorySet(ddgTexture,1);
	_texture->scale(_image->rows(),_image->cols());
	_texture->init();
    return ddgSuccess;
}

// Must be rendering inside of cloud sphere.
bool ddgSun::draw(ddgContext *ctx )
{
	if (ctx->quality() == ddgContext::DEBUG)
		return ddgSuccess;

	float day = ctx->clock()->getDayLight();
	static ddgColor4 cnow;
	if (day < 0.5)
		cnow.linterp(_cnight,_ctwilight,day*2);
	else
		cnow.linterp(_ctwilight,_cday,(day-0.5)*2);

	_texture->mode(ddgTexture::MODULATE,cnow);

	// See if sun is outside viewing volume.
	ddgMatrix4	_mm;

	glGetFloatv(GL_MODELVIEW_MATRIX,_mm);
    _curpos[0] = _mm[0]*_pos[0]+_mm[4]*_pos[1]+_mm[8]*_pos[2]+_mm[12];
    _curpos[1] = _mm[1]*_pos[0]+_mm[5]*_pos[1]+_mm[9]*_pos[2]+_mm[13];
    _curpos[2] = _mm[2]*_pos[0]+_mm[6]*_pos[1]+_mm[10]*_pos[2]+_mm[14];

	if (!ctx->visible(_curpos))
		return false;

	// Test if sun is blocked by clouds.
	// Render a pixel at the sun's location,
	// see if we can hit it, if so the sun is visible.
	// - Map suns 3D coords to screen.
	int vp[4] = {0,0,-1,-1};
	_camera->viewport(vp);

	ddgVector3 scrpos;
	_camera->map(_pos,&scrpos);

    // Draw sun effect.
	{
		// TODO: Adjust brightness by the amount of atmosphere/clouds
		// that is blocking.
		// Map pixel coords to ortho coords 0-100.
		int px = (100.0*scrpos[0]/(float)vp[2]-_scale/2.0);
		int py = (100.0*scrpos[1]/(float)vp[3]-_scale/2.0);
		// Switch to 2D Projection.
		// TODO: check if lighting is on.  Cache value.
		bool l = false;
		if ( glIsEnabled(GL_LIGHTING))
		{
			glDisable(GL_LIGHTING);
			l = true;
		}
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		glOrtho(0,100,0,100,-1.0,10000.0);
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();
		// Begin the rendering.
		// Texture mode is MODULATE.
		// Always render regardless of Zbuffer.
		// Blend with existing image.
		_texture->activate();
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_COLOR,GL_ONE);
		glDepthFunc(GL_LEQUAL);
		// Draw sun glare effect.
		glBegin(GL_QUADS);
		glTexCoord2f(0,0);
		glVertex3f(px,py,-9999);
		glTexCoord2f(1,0);
		glVertex3f( px+_scale,py,-9999);
		glTexCoord2f(1,1);
		glVertex3f( px+_scale,py+_scale,-9999);
		glTexCoord2f(0,1);
		glVertex3f( px,py+_scale,-9999);
		glEnd();
		_texture->deactivate();
		glDisable(GL_BLEND);
		if (l)
			glEnable(GL_LIGHTING);
		glDepthFunc(GL_LESS);
		// Switch back to 3D mode
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();
	}
	return ddgSuccess;
}



