/*
    Copyright (C) 1997, 1998, 1999, 2000 by Alex Pfaffe
	(Digital Dawn Graphics Inc)
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include <stdarg.h>
#include <stdio.h>

#include "render/ddgtexture.h"
#include "render/ddgtext.h"

/**
  Supports characters from 32 to 127.
 */
typedef struct ddgFontType
{
	int	width;					// Static spacing between characters also equal to width of widest char in pixels.
	int height;					// Max Height of all characters in pixels.
	int charWidth;				// Number of pixels to use for cha, this is <= width.
	int charHeight;				// Number of pixels to use for cha, this is <= height..
	int charsPerRow;			// Number of characters on each row of the texture map.
	int symbolCount;			// Total number of characters in the entire texturemap.
	char *fontFileName;			// TGA file with font texture.
	ddgImage *fontImage;
	ddgTexture *fontTexture;
	unsigned int textureId;
	unsigned int asciiMap[96];		// Maps ascii values [32-127]to texturemap positions.
} ddgFont;

ddgFont g_ddgDefaultFont = {12,12,9,11,10,96,"fonts/default.tga",NULL,NULL,0,
	{0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,
	32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,49,50,51,52,53,54,55,56,57,58,59,60,61,62,63,
	64,65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,80,81,82,83,84,85,86,87,88,89,90,91,92,93,94,95},
};

// ----------------------------------------------------------------------
bool ddgText::init(ddgContext *ctx)
{
	if ( super::init(ctx))
        return ddgFailure;

	if (NULL == g_ddgDefaultFont.fontImage)
	{
		// Initialize the font data.
		g_ddgDefaultFont.fontImage = new ddgImage();
		ddgMemorySet(ddgImage,1);

		if (g_ddgDefaultFont.fontImage->readFile(ddgStr(g_ddgDefaultFont.fontFileName)))
		{
			delete g_ddgDefaultFont.fontImage;
			g_ddgDefaultFont.fontImage = NULL;
			ddgErrorSetStr(FileRead,"Failed to load font from " << g_ddgDefaultFont.fontFileName);
			return ddgFailure;
		}
		// Convert 3 channel color Alpha to true 4 channel alpha.
		if (g_ddgDefaultFont.fontImage->channels() == 3 && g_ddgDefaultFont.fontImage->mapColorToAlpha() == ddgFailure)
			return ddgFailure;

		ddgAssert(g_ddgDefaultFont.fontImage->channels() == 4);
		if (g_ddgDefaultFont.fontImage->channels() != 4)
			return ddgFailure;

  		g_ddgDefaultFont.fontTexture = new ddgTexture(g_ddgDefaultFont.fontImage,true,false,ddgTexture::DECAL);
		g_ddgDefaultFont.fontTexture->init();
	}

	return ddgSuccess;
}

void ddgText::printf (char *str, ...)
{
	va_list arg;
	static char buf[256];

	va_start (arg, str);
	vsprintf (buf, str, arg);
	va_end (arg);
	_data.assign(buf);
}

bool ddgText::draw( ddgContext *ctx )
{
	char *p = _data.s;
	// Check that there is a string defined.
	if (!p)
		return ddgSuccess;

	ddgAsserts(g_ddgDefaultFont.fontImage,"Could not load font texture.");

	float x = posx(), y = posy();
	float h = g_ddgDefaultFont.height, w= g_ddgDefaultFont.width;
	float dx = 100.0*(g_ddgDefaultFont.charWidth)/_screen[0], dy = 100.0*h/_screen[1];
	// Number of of characters in one row of the texture.
	float tx = 0, ty = 0,
		tdxs = g_ddgDefaultFont.charWidth,
		tdys = g_ddgDefaultFont.charHeight;
	// Begin the rendering.
	// Texture mode is MODULATE.
	// Always render regardless of Zbuffer.
	// Blend with existing image.
	glDepthFunc(GL_ALWAYS);
	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE,GL_ZERO);
	glColor4ubv(background);
	glBegin(GL_QUADS);
	if (!blend)
	{
		glVertex3f(x,y,-0.51);
		glVertex3f(x+_data.l*dx,y,-0.51);
		glVertex3f(x+_data.l*dx,y+dy,-0.51);
		glVertex3f(x,y+dy,-0.51);
	}
	glEnd();
	g_ddgDefaultFont.fontTexture->activate();
	glBlendFunc(GL_ONE, blend ? GL_ONE : GL_ZERO);
	glMatrixMode(GL_TEXTURE);
	glPushMatrix();
	glLoadIdentity();
	// Set the scaling for the texture map.
	glScalef(1.0 / (float)g_ddgDefaultFont.fontImage->cols(),1.0 / (float)g_ddgDefaultFont.fontImage->rows(),1.0);

	glBegin(GL_QUADS);
	for (; p && *p && x < 100; p++)
	{
		if (*p > 31 && *p < g_ddgDefaultFont.symbolCount+32)
		{
			// Get values in pixels.
			tx = (float)(g_ddgDefaultFont.asciiMap[*p-32]%g_ddgDefaultFont.charsPerRow)*w;
			ty = ((int)g_ddgDefaultFont.asciiMap[*p-32]/g_ddgDefaultFont.charsPerRow)*h;
			// TODO Adjust for variable width characters.

			glTexCoord2f(tx,ty);
			glVertex3f(x,y,0);

			glTexCoord2f(tx+tdxs,ty);
			glVertex3f( x+dx,y,0);

			glTexCoord2f(tx+tdxs,ty+tdys);
			glVertex3f( x+dx,y+dy,0);

			glTexCoord2f(tx,ty+tdys);
			glVertex3f( x,y+dy,0);
		}
		// TODO Adjust for variable width characters.
		x +=dx;
	}
	glEnd();

	g_ddgDefaultFont.fontTexture->deactivate();
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glDisable(GL_BLEND);
	glDepthFunc(GL_LESS);
	return ddgSuccess;
}
