/*
    Copyright (C) 1997, 1998, 1999, 2000 by Alex Pfaffe
	(Digital Dawn Graphics Inc)
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include "gl/glut.h"
#include "util/ddgerror.h"
#include "render/ddgviewport.h"
#include "render/ddgkcntl.h"

ddgKeyControl::ddgKeyControl(ddgVector3 *position, ddgVector3 *orientation)
	: super(position,orientation)
{
	_delta.set(1.0,1.0,1.0);
	_beginX = 0;
	_beginY = 0;
	
	_stepForward = 0;
	_stepRight = 0;
	_moving = false;
}

void ddgKeyControl::attach( ddgViewport *vp )
{
	ddgAssert(vp);
	_eventHandler.set(this,skey,sspecialKey,smouse,smouseMotion);
	// Store the default handlers, we will call these if you don't handle the events ourselves.
	_eventHandler.defaultHandler(vp->eventHandler());
	vp->eventHandler(&_eventHandler);
	// Override the key handler with ours.
	_eventHandler.keyHandler(skey);
}

// $TODO adjust movement according to elapsed time.
bool ddgKeyControl::update( void )
{
	super::update();
	if ( _stepForward || _stepRight|| _moving)
	{
		// Setup the view.
		ddgVector3 ff;
		ddgVector3 forward, right, up;
		orientation()->angletovector(&forward,&right,&up);

		// Add to position any forward or backward steps.
		float ss = 1.0;
		if (_stepForward && _stepRight)
			ss = sqrtf(_stepForward*_stepRight);

		if (_stepForward)
		{
			ff = forward;
			ff.multiply(ss*_stepForward);
			position()->add(&ff);
		}
		// Add to position any left or right steps.
		if (_stepRight)
		{
			right.multiply(ss*_stepRight);
			right.multiply(&_delta);
			position()->add(&right);
		}

		// Reset deltas and steps.
		_delta.set(1,1,1);
		_stepForward= 0.0 ;
		_stepRight = 0.0 ;

		_dirty = true;
	}
	return _dirty;
}

/// Event handler, returns true if a change occurred.
bool ddgKeyControl::smouse(void *obj, int button, int state, int x, int y)
{
	ddgKeyControl *kc = (ddgKeyControl*)obj;
	bool h = kc->mouse(button,state,x,y);
	if (!h && kc->_eventHandler.defaultHandler())
		kc->_eventHandler.defaultHandler()->mouseHandler()(kc->_eventHandler.handlerObject(),button,state,x,y);
	return h;
}
/// Event handler, returns true if a change occurred.
bool ddgKeyControl::smouseMotion(void *obj, int x, int y)
{
	ddgKeyControl *kc = (ddgKeyControl*)obj;
	bool h = kc->mouseMotion(x,y);
	if (!h && kc->_eventHandler.defaultHandler())
		kc->_eventHandler.defaultHandler()->mouseMotionHandler()(kc->_eventHandler.handlerObject(),x,y);
	return h;
}
/// Event handler, returns true if a change occurred.
bool ddgKeyControl::sspecialKey(void *obj, int k, int x, int y)
{
	ddgKeyControl *kc = (ddgKeyControl*)obj;
	bool h = kc->specialKey(k,x,y);
	if (!h && kc->_eventHandler.defaultHandler())
		kc->_eventHandler.defaultHandler()->specialKeyHandler()(kc->_eventHandler.handlerObject(),k,x,y);
	return h;
}
/// Event handler, returns true if a change occurred.
bool ddgKeyControl::skey(void *obj, unsigned char k, int x, int y)
{
	ddgKeyControl *kc = (ddgKeyControl*)obj;
	bool h = kc->key(k,x,y);
	if (!h && kc->_eventHandler.defaultHandler())
		kc->_eventHandler.defaultHandler()->keyHandler()(kc->_eventHandler.handlerObject(),k,x,y);
	return h;
}


bool ddgKeyControl::mouse(int button, int state, int x, int y)
{
	// Scene can be spun around Y axis using left
	// mouse button movement.
	if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN) {
		_moving = true;
		_beginX = x;
		_beginY = y;
	}
	else if (button == GLUT_LEFT_BUTTON && state == GLUT_UP) {
		_moving = false;
	}
	return false;
}

bool ddgKeyControl::mouseMotion(int x, int y)
{
	if (_moving) {
		orientation()->v[YAW] -= (x - _beginX);
		orientation()->v[PITCH] += (y - _beginY);
		_beginX = x;
		_beginY = y;
		_moving = true;
		return true;
	}
	return false;
}

bool ddgKeyControl::specialKey( int k, int, int )
{
	float delta = 1;
    if (glutGetModifiers() == GLUT_ACTIVE_CTRL) 
    {
        delta *= 10;
    }
    else if (glutGetModifiers() == GLUT_ACTIVE_ALT)
    {
        delta *= 0.1;
    }
     
	switch (k)
	{
		case GLUT_KEY_DOWN:
			_stepForward -= delta;
			break;
		case GLUT_KEY_UP:
			_stepForward += delta;
			break;
		case GLUT_KEY_LEFT:
			_stepRight -= delta;
			break;
		case GLUT_KEY_RIGHT:
			_stepRight += delta;
			break;
		default:
			return false;
	}
	return true;
}

bool ddgKeyControl::key(unsigned char k, int, int )
{
    float kk = 1.0f;
    if (glutGetModifiers() == GLUT_ACTIVE_CTRL) kk = 5.0f;
    else if (glutGetModifiers() == GLUT_ACTIVE_ALT) kk = 50.0f;
    switch (k)
    {
        case ' ': orientation()->zero(); break;
        case 'z': _delta.v[2] -= kk; break;
        case 'Z': _delta.v[2] += kk; break;
        case 'x': _delta.v[0] -= kk; break;
        case 'X': _delta.v[0] += kk; break;
        case 'y': _delta.v[1] -= kk; break;
        case 'Y': _delta.v[1] += kk; break;
        case '.': orientation()->v[ROLL] += kk; break;
        case ',': orientation()->v[ROLL] -= kk; break;
		case '?':
#ifdef DDGSTREAM
			cerr << "Camera Commands:" << endl
				<< "LEFT/RIGHT ......... Move left/right." << endl
				<< "UP/DOWN ............ Move forward/back." << endl
				<< "SPACE .............. Reset orientation to (0,0,0)." << endl
				<< "x/X ................ Move along x axis." << endl
				<< "y/Y ................ Move along y axis." << endl
				<< "z/Z ................ Move along z axis." << endl
				<< "x/X ................ Move along x axis." << endl
				<< ",/. ................ Roll about forward axis" << endl;
		// Pretend we didn't handle it so that other handlers can report their help.
#endif
        default:  return false;
    }
    return (_moving = true);
}



