/*
    Copyright (C) 1997, 1998, 1999, 2000 by Alex Pfaffe
	(Digital Dawn Graphics Inc)
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#ifndef _ddgParam_Class_
#define _ddgParam_Class_

#include "util/ddgutil.h"
#include "util/ddgstr.h"

/*
 * Class which supports control of a number of parameters.
 * ddgParams can be controlled via keys.
 */
class WEXP ddgParam
{
public:
	/// The value currently held by this parameter.
    float _value;
	/// The parameters initial value.
    float _default;
	/// The minimum allowed value.
    float _min;
	/// The maximum allowed value.
    float _max;
	/// The step size for incrementing or decrementing.
    float _step;		
	/// The fast step size.
    float _fstep;
	/// The slow step size.
    float _sstep;
	/// If param is linked to an external address. By default this is _value.
	float  *_vaddr;
	/// Callback function to call when value changes.
	void (*_update)(ddgParam *p);
	/// If param is actually a boolean.
    bool  _bool:1;
	/// If param is actually an integer.
    bool  _int:1;
	/// If true, value wrap from max to min.
	bool  _wrap:1;
	/// Incrementing key is a special key.
	int _ikey;
	/// Decrementing key is a special key.
    int _dkey;
	/// The keyboard character associated with param.
    unsigned char _key;
	/// Description which is shown with help call.
	ddgStr		_desc;
	/// Set value.  Either wrapped or clamped.
	float _adjust( float v, float a, float b)
	{ return (_wrap?ddgUtil::wrap(v,a,b): ddgUtil::clamp(v,a,b)); }
public:
	ddgParam() { init(' ',"--",0.0,0.0,1.0,0.1,0.01,1.0); }
    /// Create a paramter with internal storage.
    ddgParam(unsigned char k, char *d, float v, float i = 0.0, float a = 1.0,
		float s = 0.1, float ss = 0.01, float fs = 1.0)
    {
		init(k,d,v,i,a,s,ss,fs);
	}
	/// Create a parameter linked to an address.
    ddgParam(unsigned char k, char *d, float* va, float i = 0.0, float a = 1.0,
		float s = 0.1, float ss = 0.01, float fs = 1.0)
    {
		init(k,d,*va,i,a,s,ss,fs,va);
	}
    /// Set a boolean value in the parameter.
    void setBool(void) { _bool = true; _min = 0.0; _max = 1.0; _step = _sstep = _fstep = 1.0; }
    /// Are we storing a boolean value.
    bool isBool(void)  { return _bool; }
    /// Set an integer value in the parameter.
    void setInt(void)  { _int = true; }
    /// Are we storing an integer value.
    bool isInt(void)   { return _int; }
    /// Get the float value we are storing.
    float get(void)    { return *_vaddr; }
    /// Set the float value and store it.
    float set(float v)
	{
		*_vaddr = ddgUtil::clamp(v,_min,_max);
		doUpdate();
		return *_vaddr;
	}
	/// Initialize all members of the param.
	void init( unsigned char k, char *d = "--",
		float v = 0.0, float i = 0.0, float a = 1.0,
		float s = 0.1, float ss = 0.01, float fs = 1.0, float *vaddr = NULL)
	{
		if (vaddr)
			_vaddr = vaddr;
		else
			_vaddr = &_value;
		_key = k; *_vaddr = _default = v;
        _min = i; _max = a;
        _step = s; _sstep = ss; _fstep = fs;
        _bool = _int = _wrap = false;
		_ikey = _dkey = 0;
		_desc.assign(d);
		_update = NULL;
	}
    /// Reset the parameter to its default value.
    float reset(void)  { return set(_default); }
    /// Increment the parameter by step.
    float incr(void)   { return set(_adjust(get()+_step,_min,_max)); }
    /// Decrement the parameter by step.
    float decr(void)   { return set(_adjust(get()-_step,_min,_max)); }
    /// Slow increment.
    float sincr(void)  { return set(_adjust(get()+_sstep,_min,_max)); }
    /// Slow decrement.
    float sdecr(void)  { return set(_adjust(get()-_sstep,_min,_max)); }
    /// Fast increment.
    float fincr(void)  { return set(_adjust(get()+_fstep,_min,_max)); }
    /// Fast decrement.
    float fdecr(void)  { return set(_adjust(get()-_fstep,_min,_max)); }
	/// Set the max limit.
	void max( float max ) { _max = max; }
	/// Set the min limit.
	void min( float min ) { _min = min; }
    /// Return the key that is bound to this param.
    unsigned char key(void) { return _key; }
	/// Set the description of this parameter.
	void desc( char *d ) { _desc.assign(d); }
	/// Get the description of this parameter.
	char *desc( void ) { return (_desc.s); }
	/// Set the special flag.
	void control( int d, int i = 0) { _ikey = i; _dkey = d; }
	/// Get the incrementing key.
	int controlI( void ) { return _ikey; }
    /// Get the decrementing key.
	int controlD( void ) { return _dkey; }
	/// Enable or disable wrap.
	void wrap( bool w ) { _wrap = w; }
	/// Return wrap state.
	bool wrap( void ) { return _wrap; }
	/// Set the update function.
	void update( void (*u)(ddgParam*p)) { _update = u; }
	/// Call the update function to allow method to perform some work.
	void doUpdate (void ) { if (_update) _update(this); }
}; 

/// Dummy class to explicitly indicate a Float parameter.
class WEXP ddgParamF : public ddgParam
{
public:
    /// Create a paramter with internal storage.
    ddgParamF(unsigned char k, char *d, float v, float i = 0.0, float a = 1.0,
		float s = 0.1, float ss = 0.01, float fs = 1.0) : ddgParam(k,d,v,i,a,s,ss,fs)
    {
	}
};
/**
 * Boolean version of ddgParam.
 */
class WEXP ddgParamB : public ddgParam
{
public:
	/**
	 * Create param controlled by key 'k', described by 'd',
	 * with initial value set to v.
	 */
	ddgParamB(unsigned char k, char *d, bool v ) : ddgParam(k,d,v?1.0:0.0,0.0,1.0,1.0) { setBool(); }
	///
	ddgParamB() { setBool(); }
	///
	void init( unsigned char k, char *d, bool v = false)
	{
		ddgParam::init(k,d,v); setBool();
	}
	///
	void set(bool v) { ddgParam::set(v?1.0:0.0); }
	/// Get the value as a boolean
	bool get(void)
	{
		return (ddgParam::get() != 0.0);
	}
};

/**
 * Interger version of ddgParam.
 */
class WEXP ddgParamI : public ddgParam
{
public:
	/// Create an integer parameter with given values.
	ddgParamI(unsigned char k, char *d, int v, int i, int a, int s = 1, int ss = 1 ) : ddgParam(k,d,v,i,a,s,ss) { setInt(); }
	///
	ddgParamI() { setInt(); }
	/// Initialize the value of the parameter.
	void init( unsigned char k, char *d, int v, int i, int a, int s = 1, int ss = 1, int fs = 10)
	{
		ddgParam::init(k,d,v,i,a,s,ss,fs); setInt();
	}
	/// Return the parameter value as an integer.
	int get(void)
	{
		return (int) ddgParam::get();
	}
};

/**
 * Reset when read version of ddgParamB.
 * This param automatically resets to its default value
 * when ever a value is read from it.
 */
class WEXP ddgParamR : public ddgParamB
{
public:
	/**
	 * Create param controlled by key 'k', described by 'd',
	 * with initial value set to v.
	 */
	bool get(void)
	{
		bool retval = ddgParam::get() != 0.0;
		reset();
		return retval;
	}
};


///
class WEXP ddgParamSetNode : public ddgListNode
{
///
	ddgParam* _p;	/// Pointer to a parameter.
public:
	operator ddgParam* ()	{ return _p; }
	///
	ddgParamSetNode( ddgParam *p ) : _p(p) {}
	///
	ddgParam *p()			{ return _p; }
};

/**
 * Store a list of parameters which can be used by various processing methods.
 * Note no checking is done for duplicate ddgParams (ie they have the same key).
 * The first one found is always returned.
 */
class WEXP ddgParamSet : public ddgList
{
    typedef ddgList super;
public:
	/// Add a parameter to the set.
	void add( ddgParam *p );
    /// Find the param which is associated with character k or key sk.
    ddgParam *find(unsigned char k, int sk ) // Search for parameter k
    {
		ddgParamSetNode *psn = (ddgParamSetNode *) head();
		while( psn && ((k && psn->p()->key() != k)
           ||(sk != psn->p()->controlI()
		   && sk != psn->p()->controlD()))
                ) psn = (ddgParamSetNode *) psn->next();
		return psn ? psn->p() : 0;
    }
    ///  Search for parameter k.
    float get(unsigned char k )
    {
		ddgParamSetNode *psn = (ddgParamSetNode *) head();
		while( psn && psn->p()->key() != k) psn = (ddgParamSetNode *) psn->next();
		return psn ? psn->p()->get() : 0.0f;
    }
    ///
    void report(void)
    {
		ddgParamSetNode *psn = (ddgParamSetNode *) head();
		while(psn)
        {
			ddgParam *p = psn->p();
#ifdef DDGSTREAM
			cerr << '[';
			cerr << (p->key() == 0 ?  '#' : p->key() );
			cerr << ']';
			if (p->isBool())
				cerr << (p->get() > 0.5 ? 'T':'F');
			else if (p->isInt()) cerr << (int) p->get();
			else cerr << (float)p->get();

			cerr << "   " << p->desc() << '\n';
			psn = (ddgParamSetNode *) psn->next();
#endif
        }
    }
	/// Print a nice list of commands.
    void help(void);
    /// Process a key event to update the approriate param.
    bool key(unsigned char key, int x, int y, int sk = 0 );
	// Add the entries in the given param set to this one.
	/// Entries are added at the front of the list.
	void append ( ddgParamSet *ps )
	{
		ddgParamSetNode *psn = (ddgParamSetNode *) ps->head();
		while (psn)
		{
			add(psn->p());
			psn = (ddgParamSetNode *) psn->next();
		}
	}
};
#endif
